#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# generic dimension classes
#

from math import pi, hypot, sqrt, atan2, sin, cos

import text
import point
import circle
import arc
import color
import layer
import units
import util
import tolerance
import options
import globals
import baseobject

class DimString(text.TextFormat):
    """A class for the visual presentation of the dimensional value.

The DimString class is used to present the numerical display for
the dimension. A DimString object is derived from the text.TextFormat
class, so it shares all of that classes methods and attributes.

A DimString class has the following methods

{get/set}Prefix(): Get/Set the text preceding the dimension value.
{get/set}Suffix(): Get/Set the text following the dimension value.
{get/set}Units(): Define what units the dimension will be in.
{get/set}Precision(): Get/Set how many fractional digits are displayed.
{get/set}PrintZero(): Get/Set whether or not a dimension less than one
                      unit long should have a leading 0 printed
{get/set}PrintDecimal(): Get/Set whether or not a dimension with 0
                         fractional digits prints out the decimal point.
                         """
    def __init__(self, prefix=u'', suffix=u'', unit=units.MILLIMETERS,
                 precision=3, print_zero=True, print_decimal=True):
        """Instatiate a DimString object

ds = DimString()

By default, a DimString object has the following values ...

prefix: Empty string
suffix: Empty string
unit: Millimeters
precision: 3 decimal places
print zero: True
print decimal: True
        """
        text.TextFormat.__init__(self)
        self.setPrefix(prefix)
        self.setSuffix(suffix)
        self.__unit = units.Unit()
        self.setUnits(unit)
        self.setPrecision(precision)
        self.setPrintZero(print_zero)
        self.setPrintDecimal(print_decimal)
        # self.__prefix = u''
        # self.__suffix = u''
        # self.__unit = units.Unit(units.MILLIMETERS)
        # self.__precision = 3
        # self.__print_zero = True
        # self.__print_decimal = True

    def getPrefix(self):
        """Return the prefix for the DimString object.

getPrefix()
        """
        return self.__prefix

    def setPrefix(self, prefix=u''):
        """Set the prefix for the DimString object.

setPrefix([prefix])

Invoking this method without arguments sets the prefix
to an empty string. When an argument is passed, the argument
should be a Unicode string.
        """
        _p = prefix
        if not isinstance(_p, unicode):
            _p = unicode(prefix)
        self.__prefix = _p

    prefix = property(getPrefix, setPrefix, None, "Dimension string prefix")

    def getSuffix(self):
        """Return the suffix for the DimString object.

getSuffix()
        """
        return self.__suffix

    def setSuffix(self, suffix=u''):
        """Set the suffix for the DimString object.

setSuffix([suffix])

Invoking this method without arguments sets the suffix
to an empty string. When an argument is passed, the argument
should be a Unicode string.
        """
        _s = suffix
        if not isinstance(_s, unicode):
            _s = unicode(suffix)
        self.__suffix = _s

    suffix = property(getSuffix, setSuffix, None, "Dimension string suffix")

    def getPrecision(self):
        """Return the number of decimal points used for the DimString.

getPrecision()
        """
        return self.__precision

    def setPrecision(self, p):
        """Set the number of decimal points used for the DimString.

setPrecision(p)

The valid range of "p" is 0 <= p <= 15.
        """
        _p = p
        if not isinstance(_p, int):
            _p = int(p)
        if _p < 0 or _p > 15:
            raise ValueError, "Invalid precision: %d" % _p
        self.__precision = _p

    precision = property(getPrecision, setPrecision, None,
                         "Dimension precision")

    def getUnits(self):
        """Return the current units used in the DimString().

getUnits()
        """
        return self.__unit.getUnit()

    def setUnits(self, unit):
        """The the units for the DimString.

setUnits(unit)

The value units are given in the units module.
        """
        self.__unit.setUnit(unit)

    units = property(getUnits, setUnits, None, "Dimensional units.")

    def getPrintZero(self):
        """Return whether or not a leading 0 is printed for the DimString.

getPrintZero()
        """
        return self.__print_zero

    def setPrintZero(self, pz=True):
        """Set whether or not a leading 0 is printed for the DimString.

setPrintZero([pz])

Invoking this method without arguments sets the value to True.
If called with an argument, the argument should be either True
or False.
        """
        if pz is True or pz is False:
            self.__print_zero = pz
        else:
            raise ValueError, "Invalid zero print flag: " + `pz`

    print_zero = property(getPrintZero, setPrintZero, None,
                          "Print a leading 0 for decimal dimensions")

    def getPrintDecimal(self):
        """Return whether or not the DimString will print a trailing decimal.

getPrintDecimal()
        """
        return self.__print_decimal

    def setPrintDecimal(self, pd=True):
        """Set whether or not the DimString will print a trailing decimal.

setPrintDecimal([pd])

Invoking this method without arguments sets the value to True.
If called with an argument, the argument should be either True
or False.
        """
        if pd is True or pd is False:
            self.__print_decimal = pd
        else:
            raise ValueError, "Invalid print decimal flag: " + `pd`

    print_decimal = property(getPrintDecimal, setPrintDecimal, None,
                             "Print a decimal point after the dimension value")

    def formatDimension(self, dist):
        """Return a formatted numerical value for a dimension.

formatDimension(dist)

The argument "dist" should be a float value representing the
distance in millimeters. The returned value will have the
prefix prepended and suffix appended to the numerical value
that has been formatted with the precision.
        """
        _d = dist
        if not isinstance(_d, float):
            _d = float(dist)
        _d = abs(_d)
        _fmtstr = u"%%#.%df" % self.__precision
        _dstr = _fmtstr % self.__unit.fromMillimeters(_d)
        if _d < 1.0 and self.__print_zero is False:
            _dstr = _dstr[1:]
        if _dstr.endswith('.') and self.__print_decimal is False:
            _dstr = _dstr[:-1]
        return self.__prefix + _dstr + self.__suffix

class DimBar(object):
    """The class for the dimension bar.

A dimension bar leads from the point the dimension references
out to, and possibly beyond, the point where the dimension
text bar the DimBar to another DimBar. Linear,
horizontal, vertical, and angular dimension will have two
dimension bars; radial dimensions have none.

The DimBar class has the following methods:

getEndpoints(): Get the x/y position of the DimBar start and end
{get/set}FirstEndpoint(): Get/Set the starting x/y position of the DimBar.
{get/set}SecondEndpoint(): Get/Set the ending x/y position of the DimBar.
getAngle(): Get the angle at which the DimBar slopes
getSinCosValues(): Get trig values used for transformation calculations.
    """
    def __init__(self, x1=0.0, y1=0.0, x2=0.0, y2=0.0):
        """Initialize a DimBar.

db = DimBar([x1, y1, x2, y2])

By default all the arguments are 0.0. Any arguments passed to this
method should be float values.
        """
        _x1 = x1
        if not isinstance(_x1, float):
            _x1 = float(x1)
        _y1 = y1
        if not isinstance(_y1, float):
            _y1 = float(y1)
        _x2 = x2
        if not isinstance(_x2, float):
            _x2 = float(x2)
        _y2 = y2
        if not isinstance(_y2, float):
            _y2 = float(y2)
        self.__ex1 = _x1
        self.__ey1 = _y1
        self.__ex2 = _x2
        self.__ey2 = _y2

    def getEndpoints(self):
        """Return the coordinates of the DimBar endpoints.

getEndpoints()

This method returns two tuples, each containing two float values.
The first tuple gives the x/y coordinates of the DimBar start,
the second gives the coordinates of the DimBar end.
        """
        _ep1 = (self.__ex1, self.__ey1)
        _ep2 = (self.__ex2, self.__ey2)
        return _ep1, _ep2

    def setFirstEndpoint(self, x, y):
        """Set the starting coordinates for the DimBar

setFirstEndpoint(x, y)

Arguments "x" and "y" should be float values.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.__ex1 = _x
        self.__ey1 = _y

    def getFirstEndpoint(self):
        """Return the starting coordinates of the DimBar.

getFirstEndpoint()

This method returns a tuple giving the x/y coordinates.
        """
        return self.__ex1, self.__ey1

    def setSecondEndpoint(self, x, y):
        """Set the ending coordinates for the DimBar

setSecondEndpoint(x, y)

Arguments "x" and "y" should be float values.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.__ex2 = _x
        self.__ey2 = _y

    def getSecondEndpoint(self):
        """Return the ending coordinates of the DimBar.

getSecondEndpoint()

This method returns a tuple giving the x/y coordinates.
        """
        return self.__ex2, self.__ey2

    def getAngle(self):
        """Return the angle at which the DimBar lies.

getAngle()

This method returns a float value giving the angle of inclination
of the DimBar.

The value returned will be a positive value less than 360.0.
        """
        _x1 = self.__ex1
        _y1 = self.__ey1
        _x2 = self.__ex2
        _y2 = self.__ey2
        if ((abs(_x2 - _x1) < 1e-10) and
            (abs(_y2 - _y1) < 1e-10)):
            raise ValueError, "Endpoints are equal"
        if abs(_x2 - _x1) < 1e-10: # vertical
            if _y2 > _y1:
                _angle = 90.0
            else:
                _angle = 270.0
        elif abs(_y2 - _y1) < 1e-10: # horizontal
            if _x2 > _x1:
                _angle = 0.0
            else:
                _angle = 180.0
        else:
            _angle = 180.0/pi * atan2((_y2 - _y1), (_x2 - _x1))
            if _angle < 0.0:
                _angle = _angle + 360.0
        return _angle

    def getSinCosValues(self):
        """Return sin()/cos() values based on the DimBar slope

getSinCosValues()

This method returns a tuple of two floats. The first value is
the sin() value, the second is the cos() value.
        """
        _x1 = self.__ex1
        _y1 = self.__ey1
        _x2 = self.__ex2
        _y2 = self.__ey2
        if abs(_x2 - _x1) < 1e-10: # vertical
            _cosine = 0.0
            if _y2 > _y1:
                _sine = 1.0
            else:
                _sine = -1.0
        elif abs(_y2 - _y1) < 1e-10: # horizontal
            _sine = 0.0
            if _x2 > _x1:
                _cosine = 1.0
            else:
                _cosine = -1.0
        else:
            _angle = atan2((_y2 - _y1), (_x2 - _x1))
            _sine = sin(_angle)
            _cosine = cos(_angle)
        return _sine, _cosine

class DimCrossbar(DimBar):
    """The class for the Dimension crossbar.

The DimCrossbar class is drawn between two DimBar objects for
horizontal, vertical, and generic linear dimensions. The dimension
text is place over the DimCrossbar object. Arrow heads, circles, or
slashes can be drawn at the intersection of the DimCrossbar and
the DimBar if desired. These objects are called markers.

The DimCrossbar class is derived from the DimBar class so it shares
all the methods of that class. In addition the DimCrossbar class has
the following methods:

{set/get}FirstCrossbarPoint(): Set/Get the initial location of the crossbar.
{set/get}SecondCrossbarPoint(): Set/Get the ending location of the crossbar.
getCrossbarPoints(): Get the location of the crossbar endpoints.
clearMarkerPoints(): Delete the stored coordintes of the dimension markers.
storeMarkerPoint(): Save a coordinate pair of the dimension marker.
getMarkerPoints(): Return the coordinates of the dimension marker.
    """
    def __init__(self):
        """Initialize a DimCrossbar object.

dcb = DimCrossbar()

This method takes no arguments.
        """
        DimBar.__init__(self)
        self.__mx1 = 0.0
        self.__my1 = 0.0
        self.__mx2 = 0.0
        self.__my2 = 0.0
        self.__mpts = []

    def setFirstCrossbarPoint(self, x, y):
        """Store the initial endpoint of the DimCrossbar.

setFirstCrossbarPoint(x, y)

Arguments "x" and "y" should be floats.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.__mx1 = _x
        self.__my1 = _y

    def getFirstCrossbarPoint(self):
        """Return the initial coordinates of the DimCrossbar.

getFirstCrossbarPoint()

This method returns a tuple of two floats giving the x/y coordinates.
        """
        return self.__mx1, self.__my1

    def setSecondCrossbarPoint(self, x, y):
        """Store the terminal endpoint of the DimCrossbar.

setSecondCrossbarPoint(x, y)

Arguments "x" and "y" should be floats.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.__mx2 = _x
        self.__my2 = _y

    def getSecondCrossbarPoint(self):
        """Return the terminal coordinates of the DimCrossbar.

getSecondCrossbarPoint()

This method returns a tuple of two floats giving the x/y coordinates.
        """
        return self.__mx2, self.__my2

    def getCrossbarPoints(self):
        """Return the endpoints of the DimCrossbar.

getCrossbarPoints()

This method returns two tuples, each tuple containing two float
values giving the x/y coordinates.
        """
        _mp1 = (self.__mx1, self.__my1)
        _mp2 = (self.__mx2, self.__my2)
        return _mp1, _mp2

    def clearMarkerPoints(self):
        """Delete the stored location of any dimension markers.

clearMarkerPoints()
        """
        del self.__mpts[:]

    def storeMarkerPoint(self, x, y):
        """Save a coordinate pair of the current dimension marker.

storeMarkerPoint(x, y)

Arguments "x" and "y" should be floats. Each time this method is invoked
the list of stored coordinates is appended with the values given as
arguments.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.__mpts.append((_x, _y))

    def getMarkerPoints(self):
        """Return the stored marker coordinates.

getMarkerPoints()

This method returns a list of coordinates stored with storeMarkerPoint().
Each item in the list is a tuple holding two float values - the x/y
coordinate of the point.
        """
        return self.__mpts[:]

class DimCrossarc(DimCrossbar):
    """The class for specialized crossbars for angular dimensions.

The DimCrossarc class is meant to be used only with angular dimensions.
As an angular dimension has two DimBar objects that are connected
with an arc. The DimCrossarc class is derived from the DimCrossbar
class so it shares all the methods of that class. The DimCrossarc
class has the following additional methods:

{get/set}Radius(): Get/Set the radius of the arc.
{get/set}StartAngle(): Get/Set the arc starting angle.
{get/set}EndAngle(): Get/Set the arc finishing angle.
    """
    def __init__(self, radius=0.0, start=0.0, end=0.0):
        """Initialize a DimCrossarc object.

dca = DimCrossarc([radius, start, end])

By default the arguments are all 0.0. Any arguments passed to
this method should be floats.
        """
        DimCrossbar.__init__(self)
        _r = radius
        if not isinstance(_r, float):
            _r = float(radius)
        if _r < 0.0:
            raise ValueError, "Invalid radius: %g" % _r
        _start = util.make_c_angle(start)
        _end = util.make_c_angle(end)
        self.__radius = _r
        self.__start = _start
        self.__end = _end

    def getRadius(self):
        """Return the radius of the arc.

getRadius()

This method returns a float value.
        """
        return self.__radius

    def setRadius(self, radius):
        """Set the radius of the arc.

setRadius(radius)

Argument "radius" should be a float value greater than 0.0.
        """
        _r = radius
        if not isinstance(_r, float):
            _r = float(radius)
        if _r < 0.0:
            raise ValueError, "Invalid radius: %g" % _r
        self.__radius = _r

    def getStartAngle(self):
        """Return the arc starting angle.

getStartAngle()

This method returns a float.
        """
        return self.__start

    def setStartAngle(self, angle):
        """Set the starting angle of the arc.

setStartAngle(angle)

Argument "angle" should be a float value.
        """
        _angle = util.make_c_angle(angle)
        self.__start = _angle

    def getEndAngle(self):
        """Return the arc ending angle.

getEndAngle()

This method returns a float.
        """
        return self.__end

    def setEndAngle(self, angle):
        """Set the ending angle of the arc.

setEndAngle(angle)

Argument "angle" should be a float value.
        """
        _angle = util.make_c_angle(angle)
        self.__end = _angle

    def getAngle(self):
        pass # override the DimBar::getAngle() method

    def getSinCosValues(self):
        pass # override the DimBar::getSinCosValues() method

class Dimension(object):
    """The base class for Dimensions

A Dimension object is meant to be a base class for specialized
dimensions.

Every Dimension object holds two DimString objects, so any
dimension can be displayed with two separate formatting options
and units.

A Dimension has the following methods

{get/set}DimStyle(): Get/Set the DimStyle used for this Dimension.
applyDimStyle(): Set all the Dimension attributes to the values given
                 in the DimStyle
getPrimaryDimstring(): Return the DimString used for formatting the
                       Primary dimension.
getSecondaryDimstring(): Return the DimString used for formatting the
                         Secondary dimension.
{get/set}Endpoint(): Get/Set the type of endpoints used in the Dimension
{get/set}EndpointSize(): Get/Set the size of the dimension endpoints
{get/set}DualDimMode(): Get/Set whether or not to display both the Primary
                        and Secondary DimString objects
{get/set}Offset(): Get/Set how far from the dimension endpoints to draw
                   dimension lines at the edges of the dimension.
{get/set}Extension(): Get/Set how far past the dimension crossbar line
                      to draw.
{get/set}Position(): Get/Set where the dimensional values are placed on the
                     dimension cross bar.
{get/set}Color(): Get/Set the color used to draw the dimension lines.
{get/set}Location(): Get/Set where to draw the dimensional values.
getDimensions(): Return the formatted dimensional values in this Dimension.
isModified(): Return if the dimension has had its endpoints adjusted.
modified(): Flag the dimension endpoints or circle have been modified.
reset(): Set the dimension state to be unmodified.
inRegion(): Return if the dimension is visible within some are.
calcDimValues(): Calculate the dimension lines endpoints.
onDimension(): Test if an x/y coordinate pair hit the dimension lines.
getBounds(): Return the minma and maximum locations of the dimension.
    """
    #
    # Endpoint
    #
    NO_ENDPOINT = 0
    ARROW = 1
    FILLED_ARROW = 2
    SLASH = 3
    CIRCLE = 4

    #
    # Dimension position on dimline
    #
    SPLIT = 0
    ABOVE = 1
    BELOW = 2

    __defcolor = color.Color(255,165,0)

    def __init__(self, ds, dual_mode=False, offset=0.0, extension=0.0,
                 position=SPLIT, endpoint_type=NO_ENDPOINT,
                 endpoint_size=0.0, col=None, x=0.0, y=0.0):
        """Instatiate a Dimension object

dim = Dimension(ds)

A Dimension object requires a DimStyle object to construct
itself. The DimStyle object provides a collection of values
used to format the two DimString objects and the various other
parts of a dimension.
        """
        self.__dimstring1 = DimString()
        self.__dimstring2 = DimString()
        self.setDimStyle(ds)
        self.setDualDimMode(dual_mode)
        self.setOffset(offset)
        self.setExtension(extension)
        self.setPosition(position)
        self.setEndpointType(endpoint_type)
        self.setEndpointSize(endpoint_size)
        _col = col
        if _col is None:
            _col = Dimension.__defcolor
        self.setColor(_col)
        self.setLocation(x, y)
        if self.__class__ is Dimension:
            self.applyDimStyle()
            self.reset()
        # if not isinstance(ds, DimStyle):
            # raise TypeError, "Invalid DimStyle: " + `ds`
        # self.__dimstyle = ds
        # self.__dual_mode = False # Display primary dimension only
        # self.__offset = 0.0
        # self.__extension = 0.0
        # self.__dim_pos = Dimension.SPLIT
        # self.__endpoint_type = Dimension.NO_ENDPOINT
        # self.__endpoint_size = 1.0
        # self.__color = dim_color
        # self.__dim_loc_x = None
        # self.__dim_loc_y = None
        # self.__modified = False

    def getDimStyle(self):
        """Return the DimStyle used in this Dimension.

getDimStyle()
        """
        return self.__dimstyle

    def setDimStyle(self, ds):
        """Set the DimStyle used for this Dimension.

setDimStyle(ds)

After setting the DimStyle, the values stored in it
are applied to the DimensionObject.
        """
        if not isinstance(ds, DimStyle):
            raise TypeError, "Invalid DimStyle: " + `ds`
        self.__dimstyle = ds

    dimstyle = property(getDimStyle, setDimStyle, None,
                        "Dimension DimStyle object.")

    def applyDimStyle(self):
        """Apply the values in the DimStyle to the Dimension.

applyDimStyle()
        """
        _ds = self.__dimstyle
        self.setOffset(_ds.getValue('DIM_OFFSET'))
        self.setExtension(_ds.getValue('DIM_EXTENSION'))
        self.setPosition(_ds.getValue('DIM_POSITION'))
        self.setEndpointType(_ds.getValue('DIM_ENDPOINT'))
        self.setEndpointSize(_ds.getValue('DIM_ENDPOINT_SIZE'))
        self.setColor(_ds.getValue('DIM_COLOR'))
        self.setDualDimMode(_ds.getValue('DIM_DUAL_MODE'))
        #
        # primary dimension string
        #
        _pds = self.getPrimaryDimstring()
        _pds.setPrefix(_ds.getValue('DIM_PRIMARY_PREFIX'))
        _pds.setSuffix(_ds.getValue('DIM_PRIMARY_SUFFIX'))
        _pds.setPrecision(_ds.getValue('DIM_PRIMARY_PRECISION'))
        _pds.setUnits(_ds.getValue('DIM_PRIMARY_UNITS'))
        _pds.setPrintZero(_ds.getValue('DIM_PRIMARY_LEADING_ZERO'))
        _pds.setPrintDecimal(_ds.getValue('DIM_PRIMARY_TRAILING_DECIMAL'))
        _pds.setFamily(_ds.getValue('DIM_PRIMARY_FONT_FAMILY'))
        _pds.setSize(_ds.getValue('DIM_PRIMARY_FONT_SIZE'))
        _pds.setWeight(_ds.getValue('DIM_PRIMARY_FONT_WEIGHT'))
        _pds.setStyle(_ds.getValue('DIM_PRIMARY_FONT_STYLE'))
        _pds.setColor(_ds.getValue('DIM_PRIMARY_FONT_COLOR'))
        #
        # secondary dimension string
        #
        _sds = self.getSecondaryDimstring()
        _sds.setPrefix(_ds.getValue('DIM_SECONDARY_PREFIX'))
        _sds.setSuffix(_ds.getValue('DIM_SECONDARY_SUFFIX'))
        _sds.setPrecision(_ds.getValue('DIM_SECONDARY_PRECISION'))
        _sds.setUnits(_ds.getValue('DIM_SECONDARY_UNITS'))
        _sds.setPrintZero(_ds.getValue('DIM_SECONDARY_LEADING_ZERO'))
        _sds.setPrintDecimal(_ds.getValue('DIM_SECONDARY_TRAILING_DECIMAL'))
        _sds.setFamily(_ds.getValue('DIM_SECONDARY_FONT_FAMILY'))
        _sds.setSize(_ds.getValue('DIM_SECONDARY_FONT_SIZE'))
        _sds.setWeight(_ds.getValue('DIM_SECONDARY_FONT_WEIGHT'))
        _sds.setStyle(_ds.getValue('DIM_SECONDARY_FONT_STYLE'))
        _sds.setColor(_ds.getValue('DIM_SECONDARY_FONT_COLOR'))
        self.modified()

    def getEndpointType(self):
        """Return what type of endpoints the Dimension uses.

getEndpointType()
        """
        return self.__endpoint_type

    def setEndpointType(self, e):
        """Set what type of endpoints the Dimension will use.

setEndpointType(e)

The argument "e" should be one of the following

dimension.NO_ENDPOINT => no special marking at the dimension crossbar ends
dimension.ARROW => an arrowhead at the dimension crossbar ends
dimension.FILLED_ARROW => a filled arrohead at the dimension crossbar ends
dimension.SLASH => a slash mark at the dimension crossbar ends
dimension.CIRCLE => a filled circle at the dimension crossbar ends
        """
        if (e != Dimension.NO_ENDPOINT and
            e != Dimension.ARROW and
            e != Dimension.FILLED_ARROW and
            e != Dimension.SLASH and
            e != Dimension.CIRCLE):
            raise ValueError, "Invalid endpoint value: '%s'" % str(e)
        self.__endpoint_type = e
        self.modified()

    endpoint = property(getEndpointType, setEndpointType,
                        None, "Dimension endpoint type.")

    def getEndpointSize(self):
        """Return the size of the Dimension endpoints.

getEndpointSize()
        """
        return self.__endpoint_size

    def setEndpointSize(self, size):
        """Set the size of the Dimension endpoints.

setEndpointSize(size)

The argument size should be a float greater than or
equal to 0.0
        """
        _size = size
        if not isinstance(_size, float):
            _size = float(size)
        if _size < 0.0:
            raise ValueError, "Invalid endpoint size: %g" % _size
        self.__endpoint_size = _size
        self.modified()

    def getPrimaryDimstring(self):
        """ Return the DimString used for formatting the primary dimension.

getPrimaryDimstring()
        """
        return self.__dimstring1

    def getSecondaryDimstring(self):
        """Return the DimString used for formatting the secondary dimension.

getSecondaryDimstring()
        """
        return self.__dimstring2

    def getDualDimMode(self):
        """Return if the Dimension is displaying primary and secondary values.

getDualDimMode(self)
        """
        return self.__dual_mode

    def setDualDimMode(self, mode):
        """Set the Dimension to display both primary and secondary values.

setDualDimMode(mode)

The argument "mode" should be True or False.
        """
        if mode is True or mode is False:
            self.__dual_mode = mode
            self.modified()
        else:
            raise ValueError, "Invalid dual dimension mode: '%s'" % str(mode)

    dual_mode = property(getDualDimMode, setDualDimMode, None,
                         "Display both primary and secondary dimensions")

    def getOffset(self):
        """Return the current offset value for the Dimension.

getOffset()
        """
        return self.__offset

    def setOffset(self, offset=0.0):
        """Set the offset value for the Dimension.

setOffset([offset])

Calling this method without arguments sets the value to 0.
If the argument "offset" is supplied, it should be a positive
float value.
        """
        _o = offset
        if not isinstance(_o, float):
            _o = float(offset)
        if _o < 0.0:
            raise ValueError, "Invalid dimension offset length: %g" % _o
        self.__offset = _o
        self.modified()

    offset = property(getOffset, setOffset, None, "Dimension offset.")

    def getExtension(self):
        """Get the extension length of the Dimension.

getExtension()
        """
        return self.__extension

    def setExtension(self, ext=0.0):
        """Set the extension length of the Dimension.

setExtension([ext])

If this method is called without arguments, the extension
length is set to 0. If the argument "ext" is supplied, it
should be a positive float value.
        """
        _e = ext
        if not isinstance(_e, float):
            _e = float(ext)
        if _e < 0.0:
            raise ValueError, "Invalid dimension extension length: %g" % _e
        self.__extension = _e
        self.modified()

    extension = property(getExtension, setExtension, None,
                         "Dimension extension length.")

    def getPosition(self):
        """Return how the dimension text intersects the crossbar.

getPosition()
        """
        return self.__dim_pos

    def setPosition(self, pos):
        """Set where the dimension text should be placed at the crossbar.

setPosition(pos)

The choices are:

dimension.SPLIT => In the middle of the crossbar.
dimension.ABOVE => Beyond the crossbar from the dimensioned objects.
dimension.BELOW => Between the crossbar and the dimensioned objects.
        """
        if (pos != Dimension.SPLIT and
            pos != Dimension.ABOVE and
            pos != Dimension.BELOW):
            raise ValueError, "Invalid dimension text position: '%s'" % str(pos)
        self.__dim_pos = pos
        self.modified()

    position = property(getPosition, setPosition, None,
                        "Dimension text position")

    def getColor(self):
        """Return the color of the dimension lines.

getColor()
        """
        return self.__color

    def setColor(self, c):
        """Set the color of the dimension lines.

setColor(c)
        """
        _c = c
        if not isinstance(_c, color.Color):
            _c = color.Color(c)
        self.__color = _c

    color = property(getColor, setColor, None, "Dimension Color")

    def getLocation(self):
        """Return the location of the dimensional text values.

getLocation()
        """
        return (self.__dim_loc_x, self.__dim_loc_y)

    def setLocation(self, x, y):
        """Set the location of the dimensional text values.

setLocation(x, y)

The "x" and "y" arguments should be float values. The text is
centered around that point.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.__dim_loc_x = _x
        self.__dim_loc_y = _y
        self.modified()

    location = property(getLocation, setLocation, None,
                        "Dimension location")

    def getDimensions(self, dimlen):
        """Return the formatted dimensional values.

getDimensions(dimlen)

The argument "dimlen" should be the length in millimeters.

This method returns a list of the primary and secondary
dimensional values.
        """
        _dl = dimlen
        if not isinstance(dimlen, float):
            _dl = float(dimlen)
        dims = []
        dims.append(self.__dimstring1.formatDimension(_dl))
        dims.append(self.__dimstring2.formatDimension(_dl))
        return dims

    def isModified(self):
        """Return if the dimension has adjusted itself.

isModified()
        """
        return self.__modified

    def modified(self):
        """Set the dimension to be in the modified state.

modified()
        """
        self.__modified = True

    def reset(self):
        """Set the dimension to be in the unmodified state.

reset()
        """
        self.__modified = False

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a Dimension exists with a region.

isRegion(xmin, ymin, xmax, ymax[, fully])

The first four arguments define the boundary. The optional
fifth argument "fully" indicates whether or not the Dimension
must be completely contained within the region or just pass
through it.

This method should be overriden in classes derived from Dimension.
        """
        return False

    def getBounds(self):
        """Return the minimal and maximal locations of the dimension

getBounds()

This method returns a tuple of four values - xmin, ymin, xmax, ymax.
These values give the mimimum and maximum coordinates of the dimension
object.

This method should be overriden in classes derived from Dimension.
        """
        return None, None, None, None # override

class LinearDimension(Dimension):
    """A class for Linear dimensions.

The LinearDimension class is derived from the Dimension
class, so it shares all of those methods and attributes.
A LinearDimension should be used to display the absolute
distance between two Point objects.

A LinearDimension object has the following methods:

{get/set}P1(): Get/Set the first Point for the LinearDimension.
{get/set}P2(): Get/Set the second Point for the LinearDimension.
getDimPoints(): Return the two Points used in this dimension.
getDimLayers(): Return the two Layers holding the Points.
getDimXPoints(): Get the x-coordinates of the dimension bar positions.
getDimYPoints(): Get the y-coordinates of the dimension bar positions.
getDimMarkerPoints(): Get the locaiton of the dimension endpoint markers.
calcMarkerPoints(): Calculate the coordinates of any dimension marker objects.
    """
    def __init__(self, l1, p1, l2, p2, dx, dy, ds):
        """Instantiate a LinearDimension object.

ldim = LinearDimension(l1, p1, l2, p2, dx, dy, ds)

l1: A Layer holding Point p1
p1: A Point contained in Layer l1
l2: A Layer holding Point p2
p2: A Point contained in Layer l2
dx: The "x" coordinate of the dimensional text
dy: The "y" coordinate of the dimensional text
ds: The DimStyle used for this Dimension.
        """
        Dimension.__init__(self, ds, x=dx, y=dy)
        if not isinstance(l1, layer.Layer):
            raise TypeError, "Invalid layer: " + `l1`
        if not isinstance(p1, point.Point):
            raise TypeError, "Invalid point: " + `p1`
        _lp = l1.findObject(p1)
        if _lp is not p1:
            raise ValueError, "Point p1 not found in layer!"
        if not isinstance(l2, layer.Layer):
            raise TypeError, "Invalid layer: " + `l2`
        if not isinstance(p2, point.Point):
            raise TypeError, "Invalid point: " + `p2`
        _lp = l2.findObject(p2)
        if _lp is not p2:
            raise ValueError, "Point p2 not found in layer!"
        self.__l1 = l1
        self.__p1 = p1
        self.__l2 = l2
        self.__p2 = p2
        self.__bar1 = DimBar()
        self.__bar2 = DimBar()
        self.__crossbar = DimCrossbar()
        self.applyDimStyle()
        p1.storeUser(self)
        p2.storeUser(self)

    def __eq__(self, ldim):
        """Test two LinearDimension objects for equality.
        """
        if not isinstance(ldim, LinearDimension):
            raise TypeError, "Invalid object for comparison: " + `ldim`
        _val = False
        _l1, _l2 = ldim.getDimLayers()
        _p1, _p2 = ldim.getDimPoints()
        if (self.__l1 is _l1 and
            self.__l2 is _l2 and
            self.__p1 == _p1 and
            self.__p2 == _p2):
            _val = True
        elif (self.__l1 is _l2 and
              self.__l2 is _l1 and
              self.__p1 == _p2 and
              self.__p2 == _p1):
            _val = True
        return _val

    def __ne__(self, ldim):
        """Test two LinearDimension objects for equality.
        """
        if not isinstance(ldim, LinearDimension):
            raise TypeError, "Invalid object for comparison: " + `ldim`
        _val = True
        _l1, _l2 = ldim.getDimLayers()
        _p1, _p2 = ldim.getDimPoints()
        if (self.__l1 is _l1 and
            self.__l2 is _l2 and
            self.__p1 == _p1 and
            self.__p2 == _p2):
            _val = True
        elif (self.__l1 is _l2 and
              self.__l2 is _l1 and
              self.__p1 == _p2 and
              self.__p2 == _p1):
            _val = False
        return _val

    def getP1(self):
        """Return the first Point of a LinearDimension.

getP1()
        """
        return self.__p1

    def setP1(self, l, p):
        """Set the first Point of a LinearDimension.

setP1(l, p)

There are two required arguments for this method

l: The layer holding the Point p
p: A point in Layer l
        """
        if not isinstance(l, layer.Layer):
            raise TypeError, "Invalid layer: " + `l`
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _lp = l.findObject(p)
        if _lp is not p:
            raise ValueError, "Point not found in layer!"
        _p1 = self.__p1
        self.__l1 = l
        self.__p1 = p
        self.modified()
        _p1.freeUser(self)
        p.storeUser(self)

    p1 = property(getP1, None, None, "Dimension first point.")

    def getP2(self):
        """Return the second point of a LinearDimension.

getP2()
        """
        return self.__p2

    def setP2(self, l, p):
        """Set the second Point of a LinearDimension.

setP2(l, p)

There are two required arguments for this method

l: The layer holding the Point p
p: A point in Layer l
        """
        if not isinstance(l, layer.Layer):
            raise TypeError, "Invalid layer: " + `l`
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _lp = l.findObject(p)
        if _lp is not p:
            raise ValueError, "Point not found in layer!"
        _p2 = self.__p2
        self.__l2 = l
        self.__p2 = p
        self.modified()
        _p2.freeUser(self)
        p.storeUser(self)

    p2 = property(getP2, None, None, "Dimension second point.")

    def getDimPoints(self):
        """Return both points used in the LinearDimension.

getDimPoints()

The two points are returned in a tuple.
        """
        return self.__p1, self.__p2

    def getDimBars(self):
        """Return the dimension boundary bars.

getDimBars()
        """
        return self.__bar1, self.__bar2

    def getDimCrossbar(self):
        """Return the dimension crossbar.

getDimCrossbar()
        """
        return self.__crossbar

    def getDimLayers(self):
        """Return both layers used in the LinearDimension.

getDimLayers()

The two layers are returned in a tuple.
        """
        return (self.__l1, self.__l2)

    def calculate(self):
        """Determine the length of this LinearDimension.

calculate()
        """
        return self.__p1 - self.__p2

    def applyDimStyle(self):
        """Apply the values in the DimStyle to the LinearDimension.

applyDimStyle()

This method extends the Dimension::applyDimStyle() method.
        """
        Dimension.applyDimStyle(self)
        self.modified()

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a LinearDimension exists within a region.

isRegion(xmin, ymin, xmax, ymax[, fully])

The four arguments define the boundary of an area, and the
function returns True if the LinearDimension lies within that
area. If the optional argument "fully" is used and is True,
then the dimension points and the location of the dimension
text must lie within the boundary. Otherwise, the function
returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _pxmin, _pymin, _pxmax, _pymax = self.getBounds()
        _val = False
        if (_xmin < _pxmin and
            _xmax > _pxmax and
            _ymin < _pymin and
            _ymax > _pymax):
            _val = True
        elif (not fully and
              not (_pxmax < _xmin or _pymax < _ymin or
                   _pxmin > _xmax or _pymin > _ymax)):
            _x1, _y1 = self.__p1.getCoords()
            _x2, _y2 = self.__p2.getCoords()
            _dx, _dy = self.getLocation()
            if _xmin < _x1 < _xmax and _ymin < _y1 < _ymax: # p1 point
                _val = True
            elif _xmin < _x2 < _xmax and _ymin < _y2 < _ymax: # p2 point
                _val = True
            elif _xmin < _dx < _xmax and _ymin < _dy < _ymax: # dim text
                _val = True
            else:
                #
                # bar at p1
                #
                _ep1, _ep2 = self.__bar1.getEndpoints()
                _val =_test_dim_line(_ep1, _ep2, _xmin, _ymin, _xmax, _ymax)
                if _val:
                    return _val
                #
                # bar at p2
                #
                _ep1, _ep2 = self.__bar2.getEndpoints()
                _val =_test_dim_line(_ep1, _ep2, _xmin, _ymin, _xmax, _ymax)
                if _val:
                    return _val
                #
                # crossbar
                #
                _ep1, _ep2 = self.__crossbar.getEndpoints()
                _val =_test_dim_line(_ep1, _ep2, _xmin, _ymin, _xmax, _ymax)
        return _val

    def calcDimValues(self, allpts=True):
        """Recalculate the values for dimensional display.

calcDimValues([allpts])

This method calculates where the points for the dimension
bars and crossbar are located. The argument "allpts" is
optional. By default it is True. If the argument is set to
False, then the coordinates of the dimension marker points
will not be calculated.
        """
        _allpts = allpts
        if _allpts is not True and _allpts is not False:
            raise ValueError, "Invalid argument: " + str(allpts)
        _p1, _p2 = self.getDimPoints()
        _bar1 = self.__bar1
        _bar2 = self.__bar2
        _crossbar = self.__crossbar
        _p1x, _p1y = _p1.getCoords()
        _p2x, _p2y = _p2.getCoords()
        _dx, _dy = self.getLocation()
        _offset = self.getOffset()
        _ext = self.getExtension()
        #
        # see comp.graphics.algorithms.faq section on calcuating
        # the distance between a point and line for info about
        # the following equations ...
        #
        _dpx = _p2x - _p1x
        _dpy = _p2y - _p1y
        _rnum = ((_dx - _p1x) * _dpx) + ((_dy - _p1y) * _dpy)
        _snum = ((_p1y - _dy) * _dpx) - ((_p1x - _dx) * _dpy)
        _den = pow(_dpx, 2) + pow(_dpy, 2)
        _r = _rnum/_den
        _s = _snum/_den
        _sep = abs(_s) * sqrt(_den)
        if abs(_dpx) < 1e-10: # vertical
            if _p2y > _p1y:
                _slope = pi/2.0
            else:
                _slope = -pi/2.0
        elif abs(_dpy) < 1e-10: # horizontal
            if _p2x > _p1x:
                _slope = 0.0
            else:
                _slope = -pi
        else:
            _slope = atan2(_dpy, _dpx)
        if _s < 0.0: # dim point left of p1-p2 line
            _angle = _slope + (pi/2.0)
        else: # dim point right of p1-p2 line (or on it)
            _angle = _slope - (pi/2.0)
        _sin_angle = sin(_angle)
        _cos_angle = cos(_angle)
        _x = _p1x + (_offset * _cos_angle)
        _y = _p1y + (_offset * _sin_angle)
        _bar1.setFirstEndpoint(_x, _y)
        if _r < 0.0:
            _px = _p1x + (_r * _dpx)
            _py = _p1y + (_r * _dpy)
            _x = _px + (_sep * _cos_angle)
            _y = _py + (_sep * _sin_angle)
        else:
            _x = _p1x + (_sep * _cos_angle)
            _y = _p1y + (_sep * _sin_angle)
        _crossbar.setFirstEndpoint(_x, _y)
        _x = _p1x + (_sep * _cos_angle)
        _y = _p1y + (_sep * _sin_angle)
        _crossbar.setFirstCrossbarPoint(_x, _y)
        _x = _p1x + ((_sep + _ext) * _cos_angle)
        _y = _p1y + ((_sep + _ext) * _sin_angle)
        _bar1.setSecondEndpoint(_x, _y)
        _x = _p2x + (_offset * _cos_angle)
        _y = _p2y + (_offset * _sin_angle)
        _bar2.setFirstEndpoint(_x, _y)
        if _r > 1.0:
            _px = _p1x + (_r * _dpx)
            _py = _p1y + (_r * _dpy)
            _x = _px + (_sep * _cos_angle)
            _y = _py + (_sep * _sin_angle)
        else:
            _x = _p2x + (_sep * _cos_angle)
            _y = _p2y + (_sep * _sin_angle)
        _crossbar.setSecondEndpoint(_x, _y)
        _x = _p2x + (_sep * _cos_angle)
        _y = _p2y + (_sep * _sin_angle)
        _crossbar.setSecondCrossbarPoint(_x, _y)
        _x = _p2x + ((_sep + _ext) * _cos_angle)
        _y = _p2y + ((_sep + _ext) * _sin_angle)
        _bar2.setSecondEndpoint(_x, _y)
        if _allpts:
            self.calcMarkerPoints()

    def calcMarkerPoints(self):
        """Calculate and store the dimension endpoint markers coordinates.

calcMarkerPoints()
        """
        _type = self.getEndpointType()
        _crossbar = self.__crossbar
        _crossbar.clearMarkerPoints()
        if _type == Dimension.NO_ENDPOINT or _type == Dimension.CIRCLE:
            return
        _size = self.getEndpointSize()
        _p1, _p2 = _crossbar.getCrossbarPoints()
        _x1, _y1 = _p1
        _x2, _y2 = _p2
        # print "x1: %g" % _x1
        # print "y1: %g" % _y1
        # print "x2: %g" % _x2
        # print "y2: %g" % _y2
        _sine, _cosine = _crossbar.getSinCosValues()
        if _type == Dimension.ARROW or _type == Dimension.FILLED_ARROW:
            _height = _size/5.0
            # p1 -> (x,y) = (size, _height)
            _mx = (_cosine * _size - _sine * _height) + _x1
            _my = (_sine * _size + _cosine * _height) + _y1
            _crossbar.storeMarkerPoint(_mx, _my)
            # p2 -> (x,y) = (size, -_height)
            _mx = (_cosine * _size - _sine *(-_height)) + _x1
            _my = (_sine * _size + _cosine *(-_height)) + _y1
            _crossbar.storeMarkerPoint(_mx, _my)
            # p3 -> (x,y) = (-size, _height)
            _mx = (_cosine * (-_size) - _sine * _height) + _x2
            _my = (_sine * (-_size) + _cosine * _height) + _y2
            _crossbar.storeMarkerPoint(_mx, _my)
            # p4 -> (x,y) = (-size, -_height)
            _mx = (_cosine * (-_size) - _sine *(-_height)) + _x2
            _my = (_sine * (-_size) + _cosine *(-_height)) + _y2
            _crossbar.storeMarkerPoint(_mx, _my)
        elif _type == Dimension.SLASH:
            _angle = 30.0 * (pi/180.0) # slope of slash
            _height = 0.5 * _size * sin(_angle)
            _length = 0.5 * _size * cos(_angle)
            # p1 -> (x,y) = (-_length, -_height)
            _sx1 = (_cosine * (-_length) - _sine * (-_height))
            _sy1 = (_sine * (-_length) + _cosine * (-_height))
            # p2 -> (x,y) = (_length, _height)
            _sx2 = (_cosine * _length - _sine * _height)
            _sy2 = (_sine * _length + _cosine * _height)
            #
            # shift the calculate based on the location of the
            # marker point
            #
            _mx = _sx1 + _x2
            _my = _sy1 + _y2
            _crossbar.storeMarkerPoint(_mx, _my)
            _mx = _sx2 + _x2
            _my = _sy2 + _y2
            _crossbar.storeMarkerPoint(_mx, _my)
            _mx = _sx1 + _x1
            _my = _sy1 + _y1
            _crossbar.storeMarkerPoint(_mx, _my)
            _mx = _sx2 + _x1
            _my = _sy2 + _y1
            _crossbar.storeMarkerPoint(_mx, _my)
        else:
            raise ValueError, "Unexpected endpoint type: '%s'" % str(_type)

    def onDimension(self, x, y, tol=tolerance.TOL):
        """Test an x/y coordinate pair if it could lay on the dimension.

onDimension(x, y[, tol])

This method has two required parameters:

x: The x-coordinate
y: The y-coordinate

These should both be float values.

There is an optional third parameter, "tol", giving
the maximum distance from the dimension bars that the
x/y coordinates may sit.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        if self.modified():
            self.calcDimValues(False)
            self.reset()
        _ep1, _ep2 = self.__bar1.getEndpoints()
        #
        # test p1 bar
        #
        _val = _test_on_dimension(_ep1[0], _ep1[1],
                                  _ep2[0], _ep2[1],
                                  _x, _y, _t)
        if _val:
            return _val
        #
        # test p2 bar
        #
        _ep1, _ep2 = self.__bar2.getEndpoints()
        _val = _test_on_dimension(_ep1[0], _ep1[1],
                                  _ep2[0], _ep2[1],
                                  _x, _y, _t)
        if _val:
            return _val
        #
        # test crossbar
        #
        _ep1, _ep2 = self.__crossbar.getEndpoints()
        return  _test_on_dimension(_ep1[0], _ep1[1],
                                   _ep2[0], _ep2[1],
                                   _x, _y, _t)

    def getBounds(self):
        """Return the minimal and maximal locations of the dimension

getBounds()

This method overrides the Dimension::getBounds() method
        """
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _dx, _dy = self.getLocation()
        _dxpts = []
        _dypts = []
        _ep1, _ep2 = self.__bar1.getEndpoints()
        _dxpts.append(_ep1[0])
        _dypts.append(_ep1[1])
        _dxpts.append(_ep2[0])
        _dypts.append(_ep2[1])
        _ep1, _ep2 = self.__bar2.getEndpoints()
        _dxpts.append(_ep1[0])
        _dypts.append(_ep1[1])
        _dxpts.append(_ep2[0])
        _dypts.append(_ep2[1])
        _ep1, _ep2 = self.__crossbar.getEndpoints()
        _dxpts.append(_ep1[0])
        _dypts.append(_ep1[1])
        _dxpts.append(_ep2[0])
        _dypts.append(_ep2[1])
        _xmin = min(_x1, _x2, _dx, min(_dxpts))
        _ymin = min(_y1, _y2, _dy, min(_dypts))
        _xmax = max(_x1, _x2, _dx, max(_dxpts))
        _ymax = max(_y1, _y2, _dy, max(_dypts))
        return _xmin, _ymin, _xmax, _ymax

class HorizontalDimension(LinearDimension):
    """A class representing Horizontal dimensions.

This class is derived from the LinearDimension class, so
it shares all those attributes and methods of its parent.
    """
    def __init__(self, l1, p1, l2, p2, dx, dy, ds):
        """Initialize a Horizontal Dimension.

hdim = HorizontalDimension(l1, p1, l2, p2, dx, dy, ds)

l1: A Layer holding Point p1
p1: A Point contained in Layer l1
l2: A Layer holding Point p2
p2: A Point contained in Layer l2
dx: The "x" coordinate of the dimensional text
dy: The "y" coordinate of the dimensional text
ds: The DimStyle used for this Dimension.
        """
        LinearDimension.__init__(self, l1, p1, l2, p2, dx, dy, ds)

    def calculate(self):
        """Determine the length of this HorizontalDimension.

calculate()
        """
        _p1, _p2 = self.getDimPoints()
        return abs(_p1.x - _p2.x)

    def calcDimValues(self, allpts=True):
        """Recalculate the values for dimensional display.

calcDimValues([allpts])

This method overrides the LinearDimension::calcDimValues() method.
        """
        _allpts = allpts
        if _allpts is not True and _allpts is not False:
            raise ValueError, "Invalid argument: " + str(allpts)
        _p1, _p2 = self.getDimPoints()
        _bar1, _bar2 = self.getDimBars()
        _crossbar = self.getDimCrossbar()
        _p1x, _p1y = _p1.getCoords()
        _p2x, _p2y = _p2.getCoords()
        _dx, _dy = self.getLocation()
        _offset = self.getOffset()
        _ext = self.getExtension()
        _crossbar.setFirstEndpoint(_p1x, _dy)
        _crossbar.setSecondEndpoint(_p2x, _dy)
        if _dx < min(_p1x, _p2x) or _dx > max(_p1x, _p2x):
            if _p1x < _p2x:
                if _dx < _p1x:
                    _crossbar.setFirstEndpoint(_dx, _dy)
                if _dx > _p2x:
                    _crossbar.setSecondEndpoint(_dx, _dy)
            else:
                if _dx < _p2x:
                    _crossbar.setSecondEndpoint(_dx, _dy)
                if _dx > _p1x:
                    _crossbar.setFirstEndpoint(_dx, _dy)
        _crossbar.setFirstCrossbarPoint(_p1x, _dy)
        _crossbar.setSecondCrossbarPoint(_p2x, _dy)
        if _dy < min(_p1y, _p2y):
            _bar1.setFirstEndpoint(_p1x, (_p1y - _offset))
            _bar1.setSecondEndpoint(_p1x, (_dy - _ext))
            _bar2.setFirstEndpoint(_p2x, (_p2y - _offset))
            _bar2.setSecondEndpoint(_p2x, (_dy - _ext))
        elif _dy > max(_p1y, _p2y):
            _bar1.setFirstEndpoint(_p1x, (_p1y + _offset))
            _bar1.setSecondEndpoint(_p1x, (_dy + _ext))
            _bar2.setFirstEndpoint(_p2x, (_p2y + _offset))
            _bar2.setSecondEndpoint(_p2x, (_dy + _ext))
        else:
            if _dy > _p1y:
                _bar1.setFirstEndpoint(_p1x, (_p1y + _offset))
                _bar1.setSecondEndpoint(_p1x, (_dy + _ext))
            else:
                _bar1.setFirstEndpoint(_p1x, (_p1y - _offset))
                _bar1.setSecondEndpoint(_p1x, (_dy - _ext))
            if _dy > _p2y:
                _bar2.setFirstEndpoint(_p2x, (_p2y + _offset))
                _bar2.setSecondEndpoint(_p2x, (_dy + _ext))
            else:
                _bar2.setFirstEndpoint(_p2x, (_p2y - _offset))
                _bar2.setSecondEndpoint(_p2x, (_dy - _ext))
        if _allpts:
            self.calcMarkerPoints()

class VerticalDimension(LinearDimension):
    """A class representing Vertical dimensions.

This class is derived from the LinearDimension class, so
it shares all those attributes and methods of its parent.
    """

    def __init__(self, l1, p1, l2, p2, dx, dy, ds):
        """Initialize a Vertical Dimension.

vdim = VerticalDimension(l1, p1, l2, p2, dx, dy, ds)

l1: A Layer holding Point p1
p1: A Point contained in Layer l1
l2: A Layer holding Point p2
p2: A Point contained in Layer l2
dx: The "x" coordinate of the dimensional text
dy: The "y" coordinate of the dimensional text
ds: The DimStyle used for this Dimension.
        """
        LinearDimension.__init__(self, l1, p1, l2, p2, dx, dy, ds)

    def calculate(self):
        """Determine the length of this VerticalDimension.

calculate()
        """
        _p1, _p2 = self.getDimPoints()
        return abs(_p1.y - _p2.y)

    def calcDimValues(self, allpts=True):
        """Recalculate the values for dimensional display.

calcDimValues([allpts])

This method overrides the LinearDimension::calcDimValues() method.
        """
        _allpts = allpts
        if _allpts is not True and _allpts is not False:
            raise ValueError, "Invalid argument: " + str(allpts)
        _p1, _p2 = self.getDimPoints()
        _bar1, _bar2 = self.getDimBars()
        _crossbar = self.getDimCrossbar()
        _p1x, _p1y = _p1.getCoords()
        _p2x, _p2y = _p2.getCoords()
        _dx, _dy = self.getLocation()
        _offset = self.getOffset()
        _ext = self.getExtension()
        _crossbar.setFirstEndpoint(_dx, _p1y)
        _crossbar.setSecondEndpoint(_dx, _p2y)
        if _dy < min(_p1y, _p2y) or _dy > max(_p1y, _p2y):
            if _p1y < _p2y:
                if _dy < _p1y:
                    _crossbar.setFirstEndpoint(_dx, _dy)
                if _dy > _p2y:
                    _crossbar.setSecondEndpoint(_dx, _dy)
            if _p2y < _p1y:
                if _dy < _p2y:
                    _crossbar.setSecondEndpoint(_dx, _dy)
                if _dy > _p1y:
                    _crossbar.setFirstEndpoint(_dx, _dy)
        _crossbar.setFirstCrossbarPoint(_dx, _p1y)
        _crossbar.setSecondCrossbarPoint(_dx, _p2y)
        if _dx < min(_p1x, _p2x):
            _bar1.setFirstEndpoint((_p1x - _offset), _p1y)
            _bar1.setSecondEndpoint((_dx - _ext), _p1y)
            _bar2.setFirstEndpoint((_p2x - _offset), _p2y)
            _bar2.setSecondEndpoint((_dx - _ext), _p2y)
        elif _dx > max(_p1x, _p2x):
            _bar1.setFirstEndpoint((_p1x + _offset), _p1y)
            _bar1.setSecondEndpoint((_dx + _ext), _p1y)
            _bar2.setFirstEndpoint((_p2x + _offset), _p2y)
            _bar2.setSecondEndpoint((_dx + _ext), _p2y)
        else:
            if _dx > _p1x:
                _bar1.setFirstEndpoint((_p1x + _offset), _p1y)
                _bar1.setSecondEndpoint((_dx + _ext), _p1y)
            else:
                _bar1.setFirstEndpoint((_p1x - _offset), _p1y)
                _bar1.setSecondEndpoint((_dx - _ext), _p1y)
            if _dx > _p2x:
                _bar2.setFirstEndpoint((_p2x + _offset), _p2y)
                _bar2.setSecondEndpoint((_dx + _ext), _p2y)
            else:
                _bar2.setFirstEndpoint((_p2x - _offset), _p2y)
                _bar2.setSecondEndpoint((_dx - _ext), _p2y)
        if _allpts:
            self.calcMarkerPoints()

class RadialDimension(Dimension):
    """A class for Radial dimensions.

The RadialDimension class is derived from the Dimension
class, so it shares all of those methods and attributes.
A RadialDimension should be used to display either the
radius or diamter of a Circle object.

A RadialDimension object has the following methods:

{get/set}DimCircle(): Get/Set the measured circle object.
getDimLayer(): Return the layer containing the measured circle.
{get/set}DiaMode(): Get/Set if the RadialDimension should return diameters.
getDimXPoints(): Get the x-coordinates of the dimension bar positions.
getDimYPoints(): Get the y-coordinates of the dimension bar positions.
getDimMarkerPoints(): Get the locaiton of the dimension endpoint markers.
getDimCrossbar(): Get the DimCrossbar object of the RadialDimension.
calcDimValues(): Calculate the endpoint of the dimension line.
onDimension(): Test if an x/y coordinate pair fall on the dimension line.
    """

    def __init__(self, lyr, cir, dx, dy, ds):
        """Initialize a RadialDimension object.

rdim = RadialDimension(cl, cir, dx, dy, ds)

lyr: A Layer object
cir: A Circle or Arc object
dx: The "x" coordinate of the dimensional text
dy: The "y" coordinate of the dimensional text
ds: The DimStyle used for this Dimension.
        """
        Dimension.__init__(self, ds, x=dx, y=dy)
        if not isinstance(lyr, layer.Layer):
            raise TypeError, "Invalid layer: " + `lyr`
        if not isinstance(cir, circle.Circle):
            raise TypeError, "Invalid circle: " + `cir`
        _lc = lyr.findObject(cir)
        if _lc is not cir:
            raise ValueError, "Circle not found in layer!"
        self.__layer = lyr
        self.__circle = cir
        self.__crossbar = DimCrossbar()
        self.__dia_mode = False
        self.applyDimStyle()
        cir.storeUser(self)

    def __eq__(self, rdim):
        """Compare two RadialDimensions for equality.
        """
        if not isinstance(rdim, RadialDimension):
            raise TypeError, "Invalid object for comparison: " + `rdim`
        _val = False
        _rc = rdim.getDimCircle()
        _rl = rdim.getDimLayer()
        if self.__layer is _rl and self.__circle == _rc:
            _val = True
        return _val

    def __ne__(self, rdim):
        """Compare two RadialDimensions for inequality.
        """
        if not isinstance(rdim, RadialDimension):
            raise TypeError, "Invalid object for comparison: " + `rdim`
        _val = True
        _rc = rdim.getDimCircle()
        _rl = rdim.getDimLayer()
        if self.__layer is _rl and self.__circle == _rc:
            _val = False
        return _val

    def applyDimStyle(self):
        """Apply the values in the DimStyle to the RadialDimension.

applyDimStyle()

This method extends the Dimension::applyDimStyle() method.
        """
        Dimension.applyDimStyle(self)
        _ds = self.getDimStyle()
        _pds = self.getPrimaryDimstring()
        _pds.setPrefix(_ds.getValue('RADIAL_DIM_PRIMARY_PREFIX'))
        _pds.setSuffix(_ds.getValue('RADIAL_DIM_PRIMARY_SUFFIX'))
        _sds = self.getSecondaryDimstring()
        _sds.setPrefix(_ds.getValue('RADIAL_DIM_SECONDARY_PREFIX'))
        _sds.setSuffix(_ds.getValue('RADIAL_DIM_SECONDARY_SUFFIX'))
        self.setDiaMode(_ds.getValue('RADIAL_DIM_DIA_MODE'))

    def getDiaMode(self):
        """Return if the RadialDimension will return diametrical values.

getDiaMode()

This method returns True if the diameter value is returned,
and False otherwise.
        """
        return self.__dia_mode

    def setDiaMode(self, mode=False):
        """Set the RadialDimension to return diametrical values.

setDiaMode([mode])

Calling this method without an argument sets the RadialDimension
to return radial measurements. If the argument "mode" is supplied,
it should be either True or False.

If the RadialDimension is measuring an arc, the returned value
will always be set to return a radius.
        """
        if mode is True or mode is False:
            if not isinstance(self.__circle, arc.Arc):
                self.__dia_mode = mode
                self.modified()
        else:
            raise ValueError, "Invalid diameter mode: " + str(mode)

    dia_mode = property(getDiaMode, setDiaMode, None,
                        "Draw the Dimension as a diameter")

    def getDimLayer(self):
        """Return the Layer object holding the Circle for this RadialDimension.

getDimLayer()
        """
        return self.__layer

    def getDimCircle(self):
        """Return the Circle object this RadialDimension is measuring.

getDimCircle()
        """
        return self.__circle

    def setDimCircle(self, l, c):
        """Set the Circle object measured by this RadialDimension.

setDimCircle(l, c)

The arguments for this method are:

l: A Layer containing circle c
c: A circle contained in layer l
        """
        if not isinstance(l, layer.Layer):
            raise TypeError, "Invalid layer: " + `l`
        if not isinstance(c, circle.Circle):
            raise TypeError, "Invalid circle: " + `c`
        _lc = l.findObject(c)
        if _lc is not c:
            raise ValueError, "Circle not found in layer!"
        _circ = self.__circle
        self.__layer = l
        self.__circle = c
        self.modified()
        _circ.freeUser(self)
        c.storeUser(self)

    circle = property(getDimCircle, None, None,
                      "Radial dimension circle object.")

    def getDimCrossbar(self):
        """Get the DimCrossbar object used by the RadialDimension.

getDimCrossbar()
        """
        return self.__crossbar

    def calcDimValues(self, allpts=True):
        """Recalculate the values for dimensional display.

calcDimValues([allpts])

The optional argument "allpts" is by default True. Calling
this method with the argument set to False will skip the
calculation of any dimension endpoint marker points.
        """
        _allpts = allpts
        if _allpts is not True and _allpts is not False:
            raise ValueError, "Invalid argument: " + str(allpts)
        _c = self.__circle
        _dimbar = self.__crossbar
        _cx, _cy = _c.getCenter().getCoords()
        _rad = _c.getRadius()
        _dx, _dy = self.getLocation()
        _dia_mode = self.__dia_mode
        _sep = hypot((_dx - _cx), (_dy - _cy))
        _angle = atan2((_dy - _cy), (_dx - _cx))
        _sx = _rad * cos(_angle)
        _sy = _rad * sin(_angle)
        if isinstance(_c, arc.Arc):
            assert _dia_mode is False, "dia_mode for arc radial dimension"
            _sa = _c.getStartAngle()
            _ea = _c.getEndAngle()
            _angle = (180.0/pi) * _angle
            if _angle < 0.0:
                _angle = _angle + 360.0
            if not _c.throughAngle(_angle):
                _ep1, _ep2 = _c.getEndpoints()
                if _angle < _sa:
                    _sa = (pi/180.0) * _sa
                    _sx = _rad * cos(_sa)
                    _sy = _rad * sin(_sa)
                    if _sep > _rad:
                        _dx = _cx + (_sep * cos(_sa))
                        _dy = _cy + (_sep * sin(_sa))
                if _angle > _ea:
                    _ea = (pi/180.0) * _ea
                    _sx = _rad * cos(_ea)
                    _sy = _rad * sin(_ea)
                    if _sep > _rad:
                        _dx = _cx + (_sep * cos(_ea))
                        _dy = _cy + (_sep * sin(_ea))
        if _dia_mode:
            _dimbar.setFirstEndpoint((_cx - _sx), (_cy - _sy))
            _dimbar.setFirstCrossbarPoint((_cx - _sx), (_cy - _sy))
        else:
            _dimbar.setFirstEndpoint(_cx, _cy)
            _dimbar.setFirstCrossbarPoint(_cx, _cy)
        if _sep > _rad:
            _dimbar.setSecondEndpoint(_dx, _dy)
        else:
            _dimbar.setSecondEndpoint((_cx + _sx), (_cy + _sy))
        _dimbar.setSecondCrossbarPoint((_cx + _sx), (_cy + _sy))
        if not _allpts:
            return
        #
        # calculate dimension endpoint marker coordinates
        #
        _type = self.getEndpointType()
        _dimbar.clearMarkerPoints()
        if _type == Dimension.NO_ENDPOINT or _type == Dimension.CIRCLE:
            return
        _size = self.getEndpointSize()
        _x1, _y1 = _dimbar.getFirstCrossbarPoint()
        _x2, _y2 = _dimbar.getSecondCrossbarPoint()
        _sine, _cosine = _dimbar.getSinCosValues()
        if _type == Dimension.ARROW or _type == Dimension.FILLED_ARROW:
            _height = _size/5.0
            # p1 -> (x,y) = (size, _height)
            _mx = (_cosine * _size - _sine * _height) + _x1
            _my = (_sine * _size + _cosine * _height) + _y1
            _dimbar.storeMarkerPoint(_mx, _my)
            # p2 -> (x,y) = (size, -_height)
            _mx = (_cosine * _size - _sine *(-_height)) + _x1
            _my = (_sine * _size + _cosine *(-_height)) + _y1
            _dimbar.storeMarkerPoint(_mx, _my)
            # p3 -> (x,y) = (-size, _height)
            _mx = (_cosine * (-_size) - _sine * _height) + _x2
            _my = (_sine * (-_size) + _cosine * _height) + _y2
            _dimbar.storeMarkerPoint(_mx, _my)
            # p4 -> (x,y) = (-size, -_height)
            _mx = (_cosine * (-_size) - _sine *(-_height)) + _x2
            _my = (_sine * (-_size) + _cosine *(-_height)) + _y2
            _dimbar.storeMarkerPoint(_mx, _my)
        elif _type == Dimension.SLASH:
            _angle = 30.0 * (pi/180.0) # slope of slash
            _height = 0.5 * _size * sin(_angle)
            _length = 0.5 * _size * cos(_angle)
            # p1 -> (x,y) = (-_length, -_height)
            _sx1 = (_cosine * (-_length) - _sine * (-_height))
            _sy1 = (_sine * (-_length) + _cosine * (-_height))
            # p2 -> (x,y) = (_length, _height)
            _sx2 = (_cosine * _length - _sine * _height)
            _sy2 = (_sine * _length + _cosine * _height)
            #
            # shift the calculate based on the location of the
            # marker point
            #
            _mx = _sx1 + _x1
            _my = _sy1 + _y1
            _dimbar.storeMarkerPoint(_mx, _my)
            _mx = _sx2 + _x1
            _my = _sy2 + _y1
            _dimbar.storeMarkerPoint(_mx, _my)
            _mx = _sx1 + _x2
            _my = _sy1 + _y2
            _dimbar.storeMarkerPoint(_mx, _my)
            _mx = _sx2 + _x2
            _my = _sy2 + _y2
            _dimbar.storeMarkerPoint(_mx, _my)
        else:
            raise ValueError, "Unexpected endpoint type: '%s'" % str(_type)

    def calculate(self):
        """Return the radius or diamter of this RadialDimension.

calculate()

By default, a RadialDimension will return the radius of the
circle. The setDiaMode() method can be called to set the
returned value to corresponed to a diameter.
        """
        _val = self.__circle.getRadius()
        if self.__dia_mode is True:
            _val = _val * 2.0
        return _val

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a RadialDimension exists within a region.

isRegion(xmin, ymin, xmax, ymax[, fully])

The four arguments define the boundary of an area, and the
function returns True if the RadialDimension lies within that
area. If the optional argument "fully" is used and is True,
then the dimensioned circle and the location of the dimension
text must lie within the boundary. Otherwise, the function
returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _circ = self.__circle
        _cx, _cy = _circ.getCenter().getCoords()
        _rad = _circ.getRadius()
        _dx, _dy = self.getLocation()
        _cxmin = min(_cx - _rad, _dx)
        _cymin = min(_cy - _rad, _dy)
        _cxmax = max(_cx + _rad, _dx)
        _cymax = max(_cy + _rad, _dy)
        _val = False
        if (_xmin < _cxmin and _xmax > _cxmax and
            _ymin < _cymin and _ymax > _cymax):
            _val = True
        elif (not fully and
              not (_cxmax < _xmin or
                   _cymax < _ymin or
                   _cxmin > _xmax or
                   _cymin > _ymax)):
            if _xmin < _dx < _xmax and _ymin < _dy < _ymax: # dim text
                _val = True
            elif _xmin < _cx < _xmax and _ymin < _cy < _ymax: # circle center
                _val = True
            else: # dimension line
                _p1, _p2 = self.__crossbar.getEndpoints()
                _val =_test_dim_line(_p1, _p2, _xmin, _ymin, _xmax, _ymax)
        return _val

    def onDimension(self, x, y, tol=tolerance.TOL):
        """Test an x/y coordinate pair if it could lay on the dimension.

onDimension(x, y[, tol])

This method has two required parameters:

x: The x-coordinate
y: The y-coordinate

These should both be float values.

There is an optional third parameter, "tol", giving
the maximum distance from the dimension bars that the
x/y coordinates may sit.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _p1, _p2 = self.__crossbar.getEndpoints()
        return _test_on_dimension(_p1[0], _p1[0], _p2[0], _p2[1],
                                  _x, _y, _t)

    def getBounds(self):
        """Return the minimal and maximal locations of the dimension

getBounds()

This method overrides the Dimension::getBounds() method
        """
        _ep1, _ep2 = self.__crossbar.getEndpoints()
        _ex1, _ey1 = _ep1
        _ex2, _ey2 = _ep2
        _xmin = min(_ex1, _ex2)
        _ymin = min(_ey1, _ey2)
        _xmax = max(_ex1, _ex2)
        _ymax = max(_ey1, _ey2)
        return _xmin, _ymin, _xmax, _ymax

class AngularDimension(Dimension):
    """A class for Angular dimensions.

The AngularDimension class is derived from the Dimension
class, so it shares all of those methods and attributes.

AngularDimension objects have the following methods:

{get/set}VertexPoint(): Get/Set the vertex point for the AngularDimension.
{get/set}P1(): Get/Set the first Point for the AngularDimension.
{get/set}P2(): Get/Set the second Point for the AngularDimension.
getDimPoints(): Return the two Points used in this dimension.
getDimLayers(): Return the two Layers holding the Points.
getDimXPoints(): Get the x-coordinates of the dimension bar positions.
getDimYPoints(): Get the y-coordinates of the dimension bar positions.
getDimAngles(): Get the angles at which the dimension bars should be drawn.
getDimMarkerPoints(): Get the locaiton of the dimension endpoint markers.
calcDimValues(): Calculate the endpoint of the dimension line.
onDimension(): Test if an x/y coordinate pair fall on the dimension line.
invert(): Switch the endpoints used to measure the dimension
    """

    def __init__(self, vl, vp, l1, p1, l2, p2, dx, dy, ds):
        """Initialize an AngularDimension object.

adim = AngularDimension(vl, vp, l1, p1, l2, p2, dx, dy, ds)

vl: A Layer holding Point vp
vp: A Point contained in layer vl
l1: A Layer holding Point p1
p1: A Point contained in Layer l1
l2: A Layer holding Point p2
p2: A Point contained in Layer l2
dx: The "x" coordinate of the dimensional text
dy: The "y" coordinate of the dimensional text
ds: The DimStyle used for this Dimension.
        """
        Dimension.__init__(self, ds, x=dx, y=dy)
        if not isinstance(vl, layer.Layer):
            raise TypeError, "Invalid layer: " + `vl`
        if not isinstance(vp, point.Point):
            raise TypeError, "Invalid point: " + `vp`
        _lp = vl.findObject(vp)
        if _lp is not vp:
            raise ValueError, "Vertex Point not found in layer!"
        if not isinstance(l1, layer.Layer):
            raise TypeError, "Invalid layer: " + `l1`
        if not isinstance(p1, point.Point):
            raise TypeError, "Invalid point: " + `p1`
        _lp = l1.findObject(p1)
        if _lp is not p1:
            raise ValueError, "Point p1 not found in layer!"
        if not isinstance(l2, layer.Layer):
            raise TypeError, "Invalid layer: " + `l2`
        if not isinstance(p2, point.Point):
            raise TypeError, "Invalid point: " + `p2`
        _lp = l2.findObject(p2)
        if _lp is not p2:
            raise ValueError, "Point p2 not found in layer!"
        self.__vl = vl
        self.__vp = vp
        self.__l1 = l1
        self.__p1 = p1
        self.__l2 = l2
        self.__p2 = p2
        self.applyDimStyle()
        self.__bar1 = DimBar()
        self.__bar2 = DimBar()
        self.__crossarc = DimCrossarc()
        vp.storeUser(self)
        p1.storeUser(self)
        p2.storeUser(self)

    def __eq__(self, adim):
        """Compare two AngularDimensions for equality.
        """
        if not isinstance(adim, AngularDimension):
            raise TypeError, "Invalid object for comparison: " + `adim`
        _val = False
        _vl, _l1, _l2 = adim.getDimLayers()
        _vp, _p1, _p2 = adim.getDimPoints()
        if (self.__vl is _vl and
            self.__vp == _vp and
            self.__l1 is _l1 and
            self.__p1 == _p1 and
            self.__l2 is _l2 and
            self.__p2 == _p2):
            _val = True
        return _val

    def __ne__(self, adim):
        """Compare two AngularDimensions for inequality.
        """
        if not isinstance(adim, AngularDimension):
            raise TypeError, "Invalid object for comparison: " + `adim`
        _val = True
        _vl, _l1, _l2 = adim.getDimLayers()
        _vp, _p1, _p2 = adim.getDimPoints()
        if (self.__vl is _vl and
            self.__vp == _vp and
            self.__l1 is _l1 and
            self.__p1 == _p1 and
            self.__l2 is _l2 and
            self.__p2 == _p2):
            _val = False
        return _val

    def applyDimStyle(self):
        """Set the values in the AngularDimension's DimStyle.

applyDimStyle()

This method extends the Dimension::applyDimStyle() method.
        """
        Dimension.applyDimStyle(self)
        _ds = self.getDimStyle()
        _pds = self.getPrimaryDimstring()
        _pds.setPrefix(_ds.getValue('ANGULAR_DIM_PRIMARY_PREFIX'))
        _pds.setSuffix(_ds.getValue('ANGULAR_DIM_PRIMARY_SUFFIX'))
        _sds = self.getSecondaryDimstring()
        _sds.setPrefix(_ds.getValue('ANGULAR_DIM_SECONDARY_PREFIX'))
        _sds.setSuffix(_ds.getValue('ANGULAR_DIM_SECONDARY_SUFFIX'))

    def getDimLayers(self):
        """Return the layers used in an AngularDimension.

getDimLayers()
        """
        return (self.__vl, self.__l1, self.__l2)

    def getDimPoints(self):
        """Return the points used in an AngularDimension.

getDimPoints()
        """
        return self.__vp, self.__p1, self.__p2

    def getVertexPoint(self):
        """Return the vertex point used in an AngularDimension.

getVertexPoint()
        """
        return self.__vp

    def setVertexPoint(self, l, p):
        """Set the vertex point for an AngularDimension.

setVertexPoint(l, p)

There are two required arguments for this method

l: The layer holding the Point p
p: A point in Layer l
        """
        if not isinstance(l, layer.Layer):
            raise TypeError, "Invalid layer: " + `l`
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _lp = l.findObject(p)
        if _lp is not p:
            raise ValueError, "Point not found in layer!"
        _vp = self.__vp
        self.__vl = l
        self.__vp = p
        self.modified()
        _vp.freeUser(self)
        p.storeUser(self)

    vp = property(getVertexPoint, None, None,
                  "Angular Dimension vertex point.")

    def getP1(self):
        """Return the first angle point used in an AngularDimension.

getP1()
        """
        return self.__p1

    def setP1(self, l, p):
        """Set the first Point for an AngularDimension.

setP1(l, p)

There are two required arguments for this method

l: The layer holding the Point p
p: A point in Layer l
        """
        if not isinstance(l, layer.Layer):
            raise TypeError, "Invalid layer: " + `l`
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _lp = l.findObject(p)
        if _lp is not p:
            raise ValueError, "Point not found in layer!"
        _p1 = self.__p1
        self.__l1 = l
        self.__p1 = p
        self.modified()
        _p1.freeUser(self)
        p.storeUser(self)

    p1 = property(getP1, None, None, "Dimension first point.")

    def getP2(self):
        """Return the second angle point used in an AngularDimension.

getP2()
        """
        return self.__p2

    def setP2(self, l, p):
        """Set the second Point for an AngularDimension.

setP2(l, p)

There are two required arguments for this method

l: The layer holding the Point p
p: A point in Layer l
        """
        if not isinstance(l, layer.Layer):
            raise TypeError, "Invalid layer: " + `l`
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _lp = l.findObject(p)
        if _lp is not p:
            raise ValueError, "Point not found in layer!"
        _p2 = self.__p2
        self.__l2 = l
        self.__p2 = p
        self.modified()
        _p2.freeUser(self)
        p.storeUser(self)

    p2 = property(getP2, None, None, "Dimension second point.")

    def getDimAngles(self):
        """Get the array of dimension bar angles.

geDimAngles()
        """
        _angle1 = self.__bar1.getAngle()
        _angle2 = self.__bar2.getAngle()
        return _angle1, _angle2

    def getDimRadius(self):
        """Get the radius of the dimension crossarc.

getDimRadius()
        """
        return self.__crossarc.getRadius()

    def getDimBars(self):
        """Return the dimension boundary bars.

getDimBars()
        """
        return self.__bar1, self.__bar2

    def getDimCrossarc(self):
        """Get the DimCrossarc object used by the AngularDimension.

getDimCrossarc()
        """
        return self.__crossarc

    def invert(self):
        """Switch the endpoints used in this object.

invert()

Invoking this method on an AngularDimension will result in
it measuring the opposite angle than what it currently measures.
        """
        _ltemp = self.__l1
        _ptemp = self.__p1
        self.__l1 = self.__l2
        self.__p1 = self.__p2
        self.__l2 = _ltemp
        self.__p2 = _ptemp
        self.modified()

    def calculate(self):
        """Find the value of the angle measured by this AngularDimension.

calculate()
        """
        _vx, _vy = self.__vp.getCoords()
        _p1x, _p1y = self.__p1.getCoords()
        _p2x, _p2y = self.__p2.getCoords()
        _a1 = (180.0/pi) * atan2((_p1y - _vy), (_p1x - _vx))
        if _a1 < 0.0:
            _a1 = _a1 + 360.0
        _a2 = (180.0/pi) * atan2((_p2y - _vy), (_p2x - _vx))
        if _a2 < 0.0:
            _a2 = _a2 + 360.0
        _val = _a2 - _a1
        if _a1 > _a2:
            _val = _val + 360.0
        return _val

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not an AngularDimension exists within a region.

isRegion(xmin, ymin, xmax, ymax[, fully])

The four arguments define the boundary of an area, and the
function returns True if the RadialDimension lies within that
area. If the optional argument "fully" is used and is True,
then the dimensioned circle and the location of the dimension
text must lie within the boundary. Otherwise, the function
returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _vx, _vy = self.__vp.getCoords()
        _p1x, _p1y = self.__p1.getCoords()
        _p2x, _p2y = self.__p2.getCoords()
        _dx, _dy = self.getLocation()
        _pxmin, _pymin, _pxmax, _pymax = self.getBounds()
        _val = False
        if (_xmin < _pxmin and _xmax > _pxmax and
            _ymin < _pymin and _ymax > _pymax and
            _xmin < _dx < _xmax and _ymin < _dy < _ymax):
            _val = True
        elif not fully:
            if _xmin < _dx < _xmax and _ymin < _dy < _ymax:
                _val = True
            elif _xmin < _vx < _xmax and _ymin < _vy < _ymax:
                _val = True
            elif _xmin < _p1x < _xmax and _ymin < _p1y < _ymax:
                _val = True
            elif _xmin < _p2x < _xmax and _ymin < _p2y < _ymax:
                _val = True
            else:
                #
                # bar on vp-p1 line
                #
                _ep1, _ep2 = self.__bar1.getEndpoints()
                _val =_test_dim_line(_ep1, _ep2,  _xmin, _ymin, _xmax, _ymax)
                if _val:
                    return _val
                #
                # bar at vp-p2 line
                #
                _ep1, _ep2 = self.__bar2.getEndpoints()
                _val =_test_dim_line(_ep1, _ep2, _xmin, _ymin, _xmax, _ymax)
                if _val:
                    return _val
                #
                # dimension crossarc
                #
                _r = self.__crossarc.getRadius()
                _d1 = hypot((_xmin - _vx), (_ymin - _vy))
                _d2 = hypot((_xmin - _vx), (_ymax - _vy))
                _d3 = hypot((_xmax - _vx), (_ymax - _vy))
                _d4 = hypot((_xmax - _vx), (_ymin - _vy))
                _dmin = min(_d1, _d2, _d3, _d4)
                _dmax = max(_d1, _d2, _d3, _d4)
                if _xmin < _vx < _xmax and _ymin < _vy < _ymax:
                    _dmin = -1e-10
                else:
                    if _vx > _xmax and _ymin < _vy < _ymax:
                        _dmin = _vx - _xmax
                    elif _vx < _xmin and _ymin < _vy < _ymax:
                        _dmin = _xmin - _vx
                    elif _vy > _ymax and _xmin < _vx < _xmax:
                        _dmin = _vy - _ymax
                    elif _vy < _ymin and _xmin < _vx < _xmax:
                        _dmin = _ymin - _vy
                if _dmin < _r < _dmax:
                    _rtd = 180.0/pi
                    _da = _rtd * atan2((_ymin - _vy), (_xmin - _vx))
                    if _da < 0.0:
                        _da = _da + 360.0
                    _val = self._throughAngle(_da)
                    if _val:
                        return _val
                    _da = _rtd * atan2((_ymin - _vy), (_xmax - _vx))
                    if _da < 0.0:
                        _da = _da + 360.0
                    _val = self._throughAngle(_da)
                    if _val:
                        return _val
                    _da = _rtd * atan2((_ymax - _vy), (_xmax - _vx))
                    if _da < 0.0:
                        _da = _da + 360.0
                    _val = self._throughAngle(_da)
                    if _val:
                        return _val
                    _da = _rtd * atan2((_ymax - _vy), (_xmin - _vx))
                    if _da < 0.0:
                        _da = _da + 360.0
                    _val = self._throughAngle(_da)
        return _val

    def _throughAngle(self, angle):
        """Test if the angular crossarc exists at a certain angle.

_throughAngle()

This method is private to the AngularDimension class.
        """
        _crossarc = self.__crossarc
        _angle1 = _crossarc.getStartAngle()
        _angle2 = _crossarc.getEndAngle()
        _val = False
        if _angle2 > _angle1:
            if _angle2 > angle > _angle1:
                _val = True
        else:
            if ((_angle1 < angle < (360.0 + 1e-10)) or
                (-1e-10 < angle < _angle2)):
                _val = True
        return _val

    def calcDimValues(self, allpts=True):
        """Recalculate the values for dimensional display.

calcDimValues([allpts])

The optional argument "allpts" is by default True. Calling
this method with the argument set to False will skip the
calculation of any dimension endpoint marker points.
        """
        _allpts = allpts
        if _allpts is not True and _allpts is not False:
            raise ValueError, "Invalid argument: " + str(allpts)
        _vx, _vy = self.__vp.getCoords()
        _p1x, _p1y = self.__p1.getCoords()
        _p2x, _p2y = self.__p2.getCoords()
        _dx, _dy = self.getLocation()
        _offset = self.getOffset()
        _ext = self.getExtension()
        _bar1 = self.__bar1
        _bar2 = self.__bar2
        _crossarc = self.__crossarc
        _dv1 = hypot((_p1x - _vx), (_p1y - _vy))
        _dv2 = hypot((_p2x - _vx), (_p2y - _vy))
        _ddp = hypot((_dx - _vx), (_dy - _vy))
        _crossarc.setRadius(_ddp)
        _rtd = 180.0/pi
        #
        # first dimension bar
        #
        _angle = atan2((_p1y - _vy), (_p1x - _vx))
        _sine = sin(_angle)
        _cosine = cos(_angle)
        _deg = _angle * _rtd
        if _deg < 0.0:
            _deg = _deg + 360.0
        _crossarc.setStartAngle(_deg)
        _ex = _vx + (_ddp * _cosine)
        _ey = _vy + (_ddp * _sine)
        _crossarc.setFirstCrossbarPoint(_ex, _ey)
        _crossarc.setFirstEndpoint(_ex, _ey)
        if _ddp > _dv1: # dim point is radially further to vp than p1
            _x1 = _p1x + (_offset * _cosine)
            _y1 = _p1y + (_offset * _sine)
            _x2 = _vx + ((_ddp + _ext) * _cosine)
            _y2 = _vy + ((_ddp + _ext) * _sine)
        else: # dim point is radially closer to vp than p1
            _x1 = _p1x - (_offset * _cosine)
            _y1 = _p1y - (_offset * _sine)
            _x2 = _vx + ((_ddp - _ext) * _cosine)
            _y2 = _vy + ((_ddp - _ext) * _sine)
        _bar1.setFirstEndpoint(_x1, _y1)
        _bar1.setSecondEndpoint(_x2, _y2)
        #
        # second dimension bar
        #
        _angle = atan2((_p2y - _vy), (_p2x - _vx))
        _sine = sin(_angle)
        _cosine = cos(_angle)
        _deg = _angle * _rtd
        if _deg < 0.0:
            _deg = _deg + 360.0
        _crossarc.setEndAngle(_deg)
        _ex = _vx + (_ddp * _cosine)
        _ey = _vy + (_ddp * _sine)
        _crossarc.setSecondCrossbarPoint(_ex, _ey)
        _crossarc.setSecondEndpoint(_ex, _ey)
        if _ddp > _dv2: # dim point is radially further to vp than p2
            _x1 = _p2x + (_offset * _cosine)
            _y1 = _p2y + (_offset * _sine)
            _x2 = _vx + ((_ddp + _ext) * _cosine)
            _y2 = _vy + ((_ddp + _ext) * _sine)
        else: # dim point is radially closers to vp than p2
            _x1 = _p2x - (_offset * _cosine)
            _y1 = _p2y - (_offset * _sine)
            _x2 = _vx + ((_ddp - _ext) * _cosine)
            _y2 = _vy + ((_ddp - _ext) * _sine)
        _bar2.setFirstEndpoint(_x1, _y1)
        _bar2.setSecondEndpoint(_x2, _y2)
        if not _allpts:
            return
        #
        # calculate dimension endpoint marker coordinates
        #
        _type = self.getEndpointType()
        _crossarc.clearMarkerPoints()
        if _type == Dimension.NO_ENDPOINT or _type == Dimension.CIRCLE:
            return
        _size = self.getEndpointSize()
        _a1 = _bar1.getAngle() - 90.0
        _a2 = _bar2.getAngle() - 90.0
        # print "a1: %g" % _a1
        # print "a2: %g" % _a2
        _mp1, _mp2 = _crossarc.getCrossbarPoints()
        _x1, _y1 = _mp1
        _x2, _y2 = _mp2
        # print "x1: %g" % _x1
        # print "y1: %g" % _y1
        # print "x2: %g" % _x2
        # print "y2: %g" % _y2
        _dtr = pi/180.0
        _sin1 = sin(_dtr * _a1)
        _cos1 = cos(_dtr * _a1)
        _sin2 = sin(_dtr * _a2)
        _cos2 = cos(_dtr * _a2)
        if _type == Dimension.ARROW or _type == Dimension.FILLED_ARROW:
            _height = _size/5.0
            # p1 -> (x,y) = (size, _height)
            _mx = (_cos1 * (-_size) - _sin1 * _height) + _x1
            _my = (_sin1 * (-_size) + _cos1 * _height) + _y1
            _crossarc.storeMarkerPoint(_mx, _my)
            # p2 -> (x,y) = (size, -_height)
            _mx = (_cos1 * (-_size) - _sin1 *(-_height)) + _x1
            _my = (_sin1 * (-_size) + _cos1 *(-_height)) + _y1
            _crossarc.storeMarkerPoint(_mx, _my)
            # p3 -> (x,y) = (size, _height)
            _mx = (_cos2 * _size - _sin2 * _height) + _x2
            _my = (_sin2 * _size + _cos2 * _height) + _y2
            _crossarc.storeMarkerPoint(_mx, _my)
            # p4 -> (x,y) = (size, -_height)
            _mx = (_cos2 * _size - _sin2 *(-_height)) + _x2
            _my = (_sin2 * _size + _cos2 *(-_height)) + _y2
            _crossarc.storeMarkerPoint(_mx, _my)
        elif _type == Dimension.SLASH:
            _angle = 30.0 * (pi/180.0) # slope of slash
            _height = 0.5 * _size * sin(_angle)
            _length = 0.5 * _size * cos(_angle)
            # p1 -> (x,y) = (-_length, -_height)
            _mx = (_cos1 * (-_length) - _sin1 * (-_height)) + _x1
            _my = (_sin1 * (-_length) + _cos1 * (-_height)) + _y1
            _crossarc.storeMarkerPoint(_mx, _my)
            # p2 -> (x,y) = (_length, _height)
            _mx = (_cos1 * _length - _sin1 * _height) + _x1
            _my = (_sin1 * _length + _cos1 * _height) + _y1
            _crossarc.storeMarkerPoint(_mx, _my)
            # p3 -> (x,y) = (-_length, -_height)
            _mx = (_cos2 * (-_length) - _sin2 * (-_height)) + _x2
            _my = (_sin2 * (-_length) + _cos2 * (-_height)) + _y2
            _crossarc.storeMarkerPoint(_mx, _my)
            # p4 -> (x,y) = (_length, _height)
            _mx = (_cos2 * _length - _sin2 * _height) + _x2
            _my = (_sin2 * _length + _cos2 * _height) + _y2
            _crossarc.storeMarkerPoint(_mx, _my)
        else:
            raise ValueError, "Unexpected endpoint type: '%s'" % str(_type)

    def onDimension(self, x, y, tol=tolerance.TOL):
        """Test an x/y coordinate pair hit the dimension lines and arc.

onDimension(x, y[, tol])

This method has two required parameters:

x: The x-coordinate
y: The y-coordinate

These should both be float values.

There is an optional third parameter, "tol", giving
the maximum distance from the dimension bars that the
x/y coordinates may sit.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        #
        # test vp-p1 bar
        #
        _ep1, _ep2 = self.__bar1.getEndpoints()
        _ex1, _ey1 = _ep1
        _ex2, _ey2 = _ep2
        _val = _test_on_dimension(_ex1, _ey1, _ex2, _ey2, _x, _y, _t)
        if _val:
            return _val
        #
        # test vp-p2 bar
        #
        _ep1, _ep2 = self.__bar2.getEndpoints()
        _val = _test_on_dimension(_ex1, _ey1, _ex2, _ey2, _x, _y, _t)
        if _val:
            return _val
        #
        # test the arc
        #
        _vx, _vy = self.__vp.getCoords()
        _psep = hypot((_vx - _x), (_vy - y))
        _dx, _dy = self.getLocation()
        _dsep = hypot((_vx - _dx), (_vy - _dy))
        _val = False
        if abs(_psep - _dsep) < _t:
            _crossarc = self.__crossarc
            _angle1 = _crossarc.getStartAngle()
            _angle2 = _crossarc.getEndAngle()
            _angle = (180.0/pi) * atan2((_vy - _y), (_vx - _x))
            if _angle1 < _angle2:
                if _angle1 < _angle < _angle2:
                    _val = True
            else:
                if ((_angle1 < _angle < (360.0 + 1e-10)) or
                    (-1e-10 < _angle < _angle2)):
                    _val = True
        return _val

    def getBounds(self):
        """Return the minimal and maximal locations of the dimension

getBounds()

This method overrides the Dimension::getBounds() method
        """
        _vx, _vy = self.__vp.getCoords()
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _dx, _dy = self.getLocation()
        _dxpts = []
        _dypts = []
        _ep1, _ep2 = self.__bar1.getEndpoints()
        _dxpts.append(_ep1[0])
        _dypts.append(_ep1[1])
        _dxpts.append(_ep2[0])
        _dypts.append(_ep2[1])
        _ep1, _ep2 = self.__bar2.getEndpoints()
        _dxpts.append(_ep1[0])
        _dypts.append(_ep1[1])
        _dxpts.append(_ep2[0])
        _dypts.append(_ep2[1])
        _rad = self.__crossarc.getRadius()
        _dxpts.append((_vx + _rad))
        _dypts.append((_vy + _rad))
        _dxpts.append((_vx - _rad))
        _dypts.append((_vy - _rad))
        _xmin = min(_x1, _x2, _dx, min(_dxpts))
        _ymin = min(_y1, _y2, _dy, min(_dypts))
        _xmax = max(_x1, _x2, _dx, max(_dxpts))
        _ymax = max(_y1, _y2, _dy, max(_dypts))
        return _xmin, _ymin, _xmax, _ymax

#
# class stuff for dimension styles
#

class DimStyle(object):
    """A class storing preferences for Dimensions

The DimStyle class stores a set of dimension parameters
that will be used when creating dimensions when the
particular style is active.

A DimStyle object has the following methods:

getName(): Return the name of the DimStyle.
getOption(): Return a single value in the DimStyle.
getOptions(): Return all the options in the DimStyle.
getValue(): Return the value of one of the DimStyle options.

getOption() and getValue() are synonymous.
    """

    #
    # the default values for the DimStyle class
    #

    __defaults = {
        'DIM_PRIMARY_FONT_FAMILY' : 'Sans',
        'DIM_PRIMARY_FONT_SIZE' : 12,
        'DIM_PRIMARY_FONT_WEIGHT' : text.NORMAL,
        'DIM_PRIMARY_FONT_STYLE' : text.NORMAL,
        'DIM_PRIMARY_FONT_COLOR' : color.Color(255,255,255),
        'DIM_PRIMARY_PREFIX' : u'',
        'DIM_PRIMARY_SUFFIX' : u'',
        'DIM_PRIMARY_PRECISION' : 3,
        'DIM_PRIMARY_UNITS' : units.MILLIMETERS,
        'DIM_PRIMARY_LEADING_ZERO' : True,
        'DIM_PRIMARY_TRAILING_DECIMAL' : True,
        'DIM_SECONDARY_FONT_FAMILY' : 'Sans',
        'DIM_SECONDARY_FONT_SIZE' : 12,
        'DIM_SECONDARY_FONT_WEIGHT' : text.NORMAL,
        'DIM_SECONDARY_FONT_STYLE' : text.NORMAL,
        'DIM_SECONDARY_FONT_COLOR' : color.Color(255,255,255),
        'DIM_SECONDARY_PREFIX' : u'',
        'DIM_SECONDARY_SUFFIX' : u'',
        'DIM_SECONDARY_PRECISION' : 3,
        'DIM_SECONDARY_UNITS' : units.MILLIMETERS,
        'DIM_SECONDARY_LEADING_ZERO' : True,
        'DIM_SECONDARY_TRAILING_DECIMAL' : True,
        'DIM_OFFSET' : 1.0,
        'DIM_EXTENSION' : 1.0,
        'DIM_COLOR' : color.Color(255,165,0),
        'DIM_POSITION' : Dimension.SPLIT,
        'DIM_ENDPOINT' : Dimension.NO_ENDPOINT,
        'DIM_ENDPOINT_SIZE' : 1.0,
        'DIM_DUAL_MODE' : False,
        'RADIAL_DIM_PRIMARY_PREFIX' : u'',
        'RADIAL_DIM_PRIMARY_SUFFIX' : u'',
        'RADIAL_DIM_SECONDARY_PREFIX' : u'',
        'RADIAL_DIM_SECONDARY_SUFFIX' : u'',
        'RADIAL_DIM_DIA_MODE' : False,
        'ANGULAR_DIM_PRIMARY_PREFIX' : u'',
        'ANGULAR_DIM_PRIMARY_SUFFIX' : u'',
        'ANGULAR_DIM_SECONDARY_PREFIX' : u'',
        'ANGULAR_DIM_SECONDARY_SUFFIX' : u'',
        }

    def __init__(self, name, keywords={}):
        """Instantiate a DimStyle object.

ds = DimStyle(name, keywords)

The argument "name" should be a unicode name, and the
"keyword" argument should be a dict. The keys should
be the same keywords used to set option values, such
as DIM_OFFSET, DIM_EXTENSION, etc, and the value corresponding
to each key should be set appropriately.
        """
        _n = name
        if not isinstance(_n, (str, unicode)):
            raise TypeError, "Invalid DimStyle name: "+ `_n`
        if isinstance(_n, str):
            _n = unicode(_n)
        if not isinstance(keywords, dict):
            raise TypeError, "Invalid keywords argument: " + `keywords`
        self.__opts = baseobject.ConstDict(str)
        self.__name = _n
        for _kw in keywords:
            _val = keywords[_kw]
            if _kw not in DimStyle.__defaults:
                raise KeyError, "Unknown DimStyle keyword: " + _kw
            _valid = options.test_option(_kw, _val)
            self.__opts[_kw] = _val

    def __eq__(self, obj):
        """Test a DimStyle object for equality with another DimStyle.
        """
        if not isinstance(obj, DimStyle):
            raise TypeError, "Invalid object for DimStyle comparison: " + `obj`
        return self.__opts == obj.__opts

    def __ne__(self, obj):
        """Test a DimStyle object for inequality with another DimStyle.
        """
        if not isinstance(obj, DimStyle):
            raise TypeError, "Invalid object for DimStyle comparison: " + `obj`
        return self.__opts != obj.__opts

    def getName(self):
        """Return the name of the DimStyle.

getName()
        """
        return self.__name

    name = property(getName, None, None, "DimStyle name.")

    def getOptions(self):
        """Return all the options stored within the DimStyle.

getOptions()
        """
        _keys = self.__opts.keys()
        for _key in DimStyle.__defaults:
            if _key not in self.__opts:
                _keys.append(_key)
        return _keys

    def getOption(self, key):
        """Return the value of a particular option in the DimStyle.

getOption(key)

The key should be one of the strings returned from getOptions. If
there is no value found in the DimStyle for the key, the value None
is returned.
        """
        if key in self.__opts:
            _val = self.__opts[key]
        elif key in DimStyle.__defaults:
            _val = DimStyle.__defaults[key]
        else:
            raise KeyError, "Unexpected DimStyle keyword: " + key
        return _val

    def getValue(self, key):
        """Return the value of a particular option in the DimStyle.

getValue(key)

The key should be one of the strings returned from getOptions. This
method raises a KeyError exception if the key is not found.

        """
        if key in self.__opts:
            _val = self.__opts[key]
        elif key in DimStyle.__defaults:
            _val = DimStyle.__defaults[key]
        else:
            raise KeyError, "Unexpected DimStyle keyword: " + key
        return _val

dimstyle_defaults = {
    'DIM_PRIMARY_FONT_FAMILY' : 'Sans',
    'DIM_PRIMARY_FONT_SIZE' : 12,
    'DIM_PRIMARY_FONT_WEIGHT' : text.NORMAL,
    'DIM_PRIMARY_FONT_STYLE' : text.NORMAL,
    'DIM_PRIMARY_FONT_COLOR' : color.Color(255,255,255),
    'DIM_PRIMARY_PREFIX' : u'',
    'DIM_PRIMARY_SUFFIX' : u'',
    'DIM_PRIMARY_PRECISION' : 3,
    'DIM_PRIMARY_UNITS' : units.MILLIMETERS,
    'DIM_PRIMARY_LEADING_ZERO' : True,
    'DIM_PRIMARY_TRAILING_DECIMAL': True,
    'DIM_SECONDARY_FONT_FAMILY' : 'Sans',
    'DIM_SECONDARY_FONT_SIZE' : 12,
    'DIM_SECONDARY_FONT_WEIGHT' : text.NORMAL,
    'DIM_SECONDARY_FONT_STYLE' : text.NORMAL,
    'DIM_SECONDARY_FONT_COLOR' : color.Color(255,255,255),
    'DIM_SECONDARY_PREFIX' : u'',
    'DIM_SECONDARY_SUFFIX' : u'',
    'DIM_SECONDARY_PRECISION' : 3,
    'DIM_SECONDARY_UNITS' : units.MILLIMETERS,
    'DIM_SECONDARY_LEADING_ZERO' : True,
    'DIM_SECONDARY_TRAILING_DECIMAL': True,
    'DIM_OFFSET' : 1.0,
    'DIM_EXTENSION': 1.0,
    'DIM_COLOR' : color.Color(255,165,0),
    'DIM_POSITION': Dimension.SPLIT,
    'DIM_ENDPOINT': Dimension.NO_ENDPOINT,
    'DIM_ENDPOINT_SIZE' : 1.0,
    'DIM_DUAL_MODE' : False,
    'RADIAL_DIM_PRIMARY_PREFIX' : u'',
    'RADIAL_DIM_PRIMARY_SUFFIX' : u'',
    'RADIAL_DIM_SECONDARY_PREFIX' : u'',
    'RADIAL_DIM_SECONDARY_SUFFIX' : u'',
    'RADIAL_DIM_DIA_MODE' : False,
    'ANGULAR_DIM_PRIMARY_PREFIX' : u'',
    'ANGULAR_DIM_PRIMARY_SUFFIX' : u'',
    'ANGULAR_DIM_SECONDARY_PREFIX' : u'',
    'ANGULAR_DIM_SECONDARY_SUFFIX' : u'',
    }

#
# dimension line testing
#
# test if a dimension line from (x1, y1), (x2, y2) is
# visiible in a region from (xmin, ymin, xmax, ymax)
#

def _test_dim_line(t1, t2, xmin, ymin, xmax, ymax):
    _x1, _y1 = t1
    _x2, _y2 = t2
    _val = False
    #
    # see if dim line intersects an imaginary segment
    # from (xmin, ymax) to (xmax, ymin)
    #
    _d = ((xmax - xmin)*(_y2 - _y1)) - ((ymin - ymax)*(_x2 - _x1))
    if abs(_d) > 1e-10: # not parallel
        _rn = ((ymax - _y1)*(_x2 - _x1)) - ((xmin - _x1)*(_y2 - _y1))
        _r = _rn/_d
        if 0.0 < _r < 1.0:
            _sn = ((ymax - _y1)*(xmax - xmin)) - ((xmin - _x1)*(ymin - ymax))
            _s = _sn/_d
            if 0.0 < _s < 1.0:
                _val = True
    if _val:
        return _val
    #
    # see if dim line intersects an imaginary segment
    # from (xmin, ymin) to (xmax, ymax)
    #
    _d = ((xmax - xmin)*(_y2 - _y1)) - ((ymax - ymin)*(_x2 - _x1))
    if abs(_d) > 1e-10: # not parallel
        _rn = ((ymin - _y1)*(_x2 - _x1)) - ((xmin - _x1)*(_y2 - _y1))
        _r = _rn/_d
        if 0.0 < _r < 1.0:
            _sn = ((ymax - _y1)*(xmax - xmin)) - ((xmin - _x1)*(ymax - ymin))
            _s = _sn/_d
            if 0.0 < _s < 1.0:
                _val = True
    return _val

#
# test if a projected point from an x/y coordinate pair
# is within a tolerance limited distance from the coordinate
#

def _test_on_dimension(dx1, dy1, dx2, dy2, x, y, t):
        _dx = dx2 - dx1
        _dy = dy2 - dy1
        _d = pow(_dx, 2) + pow(_dy, 2)
        _val = False
        if abs(_d) > 1e-10: # should always be true
            _r = ((x - dx1)*(_dx) + (y - dy1)*(_dy))/_d
            if _r < 0.0 : _r = 0.0
            if _r > 1.0 : _r = 1.0
            _px = dx1 + (_r * _dx)
            _py = dy1 + (_r * _dy)
            if ((abs(_px - x) < t) and (abs(_py - y) < t)):
                _val = True
        return _val
