#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# The Layer class
#

import warnings
import sys

import point
import segment
import circle
import arc
import hcline
import vcline
import acline
import cline
import ccircle
import segjoint
import leader
import polyline
import text
import dimension
import tolerance
import baseobject

class Layer(object):
    """The Layer class.

A Layer object holds all the various entities that can be
in a drawing. Each layer can have sublayers, and there is
no limit to the depth of the sublayering.

A Layer object has several attributes:

name: The Layer's name
parent: The parent Layer of the Layer
scale: The scale factor for object contained in the Layer
visibility: A flag to show the Layer's contents or not

A Layer object has the following methods:

{get/set}Name(): Get/Set the Layer's name.
{get/set}Parent(): Get/Set the Layer's parent.
hasChildren(): Test if the Layer has children.
getChildren(): Return the children of the Layer.
getEntities(): Return all of a type of entity in the Layer.
{add/del}Child(): Add/Remove a new child Layer to the Layer.
{add/del}Object(): Store/Remove a Point, Segment, etc. in the Layer.
bindObject(): Set the references of an object to the objects it uses.
usesObject(): List objects composed from other objects.
countUsers(): Return how many objects are using another object.
findObject(): Return an object in the layer equivalent to a test object.
hasPoint(): See if the Layer has a Point object at a location.
hasPointAt(): Find a Point in the Layer.
hasCircleAt(): Find a Circle in the Layer.
getExtRefs(): Return a list of objects refering to an object in this Layer.
mapPoint(): See if a non-Point object in the layer crosses some location.
addLayer(): Add the contents of one layer to another.
{get/set}Visibility(): Get/Set the visibility of the Layer.
getBoundary(): Find the maximum and minimum coordinates of the Layer.
objsInRegion(): Return all the objects in the Layer that can be seen
                within some view.
update(): Check that all objects in the layer are stored properly.
    """

    def __init__(self, name=None):
        """Initializee a Layer.

Layer(name)

The 'name' argument is optional. It should be a string if used,
otherwise a default name "Layer" is assigned to the layer.
        """
        _n = name
        if _n is None:
            _n = u'Layer'
        if not isinstance(_n, unicode):
            if isinstance(_n, str):
                _n = unicode(_n)
            else:
                raise TypeError, "Invalid layer name: "+ `_n`
        self.__name = _n
        self.__parent = None
        self.__point_tree = point.PointTree()
        self.__segment_tree = segment.SegmentTree()
        self.__circle_tree = circle.CircleTree()
        self.__arc_tree = arc.ArcTree()
        self.__hcline_tree = hcline.HCLineTree()
        self.__vcline_tree = vcline.VCLineTree()
        self.__acline_tree = acline.ACLineTree()
        self.__cline_tree = cline.CLineTree()
        self.__ccircle_tree = ccircle.CCircleTree()
        self.__chamfers = []
        self.__fillets = []
        self.__leaders = []
        self.__polylines = []
        self.__textblocks = []
        self.__linear_dimensions = []
        self.__horiz_dimensions = []
        self.__vert_dimensions = []
        self.__radial_dimensions = []
        self.__angular_dimensions = []
        self.__children = []
        self.__visibility = True
        self.__scale = 1.0
        #
        # self.__objdict keeps a reference to all objects stored in the layer
        #
        self.__objdict = {}
        self.__objects = {}
        #
        # self.__externals keeps reference to objects in this layer that are
        # used by objects stored in another layer
        #
        self.__externals = {}

    def __str__(self):
        if self.__parent is None:
            string = "Layer: %s [No Parent]" % self.__name
        else:
            string = "Layer: %s; Parent: %s" % (self.__name, self.__parent.__name)
        return string

    def __contains__(self, obj):
        """Find an object in the Layer.

This method permits the use of 'in' for test conditions.

if obj in layer:
    ....

This function tests for Point, Segment, Circle, etc. It returns
True if there is an equivalent object held in the Layer. Otherwise
the function returns False.
        """
        _seen = False
        if id(obj) in self.__objects:
            _seen = True
        if not _seen:
            if obj in self.__objdict:
                _seen = True
        if not _seen:
            self.update()
            if isinstance(obj, point.Point):
                _seen = obj in self.__point_tree
            elif isinstance(obj, segment.Segment):
                _seen = obj in self.__segment_tree
            elif isinstance(obj, arc.Arc):
                _seen = obj in self.__arc_tree
            elif isinstance(obj, circle.Circle):
                _seen = obj in self.__circle_tree
            elif isinstance(obj, hcline.HCLine):
                _seen = obj in self.__hcline_tree
            elif isinstance(obj, vcline.VCLine):
                _seen = obj in self.__vcline_tree
            elif isinstance(obj, acline.ACLine):
                _seen = obj in self.__acline_tree
            elif isinstance(obj, cline.CLine):
                _seen = obj in self.__cline_tree
            elif isinstance(obj, ccircle.CCircle):
                _seen = obj in self.__ccircle_tree
            elif isinstance(obj, segjoint.Fillet):
                _seen = obj in self.__fillets
            elif isinstance(obj, segjoint.Chamfer):
                _seen = obj in self.__chamfers
            elif isinstance(obj, leader.Leader):
                _seen = obj in self.__leaders
            elif isinstance(obj, polyline.Polyline):
                _seen = obj in self.__polylines
            elif isinstance(obj, text.TextBlock):
                _seen = obj in self.__textblocks
            elif isinstance(obj, dimension.HorizontalDimension):
                _seen = obj in self.__horiz_dimensions
            elif isinstance(obj, dimension.VerticalDimension):
                _seen = obj in self.__vert_dimensions
            elif isinstance(obj, dimension.LinearDimension):
                _seen = obj in self.__linear_dimensions
            elif isinstance(obj, dimension.RadialDimension):
                _seen = obj in self.__radial_dimensions
            elif isinstance(obj, dimension.AngularDimension):
                _seen = obj in self.__angular_dimensions
            else:
                raise TypeError, "Invalid type for in operation: " + `obj`
        return _seen

    def clear(self):
        """Remove all the entities stored in this layer

clear()
        """
        print "clearing layer: " + self.__name
        self.__point_tree.clear()
        self.__segment_tree.clear()
        self.__circle_tree.clear()
        self.__arc_tree.clear()
        self.__hcline_tree.clear()
        self.__vcline_tree.clear()
        self.__acline_tree.clear()
        self.__cline_tree.clear()
        self.__ccircle_tree.clear()
        del self.__chamfers[:]
        del self.__fillets[:]
        del self.__leaders[:]
        del self.__polylines[:]
        del self.__textblocks[:]
        del self.__linear_dimensions[:]
        del self.__horiz_dimensions[:]
        del self.__vert_dimensions[:]
        del self.__radial_dimensions[:]
        del self.__angular_dimensions[:]
        self.__visibility = True
        self.__scale = 1.0
        self.__objdict.clear()
        self.__objects.clear()
        self.__externals.clear()

    def free(self):
        """Disconnect a layer from its parent.

free()
        """
        print "freeing layer: " + self.__name
        _parent = self.__parent
        if _parent is not None:
            _parent.delChild(self)

    def getName(self):
        """Return the name of the Layer.

getName()
        """
        return self.__name

    def setName(self, name):
        """Set the name of the Layer.

setName(name)

The new must be a string, and cannot be None.
        """
        _n = name
        if _n is None:
            raise ValueError, "Layers must have a name."
        if not isinstance(_n, unicode):
            if isinstance(_n, str):
                _n = unicode(_n)
            else:
                raise TypeError, "Invalid name: " + `_n`
        self.__name = _n

    name = property(getName, setName, None, "Layer name.")

    def addLayer(self, l):
        """Add one Layer to another.

addLayer(l)

Add the contents of the Layer object "l" into the Layer. Only
the objects not found in the current Layer are added.
        """
        if not isinstance(l, Layer):
            raise TypeError, "Invalid Layer addition object: " + `l`
        _objmap = {}
        self.update()
        l.update()
        for _obj in l.getLayerEntities("point"):
            _sobj = self.__point_tree.binscan(_obj)
            if _sobj is None:
                self.addObject(_obj)
            else:
                _objmap[_obj] = _sobj
        for _obj in l.getLayerEntities("segment"):
            _sobj = self.__segment_tree.binscan(_obj)
            if _sobj is None:
                _p1, _p2 = _obj.getEndpoints()
                if _p1 in _objmap:
                    _p1 = _objmap[_p1]
                if _p2 in _objmap:
                    _p2 = _objmap[_p2]
                _ns = _obj.clone()
                _ns.setP1(_p1)
                _ns.setP2(_p2)
                _ns.reset()
                self.addObject(_ns)
            else:
                _objmap[_obj] = _sobj
        for _obj in l.getLayerEntities("circle"):
            _sobj = self.__circle_tree.binscan(_obj)
            if _sobj is None:
                _cp = _obj.getCenter()
                if _cp in _objmap:
                    _cp = _objmap[_cp]
                _nc = _obj.clone()
                _nc.setCenter(_cp)
                _nc.reset()
                self.addObject(_nc)
            else:
                _objmap[_obj] = _sobj
        for _obj in l.getLayerEntities("arc"):
            _sobj = self.__arc_tree.binscan(_obj)
            if _sobj is None:
                _cp = _obj.getCenter()
                if _cp in _objmap:
                    _cp = _objmap[_cp]
                _na = _obj.clone()
                _na.setCenter(_cp)
                _na.reset()
                self.addObject(_na)
            else:
                _objmap[_obj] = _sobj
        for _obj in l.getLayerEntities("hcline"):
            _sobj = self.__hcline_tree.binscan(_obj)
            if _sobj is None:
                _lp = _obj.getLocation()
                if _lp in _objmap:
                    _lp = _objmap[_lp]
                self.addObject(hcline.HCLine(_lp))
        for _obj in l.getLayerEntities("vcline"):
            _sobj = self.__vcline_tree.binscan(_obj)
            if _sobj is None:
                _lp = _obj.getLocation()
                if _lp in _objmap:
                    _lp = _objmap[_lp]
                self.addObject(vcline.VCLine(_lp))
        for _obj in l.getLayerEntities("acline"):
            _sobj = self.__acline_tree.binscan(_obj)
            if _sobj is None:
                _lp = _obj.getLocation()
                if _lp in _objmap:
                    _lp = _objmap[_lp]
                self.addObject(acline.ACLine(_lp, _obj.getAngle()))
        for _obj in l.getLayerEntities("cline"):
            _sobj = self.__cline_tree.binscan(_obj)
            if _sobj is None:
                _p1, _p2 = _obj.getKeypoints()
                if _p1 in _objmap:
                    _p1 = _objmap[_p1]
                if _p2 in _objmap:
                    _p2 = _objmap[_p2]
                self.addObject(cline.CLine(_p1, _p2))
        for _obj in l.getLayerEntities("ccircle"):
            _sobj = self.__ccircle_tree.binscan(_obj)
            if _sobj is None:
                _cp = _obj.getCenter()
                if _cp in _objmap:
                    _cp = _objmap[_cp]
                self.addObject(ccircle.CCircle(_cp, _obj.getRadius()))
        for _obj in l.getLayerEntities("chamfer"):
            _s1, _s2 = _obj.getSegments()
            if _s1 in _objmap:
                _s1 = _objmap[_s1]
            if _s2 in _objmap:
                _s2 = _objmap[_s2]
            _nc = segjoint.Chamfer(_s1, _s2, _obj.getLength(),
                                   _obj.getStyle(), _obj.getLinetype(),
                                   _obj.getColor(), _obj.getThickness())
            if _nc not in self.__chamfers:
                self.addObject(_nc)
        for _obj in l.getLayerEntities("fillet"):
            _s1, _s2 = _obj.getSegments()
            if _s1 in _objmap:
                _s1 = _objmap[_s1]
            if _s2 in _objmap:
                _s2 = _objmap[_s2]
            _nf = segjoint.Fillet(_s1, _s2, _obj.getRadius(),
                                  _obj.getStyle(), _obj.getLinetype(),
                                  _obj.getColor(), _obj.getThickness())
            if _nf not in self.__fillets:
                self.addObject(_nf)
        for _obj in l.getLayerEntities("leader"):
            if _obj not in self.__leaders:
                self.addObject(_obj)
        for _obj in l.getLayerEntities("polyline"):
            if _obj not in self.__polylines:
                self.addObject(_obj)
        for _obj in l.getLayerEntities("text"):
            if _obj not in self.__textblocks:
                self.addObject(_obj)
        for _obj in l.getLayerEntities("linear_dimension"):
            if _obj not in self.__linear_dimensions:
                self.addObject(_obj)
        for _obj in l.getLayerEntities("horizontal_dimension"):
            if _obj not in self.__horiz_dimensions:
                self.addObject(_obj)
        for _obj in l.getLayerEntities("vertical_dimension"):
            if _obj not in self.__vert_dimensions:
                self.addObject(_obj)
        for _obj in l.getLayerEntities("radial_dimension"):
            if _obj not in self.__radial_dimensions:
                self.addObject(_obj)
        for _obj in l.getLayerEntities("angular_dimension"):
            if _obj not in self.__angular_dimensions:
                self.addObject(_obj)

    def addObject(self, obj):
        """Add an object to this Layer.

addObject(obj)

The object should be a Point, Segment, Arc, Circle,
HCLine, VCLine, ACLine, CLine, CCircle, Chamfer,
Fillet, Leader, Polyline, or Dimension. Anything else raises
a TypeError exception.
        """
        if id(obj) in self.__objects:
            raise ValueError, "Object already in layer: " + `obj`
        self.update()
        if isinstance(obj, point.Point):
            self._addPoint(obj)
        elif isinstance(obj, segment.Segment):
            self._addSegment(obj)
        elif isinstance(obj, arc.Arc):
            self._addArc(obj)
        elif isinstance(obj, circle.Circle):
            self._addCircle(obj)
        elif isinstance(obj, hcline.HCLine):
            self._addHCLine(obj)
        elif isinstance(obj, vcline.VCLine):
            self._addVCLine(obj)
        elif isinstance(obj, acline.ACLine):
            self._addACLine(obj)
        elif isinstance(obj, ccircle.CCircle):
            self._addCCircle(obj)
        elif isinstance(obj, cline.CLine):
            self._addCLine(obj)
        elif isinstance(obj, segjoint.Chamfer):
            self._addChamfer(obj)
        elif isinstance(obj, segjoint.Fillet):
            self._addFillet(obj)
        elif isinstance(obj, leader.Leader):
            self._addLeader(obj)
        elif isinstance(obj, polyline.Polyline):
            self._addPolyline(obj)
        elif isinstance(obj, text.TextBlock):
            self._addTextBlock(obj)
        elif isinstance(obj, dimension.AngularDimension):
            self._addAngularDimension(obj)
        elif isinstance(obj, dimension.RadialDimension):
            self._addRadialDimension(obj)
        elif isinstance(obj, dimension.HorizontalDimension):
            self._addHorizontalDimension(obj)
        elif isinstance(obj, dimension.VerticalDimension):
            self._addVerticalDimension(obj)
        elif isinstance(obj, dimension.LinearDimension):
            self._addLinearDimension(obj)
        else:
            raise TypeError, "Invalid object for storage: " + `obj`

    def _addPoint(self, p):
        """Add a Point object to the Layer.

_addPoint(p)

This method is private to the Layer object.
        """
        if p not in self.__point_tree:
            self.__point_tree.store(p)
            self.__objects[id(p)] = p

    def _addSegment(self, s):
        """Add a Segment object to the Layer.

_addSegment(s)

The Segment object "s" is added to the Layer if there is not
already a Segment in the layer between the same locations.
The Segment endpoints p1 and p2 _must_ be already stored in
this layer, otherwise a ValueError is raised.

This method is private to the Layer object.
        """
        _p1, _p2 = s.getEndpoints()
        if s in self.__segment_tree:
            _p1.freeUser(s)
            _p2.freeUser(s)
        else:
            if id(_p1) not in self.__objects:
                raise ValueError, "Segment p1 Point not found in Layer."
            if id(_p2) not in self.__objects:
                raise ValueError, "Segment p2 Point not found in Layer."
            self.__segment_tree.store(s)
            self.__objects[id(s)] = s

    def _addCircle(self, c):
        """Add a Circle object to the Layer.

_addCircle(c)

The Circle object "c" is added to the Layer if there is not
already a Circle at that location with the same radius. The
Circle center must already be stored in the Layer, otherwise
a ValueError is raised.

This method is private to the layer object.
        """
        _cp = c.getCenter()
        if c in self.__circle_tree:
            _cp.freeUser(c)
        else:
            if id(_cp) not in self.__objects:
                raise ValueError, "Circle center Point not found in Layer."
            self.__circle_tree.store(c)
            self.__objects[id(c)] = c

    def _addArc(self, a):
        """Add an Arc object to the Layer.

_addArc(a)

The Arc object "a" is added to the Layer if there is not
already a Arc at that location with the same radius, start
angle, and eng_angle. The Arc center must already be stored
in the Layer, otherwise a ValueError is raised.

This method is private to the Layer object.
        """
        _cp = a.getCenter()
        if a in self.__arc_tree:
            _cp.freeUser(a)
        else:
            if id(_cp) not in self.__objects:
                raise ValueError, "Arc center Point not found in Layer."
            self.__arc_tree.store(a)
            self.__objects[id(a)] = a
            for _ep in a.getEndpoints():
                _lp = self.hasPoint(_ep)
                if _lp is None:
                    _lp = point.Point(_ep)
                    self.addObject(_lp)
                _lp.storeUser(a) # this needs to be done better ...

    def _addHCLine(self, hcl):
        """Add an HCLine object to the Layer.

_addHCLine(hcl)

The HCLine object "hcl" is added to the Layer if there is not
already a HCLine at that location. The HCLine location point
must already be stored in the Layer, otherwise a ValueError is
raised.

This method is private the the Layer object.
        """
        _lp = hcl.getLocation()
        if hcl in self.__hcline_tree:
            _lp.freeUser(hcl)
        else:
            if id(_lp) not in self.__objects:
                raise ValueError, "HCLine location Point not found in Layer."
            self.__hcline_tree.store(hcl)
            self.__objects[id(hcl)] = hcl

    def _addVCLine(self, vcl):
        """Add an VCLine object to the Layer.

_addVCLine(vcl)

The VCLine object "vcl" is added to the Layer if there is not
already a VCLine at that location. The VCLine location point
must already be stored in the Layer, otherwise a ValueError is
raised.

This method is private to the Layer object.
        """
        _lp = vcl.getLocation()
        if vcl in self.__vcline_tree:
            _lp.freeUser(vcl)
        else:
            if id(_lp) not in self.__objects:
                raise ValueError, "VCLine location Point not found in Layer."
            self.__vcline_tree.store(vcl)
            self.__objects[id(vcl)] = vcl

    def _addACLine(self, acl):
        """Add an ACLine object to the Layer.

_addACLine(acl)

The ACLine object "acl" is added to the Layer if there is not
already a ACLine at that location. The ACLine location point
must already be stored in the Layer, otherwise a ValueError is
raised.

This method is private to the Layer object.
        """
        _lp = acl.getLocation()
        if acl in self.__acline_tree:
            _lp.freeUser(acl)
        else:
            if id(_lp) not in self.__objects:
                raise ValueError, "ACLine location Point not found in Layer."
            self.__acline_tree.store(acl)
            self.__objects[id(acl)] = acl

    def _addCCircle(self, cc):
        """Add an CCircle object to the Layer.

_addCCircle(cc)

The CCircle object "cc" is added to the Layer if there is not
already a CCircle at that location and with the same radius.
The CCircle center must already be stored in the Layer, otherwise
a ValueError is raised.

This method is private to the Layer object.
        """
        _cp = cc.getCenter()
        if cc in self.__ccircle_tree:
            _cp.freeUser(cc)
        else:
            if id(_cp) not in self.__objects:
                raise ValueError, "CCircle center not found in Layer."
            self.__ccircle_tree.store(cc)
            self.__objects[id(cc)] = cc

    def _addCLine(self, cl):
        """Add an CLine object to the Layer.

_addCLine(cl)

The CLine object "cl" is added to the Layer if there is not
already a CLine at that location. The CLine location point
must already be stored in the Layer, otherwise a ValueError is
raised.

This method is private to the Layer object.
        """
        _p1, _p2 = cl.getKeypoints()
        if cl in self.__cline_tree:
            _p1.freeUser(cl)
            _p2.freeUser(cl)
        else:
            if id(_p1) not in self.__objects:
                raise ValueError, "CLine p1 Point not found in Layer."
            if id(_p2) not in self.__objects:
                raise ValueError, "CLine p2 Point not found in Layer."
            self.__cline_tree.store(cl)
            self.__objects[id(cl)] = cl

    def _addChamfer(self, ch):
        """Add a Chamfer object to the Layer.

_addChamfer(ch)

The Chamfer object "ch" is added to the Layer if there is not
already a Chamfer found joining the two Segments. Both Segment
objects in the chamfer must also be stored in this Layer.

This method is private to the Layer object.
        """
        _s1, _s2 = ch.getSegments()
        if ch in self.__chamfers:
            _s1.freeUser(ch)
            _s2.freeUser(ch)
        else:
            if id(_s1) not in self.__objects:
                raise ValueError, "Chamfer s1 Segment not found in Layer."
            if id(_s2) not in self.__objects:
                raise ValueError, "Chamfer s2 Segment not found in Layer."
            _new_chamfer = True
            for _ch in self.__chamfers:
                if _ch == ch:
                    _new_chamfer = False
                    break
            if _new_chamfer:
                self.__objects[id(ch)] = ch
                self.__chamfers.append(ch)
            else:
                _s1.freeUser(ch)
                _s2.freeUser(ch)
                

    def _addFillet(self, f):
        """Add a Fillet object to the Layer.

_addFillet(f)

The Fillet object "f" is added to the Layer if there is not
already a Fillet found joining the two Segments. Both Segment
objects in the fillet must also be stored in this Layer.

This method is private to the Layer object.
        """
        _s1, _s2 = f.getSegments()
        if f in self.__fillets:
            _s1.freeUser(f)
            _s2.freeUser(f)
        else:
            if id(_s1) not in self.__objects:
                raise ValueError, "Fillet s1 Segment not found in Layer."
            if id(_s2) not in self.__objects:
                raise ValueError, "Fillet s2 Segment not found in Layer."
            _new_fillet = True
            for _f in self.__fillets:
                if _f == f:
                    _new_fillet = False
                    break
            if _new_fillet:
                self.__objects[id(f)] = f
                self.__fillets.append(f)
            else:
                _s1.freeUser(f)
                _s2.freeUser(f)

    def _addLeader(self, l):
        """Add a Leader object to the Layer.

_addLeader(l)

The Leader object "l" is added to the Layer if there is not
already a Leader in the layer between the same locations.
The Leader endpoints p1, p2, and p3 _must_ be already stored in
this layer, otherwise a ValueError is raised.

This method is private to the Layer object.
        """
        _p1, _p2, _p3 = l.getLeaderPoints()
        if l in self.__leaders:
            _p1.freeUser(l)
            _p2.freeUser(l)
            _p3.freeUser(l)
        else:
            if id(_p1) not in self.__objects:
                raise ValueError, "Leader p1 Point not found in Layer."
            if id(_p2) not in self.__objects:
                raise ValueError, "Leader p2 Point not found in Layer."
            if id(_p3) not in self.__objects:
                raise ValueError, "Leader p3 Point not found in Layer."
            self.__leaders.append(l)
            self.__objects[id(l)] = l

    def _addPolyline(self, pl):
        """Add a Polyline object to the Layer.

_addPolyline(pl)

The Polyline object "pl" is added to the Layer if there is not
already a Polyline in the layer at the same locations. The Polyline
points _must_ be already stored in this layer, otherwise a ValueError
is raised.

This method is private to the Layer object.
        """
        _pts = pl.getPoints()
        if pl in self.__polylines:
            for _pt in _pts:
                _pt.freeUser(pl)
        else:
            for _pt in _pts:
                if id(_pt) not in self.__objects:
                    raise ValueError, "Polyline point not in layer."
            self.__polylines.append(pl)
            self.__objects[id(pl)] = pl

    def _addTextBlock(self, tb):
        """Add a TextBlock object to the Layer.

_addTextBlock(tb)

The TextBlock object "tb" is added to the layer if there is not
already a TextBlock in the layer at the same location and with the
same text.
        """
        if tb not in self.__textblocks:
            self.__textblocks.append(tb)
            self.__objects[id(tb)] = tb

    def _addAngularDimension(self, adim):
        """Add an AngularDimension object to the Layer.

_addAngularDimension(adim)

This method is private to the Layer object.
        """
        _p1, _p2, _p3 = adim.getDimPoints()        
        if adim in self.__angular_dimensions:
            _p1.freeUser(adim)
            _p2.freeUser(adim)
            _p3.freeUser(adim)
        else:
            _layer1, _layer2, _layer3 = adim.getDimLayers()
            _lp = _layer1.findObject(_p1)
            if _lp is not _p1:
                raise ValueError, "Dimension Point p1 not found in layer"
            _lp = _layer2.findObject(_p2)
            if _lp is not _p2:
                raise ValueError, "Dimension Point p2 not found in layer!"
            _lp = _layer3.findObject(_p3)
            if _lp is not _p3:
                raise ValueError, "Dimension Point p3 not found in layer!"
            self.__objects[id(adim)] = adim
            self.__angular_dimensions.append(adim)

    def _addRadialDimension(self, rdim):
        """Add a RadialDimension object to the Layer.

_addRadialDimension(rdim)

This method is private to the Layer object.
        """
        _circle = rdim.getDimCircle()        
        if rdim in self.__radial_dimensions:
            _circle.freeUser(rdim)
        else:
            _layer = rdim.getDimLayer()
            _lc = _layer.findObject(_circle)
            if _lc is not _circle:
                raise ValueError, "RadialDimension circular object not found in layer"
            self.__objects[id(rdim)] = rdim
            self.__radial_dimensions.append(rdim)

    def _addHorizontalDimension(self, hdim):
        """Add a HorizontalDimension object to the Layer.

_addHorizontalDimension(hdim)

This method is private to the Layer object.
        """
        _p1, _p2 = hdim.getDimPoints()        
        if hdim in self.__horiz_dimensions:
            _p1.freeUser(hdim)
            _p2.freeUser(hdim)
        else:
            _layer1, _layer2 = hdim.getDimLayers()
            _lp = _layer1.findObject(_p1)
            if _lp is not _p1:
                raise ValueError, "HorizontalDimension Point p1 not found in layer"
            _lp = _layer2.findObject(_p2)
            if _lp is not _p2:
                raise ValueError, "HorizontalDimension Point p2 not found in layer"
            self.__objects[id(hdim)] = hdim
            self.__horiz_dimensions.append(hdim)

    def _addVerticalDimension(self, vdim):
        """Add a VerticalDimension object to the Layer.

_addVerticalDimension(vdim)

This method is private to the Layer object.
        """
        _p1, _p2 = vdim.getDimPoints()        
        if vdim in self.__vert_dimensions:
            _p1.freeUser(vdim)
            _p2.freeUser(vdim)
        else:
            _layer1, _layer2 = vdim.getDimLayers()
            _lp = _layer1.findObject(_p1)
            if _lp is not _p1:
                raise ValueError, "VerticalDimension Point p1 not found in layer"
            _lp = _layer2.findObject(_p2)
            if _lp is not _p2:
                raise ValueError, "VerticalDimension Point p2 not found in layer"
            self.__objects[id(vdim)] = vdim
            self.__vert_dimensions.append(vdim)

    def _addLinearDimension(self, ldim):
        """Add a LinearDimension object to the Layer.

_addLinearDimension(ldim)

This method is private to the Layer object.
        """
        _p1, _p2 = ldim.getDimPoints()        
        if ldim in self.__linear_dimensions:
            _p1.freeUser(ldim)
            _p2.freeUser(ldim)
        else:
            _layer1, _layer2 = ldim.getDimLayers()
            _lp = _layer1.findObject(_p1)
            if _lp is not _p1:
                raise ValueError, "LinearDimension Point p1 not found in layer"
            _lp = _layer2.findObject(_p2)
            if _lp is not _p2:
                raise ValueError, "LinearDimension Point p2 not found in layer"
            self.__objects[id(ldim)] = ldim
            self.__linear_dimensions.append(ldim)

    def delObject(self, obj):
        """Remove an object from this Layer.

delObject(obj)

The object should be a Point, Segment, Arc, Circle,
HCLine, VCLine, ACLine, CLine, CCircle, Chamfer,
Fillet, Leader, or Dimension. Anything else raises
a TypeError exception.
        """
        if id(obj) not in self.__objects:
            raise ValueError, "Object not found in layer: " + `obj`
        self.update()
        if isinstance(obj, point.Point):
            self._delPoint(obj)
        elif isinstance(obj, segment.Segment):
            self._delSegment(obj)
        elif isinstance(obj, arc.Arc):
            self._delArc(obj)
        elif isinstance(obj, circle.Circle):
            self._delCircle(obj)
        elif isinstance(obj, hcline.HCLine):
            self._delHCLine(obj)
        elif isinstance(obj, vcline.VCLine):
            self._delVCLine(obj)
        elif isinstance(obj, acline.ACLine):
            self._delACLine(obj)
        elif isinstance(obj, cline.CLine):
            self._delCLine(obj)
        elif isinstance(obj, ccircle.CCircle):
            self._delCCircle(obj)
        elif isinstance(obj, segjoint.Chamfer):
            self._delChamfer(obj)
        elif isinstance(obj, segjoint.Fillet):
            self._delFillet(obj)
        elif isinstance(obj, leader.Leader):
            self._delLeader(obj)
        elif isinstance(obj, polyline.Polyline):
            self._delPolyline(obj)
        elif isinstance(obj, text.TextBlock):
            self._delTextBlock(obj)
        elif isinstance(obj, dimension.AngularDimension):
            self._delAngularDimension(obj)
        elif isinstance(obj, dimension.RadialDimension):
            self._delRadialDimension(obj)
        elif isinstance(obj, dimension.HorizontalDimension):
            self._delHorizontalDimension(obj)
        elif isinstance(obj, dimension.VerticalDimension):
            self._delVerticalDimension(obj)
        elif isinstance(obj, dimension.LinearDimension):
            self._delLinearDimension(obj)
        else:
            raise TypeError, "Invalid object for removal: " + `obj`

    def _delPoint(self, p):
        """Delete a Point from the Layer.

_delPoint(p)

This method is private to the Layer object.
        """
        _delete = True
        _users = p.getUsers()
        for _user in _users:
            if not isinstance(_user, dimension.Dimension):
                _delete = False
                break
        if _delete:
            _key = id(p)
            if _key in self.__externals:
                print "point in external ..."
                for _layer, _extobj in self.__externals[_key]:
                    _layer.delObject(_extobj)
            if _users is not None:
                # _dims = _users[:] # copy _users - delObject() modifies it!
                for _user in _users:
                    if id(_user) not in self.__objects:
                        print "point bound to object not in layer: " + `_user`
                        _delete = False
                if _delete:
                    for _user in _users:
                        self.delObject(_user)
            if _delete:
                del self.__objects[id(p)]
                self.__point_tree.remove(p)

    def _connectUser(self, obj, user):
        """Store a reference between an object and its user in the objdict.

_connectUser(obj, user)

This method is private to the Layer.
        """
        assert obj in self.__objdict, "Object not in objdict: " + str(obj)
        assert user in self.__objdict, "User not in objdict: " + str(user)
        if self.__objdict[obj] is None:
            self.__objdict[obj] = []
        _users = self.__objdict[obj]
        _seen = False
        for _user in _users:
            if _user is user:
                _seen = True
                break
        if not _seen:
            _users.append(user)

    def _disconnectUser(self, obj, user):
        """Remove the linkage in the objdict of user of some object

_disconnectUser(obj, user)

This method is private to the Layer.
        """
        assert obj in self.__objdict, "Object not in objdict: " + str(obj)
        assert user in self.__objdict, "User not in objdict: " + str(user)
        _users = self.__objdict[obj]
        assert _users is not None, "No users given for object: " + str(obj)
        for _i in xrange(len(_users)):
            if user is _users[_i]:
                del _users[_i]
                break
        if not len(_users):
            self.__objdict[obj] = None

    def _delSegment(self, s):
        """Delete a Segment from the Layer.

_delSegment(s)

This method is private to the Layer object.
        """
        _users = s.getUsers()
        if _users is not None: # chamfers, fillets, or hatching
            for _obj in _users:
                self.delObject(_obj) # restore other segments?
        _p1, _p2 = s.getEndpoints()
        assert id(_p1) in self.__objects, "Segment p1 Point not in objects"
        _p1.freeUser(s)
        assert id(_p2) in self.__objects, "Segment p2 Point not in objects"
        _p2.freeUser(s)
        del self.__objects[id(s)]
        self.__segment_tree.remove(s)
        self.delObject(_p1) # remove possibly unused point _p1
        self.delObject(_p2) # remove possibly unused point _p2

    def _delCircle(self, c):
        """Delete a Circle from the Layer.

_delCircle(c)

This method is private to the Layer object.
        """
        assert not isinstance(c, arc.Arc), "Arc in _delCircle()"
        _users = c.getUsers()
        if _users is not None: # dimensions or hatching
            for _user in _users:
                if isinstance(_user, dimension.RadialDimension):
                    _layer = _user.getDimLayer()
                    _layer.delObject(_user)
                else:
                    self.delObject(_user)
        _cp = c.getCenter()
        assert id(_cp) in self.__objects, "Circle center point not in objects"
        _cp.freeUser(c)
        del self.__objects[id(c)]
        self.__circle_tree.remove(c)
        self.delObject(_cp) # remove possibly unused point _cp

    def _delArc(self, a):
        """Delete an Arc from the Layer.

_delArc(a)

This method is private to the Layer object.
        """
        _users = a.getUsers() # dimensions or hatching
        if _users is not None:
            for _user in _users:
                if isinstance(_user, dimension.RadialDimension):
                    _layer = _user.getDimLayer()
                    _layer.delObject(_user)
                else:
                    self.delObject(_user)
        _cp = a.getCenter()
        assert id(_cp) in self.__objects, "Arc center point not in objects"
        _cp.freeUser(a)
        for _ep in a.getEndpoints():
            _p = self.hasPoint(_ep)
            assert _p is not None, "Arc endpoint not found in layer"
            assert id(_p) in self.__objects, "Arc endpoint not in objects"
            _p.freeUser(a)
            self.delObject(_p) # remove possibly unused point _p
        del self.__objects[id(a)]
        self.__arc_tree.remove(a)
        self.delObject(_cp) # remove possibly unused point _cp

    def _delHCLine(self, hcl):
        """Remove a HCLine object from the Layer.

_delHCLine(hcl)

This method is private to the Layer object.
        """
        _lp = hcl.getLocation()
        assert id(_lp) in self.__objects, "HCLine point not in objects"
        _lp.freeUser(hcl)
        del self.__objects[id(hcl)]
        self.__hcline_tree.remove(hcl)
        self.delObject(_lp) # remove possibly unused point _lp

    def _delVCLine(self, vcl):
        """Remove a VCLine object from the Layer.

_delVCLine(vcl)

This method is private to the Layer object.
        """
        _lp = vcl.getLocation()
        assert id(_lp) in self.__objects, "VCLine point not in objects"
        _lp.freeUser(vcl)
        del self.__objects[id(vcl)]
        self.__vcline_tree.remove(vcl)
        self.delObject(_lp) # remove possibly unused point _lp

    def _delACLine(self, acl):
        """Remove an ACLine object from the Layer.

_delACLine(acl)

This method is private to the Layer object.
        """
        _lp = acl.getLocation()
        assert id(_lp) in self.__objects, "ACLine point not in objects"
        _lp.freeUser(acl)
        del self.__objects[id(acl)]
        self.__acline_tree.remove(acl)
        self.delObject(_lp) # remove possibly unused point _lp

    def _delCLine(self, cl):
        """Delete a CLine from the Layer.

_delCLine(s)

This method is private to the Layer object.
        """
        _p1, _p2 = cl.getKeypoints()
        assert id(_p1) in self.__objects, "CLine point p1 not in objects"
        _p1.freeUser(cl)
        assert id(_p2) in self.__objects, "CLine point p2 not in objects"
        _p2.freeUser(cl)
        del self.__objects[id(cl)]
        self.__cline_tree.remove(cl)
        self.delObject(_p1) # remove possibly unused point _p1
        self.delObject(_p2) # remove possibly unused point _p2

    def _delCCircle(self, cc):
        """Delete a CCircle from the Layer.

_delCCircle(cc)

This method is private to the Layer object.
        """
        _cp = cc.getCenter()
        assert id(_cp) in self.__objects, "CCircle center point not in objects"
        _cp.freeUser(cc)
        del self.__objects[id(cc)]
        self.__ccircle_tree.remove(cc)
        self.delObject(_cp) # remove possibly unused point _cp

    def _delChamfer(self, ch):
        """Remove a Chamfer from the Layer.

_delChamfer(ch)

This method is private to the Layer.
        """
        _chamfers = self.__chamfers
        _idx = None
        for _i in xrange(len(_chamfers)):
            if ch is _chamfers[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost chamfer from list"
        _users = ch.getUsers()
        if _users is not None: # could be hatching ...
            for _user in _users:
                self.delObject(_user)
        _s1, _s2 = ch.getSegments()
        assert id(_s1) in self.__objects, "Chamfer s1 segment not in objects"
        _s1.freeUser(ch)
        assert id(_s2) in self.__objects, "Chamfer s2 segment not in objects"
        _s2.freeUser(ch)
        del self.__objects[id(ch)]
        del _chamfers[_idx] # restore the things the chamfer connects?

    def _delFillet(self, f):
        """Remove a Fillet from the Layer.

_delFillet(f)

This method is private to the Layer.
        """
        _fillets = self.__fillets
        _idx = None
        for _i in xrange(len(_fillets)):
            if f is _fillets[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost fillet from list"
        _users = f.getUsers()
        if _users is not None: # could be hatching ...
            for _user in _users:
                self.delObject(_user)
        _s1, _s2 = f.getSegments()
        assert id(_s1) in self.__objects, "Fillet s1 segment not in objects"
        _s1.freeUser(f)
        assert id(_s2) in self.__objects, "Fillet s2 segment not in objects"
        _s2.freeUser(f)
        del self.__objects[id(f)]
        del _fillets[_idx] # restore the things the fillet connects?

    def _delLeader(self, l):
        """Delete a Leader from the Layer.

_delLeader(l)

This method is private to the Layer object.
        """
        _leaders = self.__leaders
        _idx = None
        for _i in xrange(len(_leaders)):
            if l is _leaders[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost leader from list"
        _p1, _p2, _p3 = l.getLeaderPoints()
        assert id(_p1) in self.__objects, "Leader p1 Point not in objects"
        _p1.freeUser(l)
        assert id(_p2) in self.__objects, "Leader p2 Point not in objects"
        _p2.freeUser(l)
        assert id(_p3) in self.__objects, "Leader p3 Point not in objects"
        _p3.freeUser(l)
        del self.__objects[id(l)]
        del _leaders[_idx]
        self.delObject(_p1) # remove possibly unused point _p1
        self.delObject(_p2) # remove possibly unused point _p2
        self.delObject(_p3) # remove possibly unused point _p3

    def _delPolyline(self, pl):
        """Delete a Polyline from the Layer.

_delPolyline(pl)

This method is private to the Layer object.
        """
        _polylines = self.__polylines
        _idx = None
        for _i in xrange(len(_polylines)):
            if pl is _polylines[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost polyline from list"
        _users = pl.getUsers()
        if _users is not None: # could be hatching
            for _obj in _users:
                self.delObject(_obj)
        _pts = pl.getPoints()
        for _pt in _pts:
            assert id(_pt) in self.__objects, "Polyline point not in objects"
            _pt.freeUser(pl)
        del self.__objects[id(pl)]
        del _polylines[_idx]
        for _pt in _pts:
            self.delObject(_pt) # remove possibly unused point _pt

    def _delTextBlock(self, tb):
        """Delete a TextBlock from the Layer.

_delTextBlock(tb)

This method is private to the Layer object.
        """
        _tbs = self.__textblocks
        _idx = None
        for _i in xrange(len(_tbs)):
            if tb is _tbs[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost textblock in list"
        del self.__objects[id(tb)]
        del _tbs[_idx]

    def _delAngularDimension(self, adim):
        """Delete an AngularDimension from the Layer.

_delAngularDimension(adim)

This method is private to the Layer object.
        """
        _adims = self.__angular_dimensions
        _idx = None
        for _i in xrange(len(_adims)):
            if adim is _adims[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost adim in list"
        _layer1, _layer2, _layer3 = adim.getDimLayers()
        _p1, _p2, _p3 = adim.getDimPoints()
        if _layer1 is self:
            assert id(_p1) in self.__objects, "ADim p1 not in objects"
        _p1.freeUser(adim)        
        if _layer2 is self:
            assert id(_p2) in self.__objects, "ADim p2 not in objects"
        _p2.freeUser(adim)
        if _layer3 is self:
            assert id(_p3) in self.__objects, "ADim p3 not in objects"
        _p3.freeUser(adim)
        del self.__objects[id(adim)]
        del _adims[_idx]

    def _delRadialDimension(self, rdim):
        """Delete a RadialDimension from the Layer.

_delRadialDimension(rdim)

This method is private to the Layer object.
        """
        _rdims = self.__radial_dimensions
        _layer = rdim.getDimLayer()
        _circ = rdim.getDimCircle()
        if _layer is self:
            assert id(_circ) in self.__objects, "RDim circle not in objects"
        _circ.freeUser(rdim)
        for _i in xrange(len(_rdims)):
            if rdim is _rdims[_i]:
                del _rdims[_i]
                break
        del self.__objects[id(rdim)]

    def _delHorizontalDimension(self, hdim):
        """Delete an HorizontalDimension from the Layer.

_delHorizontalDimension(hdim)

This method is private to the Layer object.
        """
        _hdims = self.__horiz_dimensions
        _idx = None
        for _i in xrange(len(_hdims)):
            if hdim is _hdims[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost hdim in list"
        _layer1, _layer2 = hdim.getDimLayers()
        _p1, _p2 = hdim.getDimPoints()
        if _layer1 is self:
            assert id(_p1) in self.__objects, "HDim p1 not in objects"
        _p1.freeUser(hdim)
        if _layer2 is self:
            assert id(_p2) in self.__objects, "HDim p2 not in objects"
        _p2.freeUser(hdim)
        del self.__objects[id(hdim)]
        del _hdims[_idx]

    def _delVerticalDimension(self, vdim):
        """Delete an VerticalDimension from the Layer.

_delVerticalDimension(vdim)

This method is private to the Layer object.
        """
        _vdims = self.__vert_dimensions
        _idx = None
        for _i in xrange(len(_vdims)):
            if vdim is _vdims[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost vdim in list"
        _layer1, _layer2 = vdim.getDimLayers()
        _p1, _p2 = vdim.getDimPoints()
        if _layer1 is self:
            assert id(_p1) in self.__objects, "VDim p1 not in objects"
        _p1.freeUser(vdim)
        if _layer2 is self:
            assert id(_p2) in self.__objects, "VDim p2 not in objects"
        _p2.freeUser(vdim)
        del self.__objects[id(vdim)]
        del _vdims[_idx]

    def _delLinearDimension(self, ldim):
        """Delete an LinearDimension from the Layer.

_delLinearDimension(ldim)

This method is private to the Layer object.
        """
        _ldims = self.__linear_dimensions
        _idx = None
        for _i in xrange(len(_ldims)):
            if ldim is _ldims[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost ldim in list"
        _layer1, _layer2 = ldim.getDimLayers()
        _p1, _p2 = ldim.getDimPoints()
        if _layer1 is self:
            assert id(_p1) in self.__objects, "LDim p1 not in objects"
        _p1.freeUser(ldim)
        if _layer2 is self:
            assert id(_p2) in self.__objects, "LDim p2 not in objects"
        _p2.freeUser(ldim)
        del self.__objects[id(ldim)]
        del _ldims[_idx]

    def usesObject(self, obj):
        """Return a list of objects tied to another object

usesObject(obj)

The object "obj" should be a Point, Segment, Circle, Arc,
Chamfer, or Fillet.
        """
        _users = []
        if isinstance(obj, baseobject.Subpart):
            _users.extend(obj.getUsers())
        return _users

    def countUsers(self, obj):
        """Return the number of objects using a particular object.

countUsers(obj)

This method returns an integer value.
        """
        if id(obj) not in self.__objects:
            raise ValueError, "Object not stored in layer: " + `obj`
        _count = 0
        if isinstance(obj, baseobject.Subpart):
            _count = obj.countUsers()
        return _count

    def findObject(self, obj):
        """Return an object in the layer that is equivalent to a test object.

findObject(obj)

This method returns None if a suitable object is not found.
        """
        _retobj = None
        if id(obj) in self.__objects:
            _retobj = obj
        else:
            self.update()
            _objtree = None
            _objlist = None
            if isinstance(obj, point.Point):
                _objtree = self.__point_tree
            elif isinstance(obj, segment.Segment):
                _objtree = self.__segment_tree
            elif isinstance(obj, arc.Arc):
                _objtree = self.__arc_tree
            elif isinstance(obj, circle.Circle):
                _objtree = self.__circle_tree
            elif isinstance(obj, hcline.HCLine):
                _objtree = self.__hcline_tree
            elif isinstance(obj, vcline.VCLine):
                _objtree = self.__vcline_tree
            elif isinstance(obj, acline.ACLine):
                _objtree = self.__acline_tree
            elif isinstance(obj, cline.CLine):
                _objtree = self.__cline_tree
            elif isinstance(obj, ccircle.CCircle):
                _objtree = self.__ccircle_tree
            elif isinstance(obj, segjoint.Fillet):
                _objlist = self.__fillets
            elif isinstance(obj, segjoint.Chamfer):
                _objlist = self.__chamfers
            elif isinstance(obj, leader.Leader):
                _objlist = self.__leaders
            elif isinstance(obj, polyline.Polyline):
                _objlist = self.__polylines
            elif isinstance(obj, text.TextBlock):
                _objlist = self.__textblocks
            elif isinstance(obj, dimension.HorizontalDimension):
                _objlist = self.__horiz_dimensions
            elif isinstance(obj, dimension.VerticalDimension):
                _objlist = self.__vert_dimensions
            elif isinstance(obj, dimension.LinearDimension):
                _objlist =  self.__linear_dimensions
            elif isinstance(obj, dimension.RadialDimension):
                _objlist = self.__radial_dimensions
            elif isinstance(obj, dimension.AngularDimension):
                _objlist = self.__angular_dimensions
            else:
                raise TypeError, "Invalid type for in operation: " + `obj`
            if _objtree is not None:
                _retobj = _objtree.binscan(obj)
            elif _objlist is not None:
                _seen = True
                try:
                    _idx = _objlist.index(obj)
                except ValueError:
                    _seen = False
                if _seen:
                    _retobj = _objlist[_idx]
        return _retobj

    def hasPoint(self, p, tol=tolerance.TOL):
        """Find a Point in the layer.

hasPoint(p [,tol])

There is a single required argument:

p: Either a Point object or a tuple of two-floats

There is one optional argument:

tol: A float equal or greater than 0

The function returns an existing Point object found at
the same location, or None.
        """
        _p = p
        if not isinstance(_p, point.Point):
            _p = point.Point(p)
        _lp = None
        if id(_p) in self.__objects:
            _lp = _p
        else:
            _t = tolerance.toltest(tol)
            self.update()
            _lp = self.__point_tree.binscan(_p, _t)
        return _lp

    def hasPointAt(self, x, y):
        """Find a Point in the layer at a specific location.

hasPointAt(x, y)

The two required arguments are:

x: A float giving the "x" coordinate
y: A float giving the "y" coordinate

This method returns a the Point if one is found at
the location, or None otherwise.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.update()
        return self.__point_tree.atCoords(_x, _y)

    def hasCircleAt(self, r, x, y):
        """Find a Circle of some radius in the layer at a specific location.

hasCircleAt(r, x, y)

The three required arguments are:

r: A float giving the Circle radius
x: A float giving the "x" coordinate
y: A float giving the "y" coordinate

This method returns a the Circle if one is found at
the location, or None otherwise.
        """
        _r = r
        if not isinstance(_r, float):
            _r = float(_r)
        if _r < 1e-10:
            raise ValueError, "Invalid radius: %g" % _r
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.update()
        return self.__circle_tree.atCoords(_r, _x, _y)

    def extRef(self, obj, ext, l):
        """Store a reference of an object in another layer in this layer.

extRef(obj, ext, l)

There are three required arguments to this function:

obj: An object in this layer - currently this must be a Point or a Circle
ext: An object in another layer
l: The layer holding the external object

The object in the other layer must call this function so
that a reference to itself can be associated to the object
stored in this layer.
        """
        warnings.warn("extRef() is deprecated", DeprecationWarning,
                      stacklevel=2)
        if not isinstance(obj, (point.Point, circle.Circle)):
            raise TypeError, "Invalid object: " + `obj`
        if id(obj) not in self.__objects:
            raise ValueError, "Object not found in layer: " + `obj`
        if ext is None:
            raise ValueError, "External object cannot be None."
        if not isinstance(l, Layer):
            raise TypeError, "Invalid layer argument: " + `l`
        if l is self:
            raise ValueError, "extRef() called for self as layer"
        _key = id(obj)
        if _key not in self.__externals:
            self.__externals[_key] = []
        self.__externals[_key].append((l, ext))

    def extDeref(self, obj, ext):
        """Release a reference to an object in another layer from a object.

extDeref(obj, ext)

There are two required arguments to this function:

obj: An object in this layer - Currently this must be a Point or a Circle
ext: An object in another layer

The object in the other layer must call this function so
that a reference to itself can be removed from the external
reference dictionary kept in this layer.
        """
        warnings.warn("extRef() is deprecated", DeprecationWarning,
                      stacklevel=2)
        if not isinstance(obj, (point.Point, circle.Circle)):
            raise TypeError, "Invalid object: " + `obj`
        if id(obj) not in self.__objects:
            raise ValueError, "Object not found in this layer: " + `obj`
        _key = id(obj)
        if _key in self.__externals:
            _users = self.__externals[_key]
            for _i in xrange(len(_users)):
                _layer, _extobj = _users[_i]
                if ext is _extobj:
                    del _users[_i]
                    break
            if not len(self.__externals[_key]):
                del self.__externals[_key]

    def getExtRefs(self, obj):
        """Return the list of external objects tied to an object in this layer.

getExtRefs(obj)

The argument obj must be an object in this layer. This method returns
a list of objects.
        """
        warnings.warn("getExtRef() is deprecated", DeprecationWarning,
                      stacklevel=2)
        _extobjs = []
        if isinstance(obj, baseobject.Subpart):
            for _user in obj.getUsers():
                if id(_user) not in self.__objects:
                    _extobjs.append(_user)
        return _extobjs

    def mapPoint(self, p, tol=tolerance.TOL, count=2):
        """Find a Point in the layer

mapPoint(p [,tol, count])

There is a single required argument:

p: Either a Point object or a tuple of two-floats

There are two optional arguments:

tol: A float equal or greater than 0 for distance tolerance comparisons.
count: An integer value indicating the largest number of objects to
       return. By default this value is 2.

Setting "count" to None or a negative value will result in the maximum
number of objects being unlimited.

This method tests the objects in the Layer to see if the
Point can can be mapped on to any of them. The returned list
consists of tuples in the form:

(obj, pt)

Where "obj" is the object the point was mapped to and "pt"
is the projected point on the object.
        """
        _scan = True
        _hits = []
        _p = p
        if not isinstance(_p, point.Point):
            _p = point.Point(p)
        _t = tolerance.toltest(tol)
        _count = count
        if _count is None:
            _count = sys.maxint
        else:
            if not isinstance(_count, int):
                _count = int(count)
            if _count < 0:
                _count = sys.maxint
        if _count == 0: # bail out, but why set count to this value?
            return _hits
        self.update()
        #
        # scan for non-point objects first, then look at points,
        # because there will not be any way the same object can
        # be found in differnt trees/lists
        #
        if _scan:
            for _tup in self.__segment_tree.mapPoint(_p, _t, _count):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan:
            for _tup in self.__circle_tree.mapPoint(_p, _t, _count):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan:
            for _tup in self.__arc_tree.mapPoint(_p, _t, _count):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan: # max of one object with hcline_tree
            for _tup in self.__hcline_tree.mapPoint(_p, _t):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan: # max of one object with vcline_tree
            for _tup in self.__vcline_tree.mapPoint(_p, _t):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan:
            for _tup in self.__acline_tree.mapPoint(_p, _t, _count):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan:
            for _tup in self.__ccircle_tree.mapPoint(_p, _t, _count):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan:
            for _tup in self.__cline_tree.mapPoint(_p, _t, _count):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan:
            for _leader in self.__leaders:
                _pt = _leader.mapPoint(_p, _t)
                if _pt is not None:
                    _hits.append((_leader, _pt))
                    if len(_hits) == _count:
                        break
        if _scan:
            for _polyline in self.__polylines:
                _pt = _polyline.mapPoint(_p, _t)
                if _pt is not None:
                    _hits.append((_polyline, _pt))
                    if len(_hits) == _count:
                        break
        _px, _py = _p.getCoords()
        if _scan:
            for _ldim in self.__linear_dimensions:
                if _ldim.onDimension(_px, _py, _t):
                    _hits.append((_ldim, None))
                    if len(_hits) == _count:
                        break
        if _scan:
            for _hdim in self.__horiz_dimensions:
                if _hdim.onDimension(_px, _py, _t):
                    _hits.append((_hdim, None))
                    if len(_hits) == _count:
                        break
        if _scan:
            for _vdim in self.__vert_dimensions:
                if _vdim.onDimension(_px, _py, _t):
                    _hits.append((_vdim, None))
                    if len(_hits) == _count:
                        break
        if _scan:
            for _rdim in self.__radial_dimensions:
                if _rdim.onDimension(_px, _py, _t):
                    _hits.append((_rdim, None))
                    if len(_hits) == _count:
                        break
        if _scan:
            for _adim in self.__angular_dimensions:
                if _adim.onDimension(_px, _py, _t):
                    _hits.append((_adim, None))
                    if len(_hits) == _count:
                        break
        #
        # scan for point, but do not append any object that
        # has already been added to the hit list
        #
        if _scan:
            _objs = {}
            for _obj, _pt in _hits:
                _objs[id(_obj)] = True
            _lp = self.hasPoint(_p, _t)
            if _lp is not None:
                assert id(_lp) in self.__objects, "Point not in objects"
                _users = _lp.getUsers()
                if _users is not None:
                    for _user in _users:
                        _uid = id(_user)
                        if _uid not in _objs:
                            _objs[_uid] = True
                            _hits.append((_user, _lp))
                            if len(_hits) == _count:
                                _scan = False
                                break
                    
        return _hits

    def getLayerEntities(self, entity):
        """Get all of a particular type of entity in the Layer.

getLayerEntities(entity)

The argument "entity" should be one of the following:
point, segment, circle, arc, hcline, vcline, acline,
cline, ccircle, chamfer, fillet, leader, polyline,
textblock, linear_dimension, horizontal_dimenions,
vertical_dimension, radial_dimension, or angular_dimension.
        """
        if entity == "point":
            objs = self.__point_tree[:]
        elif entity == "segment":
            objs = self.__segment_tree[:]
        elif entity == "circle":
            objs = self.__circle_tree[:]
        elif entity == "arc":
            objs = self.__arc_tree[:]
        elif entity == "hcline":
            objs = self.__hcline_tree[:]
        elif entity == "vcline":
            objs = self.__vcline_tree[:]
        elif entity == "acline":
            objs = self.__acline_tree[:]
        elif entity == "cline":
            objs = self.__cline_tree[:]
        elif entity == "ccircle":
            objs = self.__ccircle_tree[:]
        elif entity == "chamfer":
            objs = self.__chamfers[:]
        elif entity == "fillet":
            objs = self.__fillets[:]
        elif entity == "leader":
            objs = self.__leaders[:]
        elif entity == "polyline":
            objs = self.__polylines[:]
        elif entity == "text":
            objs = self.__textblocks[:]
        elif entity == "linear_dimension":
            objs = self.__linear_dimensions[:]
        elif entity == "horizontal_dimension":
            objs = self.__horiz_dimensions[:]
        elif entity == "vertical_dimension":
            objs = self.__vert_dimensions[:]
        elif entity == "radial_dimension":
            objs = self.__radial_dimensions[:]
        elif entity == "angular_dimension":
            objs = self.__angular_dimensions[:]
        else:
            raise ValueError, "Invalid layer entity: " + `entity`
        return objs

    def addChild(self, child):
        """Add a child Layer to a Layer.

addChild(child)

There is no limit to the number of children for a Layer.
        """
        if not isinstance(child, Layer):
            raise TypeError, "Invalid child Layer: " + `child`
        child.__parent = self
        self.__children.append(child)

    def delChild(self, child):
        """Remove a child Layer from the Layer.

delChild(child)

The child layer will not be removed if it has children
of its own.
        """
        if not isinstance(child, Layer):
            raise TypeError, "Invalid child Layer: " + `child`
        _seen = False
        _index = 0
        for _kid in self.__children:
            if child is _kid:
                _seen = True
                break
            _index = _index + 1
        if not _seen:
            raise ValueError, "Invalid child Layer: " + `child`
        if not child.hasChildren():
            del self.__children[_index]

    def hasChildren(self):
        """Test if there are children of this Layer.

hasChildren()

The function returns 1 if the Layer has children, 0 otherwise.
        """
        _val = False
        if len(self.__children) > 0:
            _val = True
        return _val

    def getChildren(self):
        """Get the children of a Layer.

getChildren()

This function returns a list of children Layers for a Layer.
        """
        return self.__children[:]

    def getParent(self):
        """Get the Layer's parent.

getParent()
        """
        return self.__parent

    def setParent(self, parent):
        """Set the Layer's parent.

setParent(parent)

Set the parent layer for a Layer object. The parent cannot
be set to None.
        """
        if parent is None:
            raise ValueError, "Layer parent cannot be set to None."
        if not isinstance(parent, Layer):
            raise TypeError, "Invalid parent layer: " + `parent`
        self.__parent = parent

    parent = property(getParent, setParent, None, "Parent of a layer.")

    def getVisibility(self):
        """Return the visibility value of the Layer.

getVisibility()
        """
        return self.__visibility

    def setVisibility(self, val):
        """Set the visibility of the Layer.

setVisibility(val)

The argument "val" should be either "True" or "False".
        """
        if val is not True and val is not False:
            raise ValueError, "Invalid visibility value: " + `val`
        self.__visibility = val

    visibility = property(getVisibility, setVisibility, None, "Visibility.")

    def getScale(self):
        """Return the scale factor of the Layer.

getScale()
        """
        return self.__scale

    def setScale(self, scale):
        """Set the scale factor for the Layer.

setScale(scale)

The scale factor must be a positive float value greater than 0.0
        """
        _s = scale
        if not isinstance(_s, float):
            _s = float(scale)
        if _s < 1e-10:
            raise ValueError, "Invalid scale factor: %g" % _s
        self.__scale = _s

    scale = property(getScale, setScale, None, "Layer scale factor.")

    def getBoundary(self):
        """Return the maximum and minimum values of the object in the Layer.

getBoundary()

The function returns a tuple holding four float values:

(xmin, ymin, xmax, _ymax)

A default value of (-1.0, -1.0, 1.0, 1.0) is returned for a Layer
containing no objects.
        """
        _xmin = None
        _ymin = None
        _xmax = None
        _ymax = None
        for _obj in self.__point_tree:
            _x, _y = _obj.getCoords()
            if _xmin is None or _x < _xmin:
                _xmin = _x
            if _ymin is None or _y < _ymin:
                _ymin = _y
            if _xmax is None or _x > _xmax:
                _xmax = _x
            if _ymax is None or _y > _ymax:
                _ymax = _y
        _ctrees = self.__circle_tree + self.__arc_tree + self.__ccircle_tree
        for _obj in _ctrees:
            _x, _y = _obj.getCenter().getCoords()
            _rad = _obj.getRadius()
            _val = _x - _rad
            if _xmin is None or _val < _xmin:
                _xmin = _val
            _val = _y - _rad
            if _ymin is None or _val < _ymin:
                _ymin = _val
            _val = _x + _rad
            if _xmax is None or _val > _xmax:
                _xmax = _val
            _val = _y + _rad
            if _ymax is None or _val > _ymax:
                _ymax = _val
        _dims = (self.__linear_dimensions +
                 self.__horiz_dimensions +
                 self.__vert_dimensions +
                 self.__radial_dimensions +
                 self.__angular_dimensions)
        for _dim in _dims:
            _dx, _dy = _dim.getLocation()
            _dxmin, _dymin, _dxmax, _dymax = _dim.getBounds()
            if _xmin is None or _dxmin < _xmin:
                _xmin = _dxmin
            if _ymin is None or _dymin < _ymin:
                _ymin = _dymin
            if _xmax is None or _dxmax > _xmax:
                _xmax = _dxmax
            if _ymax is None or _dymax > _ymax:
                _ymax = _dymax
        for _textblock in self.__textblocks:
            _x, _y = _textblock.getLocation() # need to get boundary of text
            if _xmin is None or _x < _xmin:
                _xmin = _x
            if _ymin is None or _y < _ymin:
                _ymin = _y
            if _xmax is None or _x > _xmax:
                _xmax = _x
            if _ymax is None or _y > _ymax:
                _ymax = _y
        if _xmin is None: _xmin = -1.0
        if _ymin is None: _ymin = -1.0
        if _xmax is None: _xmax = 1.0
        if _ymax is None: _ymax = 1.0
        return _xmin, _ymin, _xmax, _ymax

    def objsInRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return a all the objects in the Layer visible within the bounds.

objsInRegion(xmin, ymin, xmax, ymax[, fully])

The function has four required arguments:

xmin: The minimum x-value of the region
ymin: The minimum y-value of the region
xmax: The maximum x-value of the region
ymax: The maximum y-value of the region

There is a single optional argument:

fully: A True/False value indicating if the object must be
       entirely within the region [fully=True], or can
       merely pass through [fully=False]. The default value
       is False.

The function returns a list of objects.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if xmax < xmin:
            raise ValueError, "Value error: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Value error: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        self.update()
        _objs = []
        for _obj in self.__point_tree:
            _x, _y = _obj.getCoords()
            if _x > _xmax:
                break
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__segment_tree:
            _p1, _p2 = _obj.getEndpoints()
            _p1x, _p1y = _p1.getCoords()
            _p2x, _p2y = _p2.getCoords()
            _pxmin = min(_p1x, _p2x) - 1e-10
            if _pxmin > _xmax:
                break
            _pymin = min(_p1y, _p2y)
            _pxmax = max(_p1x, _p2x)
            _pymax = max(_p1y, _p2y)
            if _xmin > _pxmax or _ymin > _pymax or _ymax < _pymin:
                continue
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax, fully):
                _objs.append(_obj)
        _ctrees = (self.__circle_tree +
                   self.__arc_tree +
                   self.__ccircle_tree)
        for _obj in _ctrees:
            _x, _y = _obj.getCenter().getCoords()
            _rad = _obj.getRadius()
            _cxmin = _x - _rad
            if _cxmin > _xmax:
                break
            _cymin = _y - _rad
            _cxmax = _x + _rad
            _cymax = _y + _rad
            if _xmin > _cxmax or _ymin > _cymax or _ymax < _cymin:
                continue
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax, fully):
                _objs.append(_obj)
        for _obj in self.__hcline_tree:
            _x, _y = _obj.getLocation().getCoords()
            if _y > _ymax:
                break
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__vcline_tree:
            _x, _y = _obj.getLocation().getCoords()
            if _x > _xmax:
                break
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__acline_tree:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__cline_tree:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__chamfers:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__leaders:
            _p1, _p2, _p3 = _obj.getLeaderPoints()
            _p1x, _p1y = _p1.getCoords()
            _p2x, _p2y = _p2.getCoords()
            _p3x, _p3y = _p3.getCoords()
            if _xmin > max(_p1x, _p2x, _p3x):
                break
            if ((_xmax < min(_p1x, _p2x, _p3x)) or
                (_ymax < min(_p1y, _p2y, _p3y)) or
                (_ymin > max(_p1y, _p2y, _p3y))):
                continue
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__polylines:
            _pxmin, _pymin, _pxmax, _pymax = _obj.getBounds()
            if _xmin > _pxmax:
                break
            if _xmax < _pxmin or _ymax < _pymin or _ymin > _pymax:
                continue
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__fillets:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__linear_dimensions:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__horiz_dimensions:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__vert_dimensions:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__radial_dimensions:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__angular_dimensions:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__textblocks:
            _x, _y = _obj.getLocation()
            if _xmin < _x < _xmax and _ymin < _y < _ymax:
                _objs.append(_obj)
        return _objs

    def update(self):
        """Check that the objects in this layer are stored correctly.

update()

This function checks that the objects held in this layer are kept
in the proper order. Also, any duplicated objects that may have
be created due to modifying entities in the layer are removed.
        """
        if (self.__segment_tree.isModified() or
            self.__circle_tree.isModified() or
            self.__arc_tree.isModified() or
            self.__hcline_tree.isModified() or
            self.__vcline_tree.isModified() or
            self.__acline_tree.isModified() or
            self.__cline_tree.isModified() or
            self.__ccircle_tree.isModified()):
            self._sweepPointUsers()
        self._sweepPointTree()
        self._sweepSegmentTree()
        self._sweepCircleTree()
        self._sweepArcTree()
        self._sweepSPCLineTree(self.__hcline_tree)
        self._sweepSPCLineTree(self.__vcline_tree)
        self._sweepSPCLineTree(self.__acline_tree)
        self._sweepCLineTree()
        self._sweepCCircleTree()
        self._sweepLeaderList()
        self._sweepPolylineList()

    def bindObject(self, obj):
        """Make the object connect itself to the other objects it uses.

bindObject(obj)

The argument "obj" must be an object stored in the layer.
        """
        warnings.warn("bindObject() is deprecated - unneeded now",
                      DeprecationWarning, stacklevel=2)
        if id(obj) not in self.__objects:
            raise ValueError, "Object not in layer: " + `obj`
        if isinstance(obj, segment.Segment):
            for _pt in obj.getEndpoints():
                assert id(_pt) in self.__objects, "Segment point not in objects"
        elif isinstance(obj, arc.Arc):
            _cp = obj.getCenter()
            assert id(_cp) in self.__objects, "Arc center not in objects"
            for _ep in obj.getEndpoints():
                _lp = self.hasPoint(_ep)
                assert _lp is not None, "Arc endpoint not in objects"
        elif isinstance(obj, circle.Circle):
            _cp = obj.getCenter()
            assert id(_cp) in self.__objects, "Circle center not in objects"
        elif isinstance(obj, (hcline.HCLine,
                               vcline.VCLine,
                               acline.ACLine)):
            _lp = obj.getLocation()
            assert id(_lp) in self.__objects, "SPCLine point not in objects"
        elif isinstance(obj, cline.CLine):
            for _pt in obj.getKeypoints():
                assert id(_pt) in self.__objects, "CLine point not in objects"
        elif isinstance(obj, leader.Leader):
            for _pt in obj.getLeaderPoints():
                assert id(_pt) in self.__objects, "Leader point not in objects"
        elif isinstance(obj, polyline.Polyline):
            for _pt in obj.getPoints():
                assert id(_pt) in self.__objects, "Polyline point not in objects"
        elif isinstance(obj, (segjoint.Chamfer, segjoint.Fillet)):
            for _seg in obj.getSegments():
                assert id(_seg) in self.__objects, "Segjoint segment not in objects"
        elif isinstance(obj, dimension.LinearDimension):
            _l1, _l2 = obj.getDimLayers()
            _p1, _p2 = obj.getDimPoints()
            if _l1 is self:
                assert id(_p1) in self.__objects, "Dim P1 not in objects"
            if _l2 is self:
                assert id(_p2) in self.__objects, "Dim P2 not in objects"
        elif isinstance(obj, dimension.RadialDimension):
            _l = obj.getDimLayer()
            _c = obj.getDimCircle()
            if _l is self:
                assert id(_c) in self.__objects, "Dim circle not in objects"
        elif isinstance(obj, dimension.AngularDimension):
            _vl, _l1, _l2 = obj.getDimLayers()
            _vp, _p1, _p2 = obj.getDimPoints()
            if _vl is self:
                assert id(_vp) in self.__objects, "Dim vertex not in objects"
            if _l1 is self:
                assert id(_p1) in self.__objects, "Dim P1 not in objects"
            if _l2 is self:
                assert id(_p2) in self.__objects, "Dim P2 not in objects"
        else:
            pass #
        self._sweepPointUsers()

    def _sweepPointUsers(self):
        """Check that all objects associated with a point still do.

_sweepPointUsers()

This method is private to the Layer object.
        """
        for _pt in self.__point_tree:
            assert id(_pt) in self.__objects, "Lost point from objects"
            _users = _pt.getUsers()
            _freeobjs = []
            if _users is not None:
                for _user in _users:
                    if isinstance(_user, segment.Segment):
                        _p1, _p2 = _user.getEndpoints()
                        if _pt is not _p1 and _pt is not _p2:
                           _freeobjs.append(_user)
                    elif isinstance(_user, arc.Arc):
                        _cp = _user.getCenter()
                        _ep1, _ep2 = _user.getEndpoints()
                        if _pt is not _cp and _pt != _ep1 and _pt != _ep2:
                            _freeobjs.append(_user)
                    elif isinstance(_user, circle._Circle):
                        if _pt is not _user.getCenter():
                            _freeobjs.append(_user)
                    elif isinstance(_user, (hcline.HCLine,
                                            vcline.VCLine,
                                            acline.ACLine)):
                        _lp = _user.getLocation()
                        if _pt is not _lp:
                            _freeobjs.append(_user)
                    elif isinstance(_user, cline.CLine):
                        _p1, _p2 = _user.getKeypoints()
                        if _pt is not _p1 and _pt is not _p2:
                            _freeobjs.append(_user)
                    elif isinstance(_user, leader.Leader):
                        _p1, _p2, _p3 = _user.getLeaderPoints()
                        if ((_pt is not _p1) and
                            (_pt is not _p2) and
                            (_pt is not _p3)):
                            _freeobjs.append(_user)
                    elif isinstance(_user, polyline.Polyline):
                        if _pt not in _user.getPoints():
                            _freeobjs.append(_user)
                    elif isinstance(_user, dimension.Dimension):
                        if isinstance(_user, dimension.LinearDimension):
                            _p1, _p2 = _user.getDimPoints()
                            if _pt is not _p1 and _pt is not _p2:
                                _freeobjs.append(_user)
                        elif isinstance(_user, dimension.AngularDimension):
                            _p1, _p2, _p3 = _user.getDimPoints()
                            if (_pt is not _p1 and
                                _pt is not _p2 and
                                _pt is not _p3):
                                _freeobjs.append(_user)
            if len(_freeobjs):
                for _obj in _freeobjs:
                    print "Failed to free object from point: " + `_obj`
                    _pt.freeUser(_obj)

    def _sweepPointTree(self):
        """Validate the layer Point storage.

_sweepPointTree()

This method is private to the Layer object.
        """
        _modpts = self.__point_tree.getModified()
        if len(_modpts):
            _deadpts = self.__point_tree.clean()
            for _pt in _deadpts:
                _lp = self.__point_tree.binscan(_pt)
                assert _lp is not None, "lost point in point_tree: " + str(_pt)
                _users = _pt.getUsers()
                if _users is not None:
                    for _user in _users:
                        if isinstance(_user, segment.Segment):
                            _p1, _p2 = _user.getEndpoints()
                            if _pt is _p1:
                                _user.setP1(_lp)
                            elif _pt is _p2:
                                _user.setP2(_lp)
                            else:
                                pass
                        elif isinstance(_user, circle._Circle):
                            _user.setCenter(_lp)
                        elif isinstance(_user, (hcline.HCLine,
                                                vcline.VCLine,
                                                acline.ACLine)):
                            _user.setLocation(_lp)
                        elif isinstance(_user, cline.CLine):
                            _p1, _p2 = _user.getKeypoints()
                            if _pt is _p1:
                                _user.setP1(_lp)
                            elif _pt is _p2:
                                _user.setP2(_lp)
                            else:
                                pass
                        elif isinstance(_user, leader.Leader):
                            _p1, _p2, _p3 = _user.getLeaderPoints()
                            if _pt is _p1:
                                _user.setP1(_lp)
                            elif _pt is _p2:
                                _user.setP2(_lp)
                            elif _pt is _p3:
                                _user.setP3(_lp)
                            else:
                                pass
                        elif isinstance(_user, polyline.Polyline):
                            for _i in xrange(len(_user)):
                                _pi = _user.getPoint(_i)
                                if _pi is _pt:
                                    _user.setPoint(_i, _lp)
                                    break
                        elif isinstance(_user, dimension.Dimension):
                            if isinstance(_user, dimension.LinearDimension):
                                _p1, _p2 = _user.getDimPoints()
                                if _pt is _p1:
                                    _user.setP1(_lp)
                                elif _pt is _p2:
                                    _user.setP2(_lp)
                                else:
                                    pass
                            elif isinstance(_user, dimension.AngularDimension):
                                _cp, _p1, _p2 = _user.getDimPoints()
                                if _pt is _cp:
                                    _user.setCenterPoint(_lp)
                                elif _pt is _p1:
                                    _user.setP1(_lp)
                                elif _pt is _p2:
                                    _user.setP2(_lp)
                                else:
                                    pass
                        else:
                            raise TypeError, "Missing type for keypoint modification:" + `_user`
                _key = id(_pt)
                if _key in self.__externals:
                    print "point in externals: " + `_pt`
                    for _layer, _extobj in self.__externals[_key]:
                        if isinstance(_extobj, dimension.LinearDimension):
                            _p1, _p2 = _extobj.getDimPoints()
                            if _pt is _p1:
                                _extobj.setP1(_lp)
                            elif _pt is _p2:
                                _extobj.setP2(_lp)
                            else:
                                pass
                        elif isinstance(_extobj, dimension.RadialDimension):
                            _extobj.setCenterPoint(_lp)
                        elif isinstance(_extobj, dimension.AngularDimension):
                            _cp, _p1, _p2 = _extobj.getDimPoints()
                            if _pt is _cp:
                                _extobj.setCenterPoint(_lp)
                            elif _pt is _p1:
                                _extobj.setP1(_lp)
                            elif _pt is _p2:
                                _extobj.setP2(_lp)
                            else:
                                pass
            for _pt in _modpts:
                if _pt not in _deadpts:
                    _objlist = _pt.getUsers()
                    if _objlist is not None:
                        for _obj in _objlist:
                            _obj.modified()
                    _pt.reset()

    def _sweepSegmentTree(self):
        """Validate the layer segment storage.

_sweepSegmentTree()

This method is private to the Layer.
        """
        _modsegs = self.__segment_tree.getModified()
        if len(_modsegs):
            _deadsegs = self.__segment_tree.clean()
            for _seg in _modsegs:
                if _seg in _deadsegs:
                    continue
                if _seg.length() < 1e-10:
                    _deadsegs.append(_seg)
                    continue
                _seg.reset()
                for _pt in _seg.getEndpoints():
                    assert id(_pt) in self.__objects, "Segment point not in objects:" + `_pt`
            for _seg in _deadsegs:
                _keepseg = self.__segment_tree.binscan(_seg)
                assert _keepseg is not None, "Lost segment: " + str(_seg)
                _p1, _p2 = _seg.getEndpoints()
                _p1.freeUser(_seg)
                _p2.freeUser(_seg)
                _users = _seg.getUsers()
                if _users is not None: # fixme - hatching ...
                    for _user in _users:
                        self.delObject(_user)
                del self.__objects[id(_seg)]
                self.delObject(_p1)
                self.delObject(_p2)

    def _sweepCircleTree(self):
        """Validate the layer Circle storage.

_sweepCircleTree(spctree)

This method is private to the Layer.
        """
        _modcircs = self.__circle_tree.getModified()
        if len(_modcircs):
            _deadcircs = self.__circle_tree.clean()
            for _circ in _modcircs:
                if _circ in _deadcircs:
                    continue
                if _circ.getRadius() < 1e-10: # this probably can't happen
                    _deadcircs.append(_circ)
                    continue
                _circ.reset()
                _cp = _circ.getCenter()
                assert id(_cp) in self.__objects, "Circle center not in objects:" + `_cp`
                _users = _circ.getUsers()
                if _users is not None: # radial dimension or hatching
                    for _user in _users:
                        _user.modified()
                _key = id(_circ)
                if _key in self.__externals:
                    print "circle in externals: " + `_circ`
                    for _layer, _extobj in self.__externals[_key]:
                        _extobj.modified()
            for _circ in _deadcircs:
                _keepcirc = self.__circle_tree.binscan(_circ)
                assert _keepcirc is not None, "Lost Circle: " + `_circ`
                _key = id(_circ)
                if _key in self.__externals:
                    print "circle in externals: " + `_circ`
                    for _layer, _extobj in self.__externals[_key]:
                        _extobj.setDimCircle(_keepcirc)
                _users = _circ.getUsers()
                if _users is not None:
                    for _user in _users:
                        if isinstance(_user, dimension.RadialDimension):
                            _user.setDimCircle(_keepcirc)
                        else:
                            pass # fixme - hatching ...
                _cp = _circ.getCenter()
                _cp.freeUser(_circ)
                del self.__objects[id(_circ)]
                self.delObject(_cp)

    def _sweepArcTree(self):
        """Validate the layer Arc storage.

_sweepArcTree(spctree)

This method is private to the Layer.
        """
        _modarcs = self.__arc_tree.getModified()
        if len(_modarcs):
            _deadarcs = self.__arc_tree.clean()
            for _arc in _modarcs:
                if _arc in _deadarcs:
                    continue
                if _arc.getRadius() < 1e-10: # probably can't happen ...
                    _deadarcs.append(_arc)
                    continue
                _arc.reset()
                _cp = _arc.getCenter()
                assert id(_cp) in self.__objects, "Arc center not in objects:" + `_cp`
                for _ep in _arc.getEndpoints():
                    _lp = self.hasPoint(_ep)
                    if _lp is None:
                        print "Arc missing point: " + str(_arc)
                        print "Endpoint: " + str(_ep)
                        continue
                    assert id(_lp) in self.__objects, "Arc endpoint not in objects:" + `_lp`
                _users = _arc.getUsers()
                if _users is not None:
                    for _user in _users: # radial/angular dimension or hatching
                        _user.modified()
                _key = id(_arc)
                if _key in self.__externals:
                    print "arc in externals: " + `_arc`
                    for _layer, _extobj in self.__externals[_key]:
                        _extobj.modified()
            for _arc in _deadarcs:
                _keeparc = self.__arc_tree.binscan(_arc)
                assert _keeparc is not None, "Lost Arc: " + `_arc`
                _key = id(_arc)
                if _key in self.__externals:
                    print "arc in externals: " + `_arc`
                    for _layer, _extobj in self.__externals[_key]:
                        _extobj.setDimCircle(_keeparc)
                _users = _arc.getUsers()
                if _users is not None:
                    for _user in _users:
                        if isinstance(_user, dimension.RadialDimension):
                            _user.setDimCircle(_keeparc)
                        elif isinstance(_user, dimension.AngularDimension):
                            pass # fixme
                        else:
                            pass # fixme - hatching ...
                _cp = _arc.getCenter()
                _cp.freeUser(_arc)
                for _ep in _arc.getEndpoints():
                    _p = self.hasPoint(_ep)
                    _p.freeUser(_arc)
                    self.delObject(_p)
                del self.__objects[id(_arc)]
                self.delObject(_cp)

    def _sweepSPCLineTree(self, spctree):
        """Validate the storage of single point construction line objects.

_sweepSPCLineTree(spctree)

This method is private to the Layer.
        """
        _modspclines = spctree.getModified()
        if len(_modspclines):
            _deadspclines = spctree.clean()
            for _spcline in _modspclines:
                if _spcline not in _deadspclines:
                    _spcline.reset()
                    _pt =  _spcline.getLocation()
                    assert id(_pt) in self.__objects, "spcline point not in objects:" + str(_spcline)
            for _spcline in _deadspclines:
                _pt = _spcline.getLocation()
                _pt.freeUser(_spcline)
                del self.__objects[id(_spcline)]
                self.delObject(_pt)

    def _sweepCLineTree(self):
        """Validate the layer CLine storage.

_sweepCLineTree()

This method is private to the Layer.
        """
        _modclines = self.__cline_tree.getModified()
        if len(_modclines):
            _deadclines = self.__segment_tree.clean()
            for _cline in _modclines:
                if _cline not in _deadclines:
                    _cline.reset()
                    for _pt in _cline.getKeypoints():
                        assert id(_pt) in self.__objects, "CLine keypoint not in objects:" + `_pt`
            for _cline in _deadclines:
                _p1, _p2 = _cline.getKeypoints()
                _p1.freeUser(_cline)
                _p2.freeUser(_cline)
                del self.__objects[_id(_cline)]
                self.delObject(_p1)
                self.delObject(_p2)

    def _sweepCCircleTree(self):
        """Validate the layer CCircle storage.

_sweepCCircleTree()

This method is private to the Layer object.
        """
        _modccs = self.__ccircle_tree.getModified()
        if len(_modccs):
            _deadccs = self.__ccircle_tree.clean()
            for _cc in _modccs:
                if _cc in _deadccs:
                    continue
                if _cc.getRadius() < 1e-10: # probably not possible
                    _deadccs.append(_cc)
                else:
                    _cc.reset()
                    _cp = _cc.getCenter()
                    assert id(_cp) in self.__objects, "CCircle center not in objjects:" + `_cp`
            for _cc in _deadccs:
                _cp = _cc.getCenter()
                _cp.freeUser(_cc)
                del self.__objects[id(_cc)]
                self.delObject(_cp)

    def _sweepLeaderList(self):
        """Validate the layer Leader storage

_sweepLeaderList()

This method is private to the Layer.
        """
        _mods = []
        for _leader in self.__leaders:
            if _leader.isModified():
                _mods.append(_leader)
        if len(_mods):
            self._sweepPointUsers()
            for _leader in _mods:
                _leader.reset()
                _p1, _p2, _p3 = _leader.getLeaderPoints()
                assert id(_p1) in self.__objects, "Leader P1 point not in objects."
                assert id(_p2) in self.__objects, "Leader P2 point not in objects."
                assert id(_p3) in self.__objects, "Leader P3 point not in objects."
    def _sweepPolylineList(self):
        """Validate the layer Polyline storage

_sweepPolylineList()

This method is private to the Layer.
        """
        _mods = []
        for _polyline in self.__polylines:
            if _polyline.isModified():
                _mods.append(_polyline)
        if len(_mods):
            self._sweepPointUsers()
            for _polyline in _mods:
                _polyline.reset()
                for _pt in _polyline.getPoints():
                    assert id(_pt) in self.__objects, "Polyline point not in objects."
