#
# Copyright (c) 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# classes for polyline objects
#

import graphicobject
import tolerance
import style
import linetype
import color
import point
import baseobject
import util

class _Polyline(object):
    """A class representing a polyline. A polyline is essentially
a number of segments that connect end to end.

A _Polyline has the following methods:

getPoints(): Return the points of the _Polyline.
{get/set}Point(): Get/Set one of the points of the _Polyline
addPoint(): Add a new point into the _Polyline.
delPoint(): Remove a point from the _Polyline.
move(): Move the _Polyline.
length(): Get the _Polyline length.
mapPoint(): Test if a point is within some distance to a _Polyline.
mapCoords(): Test if a coordinate pair is within some distance to a _Polyline.
inRegion(): Test if the _Polyline is visible in some area.
clone(): Make an identical copy of a _Polyline.
    """

    # member functions

    def __init__(self, plist):
        """Initialize a _Polyline object.

_Polyline(plist)

The argument "plist" is a list object containing _Point objects.
There should be two or more _Points in the list.
        """
        if not isinstance(plist, list):
            raise TypeError, "Invalid list: " + str(plist)
        _pts = []
        _count = len(plist)
        if _count < 2:
            raise ValueError, "Invalid list count: %d" % _count
        for _i in xrange(_count):
            _obj = plist[_i]
            if not isinstance(_obj, point._Point):
                _obj = point._Point(plist[_i])
            _pts.append(_obj)
        _pmap = {}
        for _pt in _pts:
            _seen_pt = None
            for _mpt in _pmap:
                if _mpt == _pt:
                    _seen_pt = _mpt
                    break
            _pmap[_pt] = _seen_pt
        _plist = []
        for _pt in _pts:
            _mpt = _pmap[_pt]
            if _mpt is None:
                _plist.append(_pt)
            else:
                _plist.append(_mpt)
        if len(_plist) < 2:
            raise ValueError, "Invalid point list: " + str(plist)
        self.__pts = _plist
        self.__modified = False

    def __len__(self):
        return len(self.__pts)

    def __str__(self):
        return "_Polyline" # fixme

    def __eq__(self, pline):
        """Compare two _Polyline objects for equality.
        """
        if not isinstance(pline, _Polyline):
            raise TypeError, "Invalid _Polyline: " + str(pline)
        _val = False
        _ppts = pline.getPoints()
        _pcount = len(_ppts)
        _spts = self.__pts
        _scount = len(_spts)
        if _pcount == _scount:
            _val = True
            for _i in xrange(_scount):
                if _ppts[_i] != _spts[_i]:
                    _val = False
                    break
        return _val

    def __ne__(self, pline):
        """Compare two _Polyline objects for inequality.
        """
        if not isinstance(pline, _Polyline):
            raise TypeError, "Invalid _Polyline: " + str(pline)
        _val = True
        _ppts = pline.getPoints()
        _pcount = len(_ppts)
        _spts = self.__pts
        _scount = len(_spts)
        if _pcount == _scount:
            _val = False
            for _i in xrange(_scount):
                if _ppts[_i] != _spts[_i]:
                    _val = True
                    break
        return _val

    def getPoints(self):
        """Get the points of the _Polyline.

getPoints()

This function returns a list containing all the _Point
objects that define the _Polyline.
        """
        return self.__pts[:]

    def getNumPoints(self):
        """Return the number of _Point objects defining the _Polyline.

getNumPoints()
        """
        return len(self)

    def getPoint(self, i):
        """Return a single _Point object used for defining the _Polyline.

getPoint(i)

The argument "i" must be an integer, and its value represents
the i'th _Point used to define the _Polyline.
        """
        return self.__pts[i]

    def setPoint(self, i, p):
        """Set a _Point of the _Polyline.

setPoint(i, p)

The argument "i" must be an integer, and its value represents
the i'th _Point used to define the _Polyline. Argument "p"
must be a _Point.
        """
        if not isinstance(p, point._Point):
            raise TypeError, "Invalid _Point for _Polyline point: " + str(p)
        self.__pts[i] = p

    def addPoint(self, i, p):
        """Add a _Point to the _Polyline.

addPoint(i, p)

The argument "i" must be an integer, and argument "p" must be a
_Point. The _Point is added into the list of points comprising
the _Polyline as the i'th point.
        """
        if not isinstance(p, point._Point):
            raise TypeError, "Invalid _Point for _Polyline point: " + str(p)
        self.__pts.insert(i, p)

    def delPoint(self, i):
        """Remove a _Point from the _Polyline.

delPoint(i)

The argument "i" represents the index of the point to remove from
the list of points defining the _Polyline. The point will be
removed only if the polyline will still have at least two _Points.
        """
        if len(self.__pts) > 2:
            del self.__pts[i]

    def move(self, dx, dy):
        """Move a _Polyline.

move(dx, dy)

The first argument gives the "x" coordinate displacement,
and the second gives the "y" coordinate displacement. Both
values should be floats.
        """
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        for _pt in self.__pts:
            _pt.move(_dx, _dy)

    def length(self):
        """Return the length of the _Polyline.

length()

The length is the sum of the lengths of all the sub-segments
in the _Polyline
        """
        _length = 0.0
        _pts = self.__pts
        _count = len(_pts) - 1
        for _i in xrange(_count):
            _sublength = _pts[_i + 1] - _pts[_i]
            _length = _length + _sublength
        return _length

    def getBounds(self):
        """Return the bounding rectangle around a _Polyline.

getBounds()

This method returns a tuple of four values:

(xmin, ymin, xmax, ymax)
        """
        _pts = self.__pts
        _pxmin = None
        _pymin = None
        _pxmax = None
        _pymax = None
        for _pt in _pts:
            _px, _py = _pt.getCoords()
            if _pxmin is None or _px < _pxmin:
                _pxmin = _px
            if _pymin is None or _py < _pymin:
                _pymin = _py
            if _pxmax is None or _px > _pxmax:
                _pxmax = _px
            if _pymax is None or _py > _pymax:
                _pymax = _py
        return _pxmin, _pymin, _pxmax, _pymax

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest _Point on the _Polyline to the point p.

mapPoint(p [, tol])

The function has one required argument:

p: A _Point object or a tuple of two floats

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by _Point object to an
actual _Point on the _Polyline. If the distance between the actual
_Point and the _Point used as an argument is less than the Tolerance,
the actual _Point is returned. Otherwise, this function returns None.
        """
        _p = p
        if not isinstance(_p, point._Point):
            _p = point._Point(p)
        _t = tolerance.toltest(tol)
        _x, _y = _p.getCoords()
        return self.mapCoords(_x, _y, _t)

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest _Point on the _Polyline by the x/y coordinates.

mapCoords(x, y[, tol])

The function has two required arguments:

x: A Float value giving the "x" coordinate
y: A Float value giving the "y" coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by coordinate pair to a
_Point object on the Polyline. If the distance between the actual
_Point and the coordinates used as an argument is less than the tolerance,
the actual _Point is returned. Otherwise, this method returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _count = len(self.__pts) - 1
        _mp = None
        for _i in xrange(_count):
            _x1, _y1 = self.__pts[_i].getCoords()
            _x2, _y2 = self.__pts[_i + 1].getCoords()
            _pt = util.map_coords(_x, _y, _x1, _y1, _x2, _y2, _t)
            if _pt is not None:
                _px, _py = _pt
                _mp = point._Point(_px, _py)
                break
        return _mp

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a _Polyline exists within a region.

isRegion(xmin, ymin, xmax, ymax[, fully])

The four arguments define the boundary of an area, and the
method returns True if the _Polyline lies within that area. If
the optional argument "fully" is used and is True, then both
endpoints of the _Polyline must lie within the boundary.
Otherwise, the method returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        # print "xmin: %g" % _xmin
        # print "ymin: %g" % _ymin
        # print "xmax: %g" % _xmax
        # print "ymax: %g" % _ymax
        _pxmin, _pymin, _pxmax, _pymax = self.getBounds()
        # print "pxmin: %g" % _pxmin
        # print "pymin: %g" % _pymin
        # print "pxmax: %g" % _pxmax
        # print "pymax: %g" % _pymax
        if (_xmin < _pxmin and _xmax > _pxmax and
            _ymin < _pymin and _ymax > _pymax):
            return True
        if not fully:
            _val = False
            _pts = self.__pts
            _segs = len(_pts) - 1
            for _i in xrange(_segs):
                _x1, _y1 = _pts[_i - 1].getCoords()
                _x2, _y2 = _pts[_i].getCoords()
                _val = util.in_region(_x1, _y1, _x2, _y2,
                                      _xmin, _ymin, _xmax, _ymax)
                if _val:
                    break
        return _val

    def clone(self):
        """Create an identical copy of a _Polyline.

clone()
        """
        _cppts = []
        for _pt in self.__pts:
            _cppts.append(_pt.clone())
        return _Polyline(_cppts)

#
# Polyline defaults
#

def_linetype = linetype.Linetype('Solid_Default', None)
def_color = color.Color(0xffffff)
def_thickness = 1.0

def_pline_style = style.Style("Polyline_Default", def_linetype,
                               def_color, def_thickness)

class Polyline(_Polyline, baseobject.ModObject, baseobject.Subpart,
               graphicobject.GraphicObject):
    """The storable and displayed representation of _Polyline objects.

The Polyline class is derived from _Polyline, ModObject, and
GraphicObject classes, so it shares all the attributes and
methods of those classes.
    """

    #
    # The default style for the Polyline class
    #
    
    __defstyle = style.Style(u'Polyline Default Style',
                            linetype.Linetype(u'Solid', None),
                            color.Color(0xffffff),
                            1.0)

    def __init__(self, plist, st=None, lt=None, col=None, t=None):
        """Initialize a Polyline.

Polyline(plist[, st, lt, col, t])

There is one required argument:

plist: A list of Point objects giving the points in the Polyli

The following arguments are optional:
st: A Style object
lt: A Linetype object that overrides the Style linetype
col: A Color object that overrides the Style color
t: A float that overrides the Style thickness
        """
        if not isinstance(plist, list):
            raise TypeError, "Invalid list: " + str(plist)
        _count = len(plist)
        if _count < 2:
            raise ValueError, "Invalid list count: %d" % _count
        _pts = []
        for _i in xrange(_count):
            _obj = plist[_i]
            if not isinstance(_obj, point.Point):
                _obj = point.Point(plist[_i])
            _pts.append(_obj)
        _st = st
        if _st is None:
            _st = Polyline.__defstyle
        baseobject.ModObject.__init__(self)
        baseobject.Subpart.__init__(self)
        _Polyline.__init__(self, _pts)
        graphicobject.GraphicObject.__init__(self, _st, lt, col, t)
        for _pt in _pts:
            _pt.storeUser(self)

    def isModified(self):
        """Indicate if the Polyline has been changed.

isModified()

This method extends the ModObject:isModified() method.
        """
        for _pt in self.getPoints():
            if _pt.isModified():
                self.modified()
                break
        return baseobject.ModObject.isModified(self)

    def setPoint(self, i, p):
        """Set a Point in the Polyline.

setPoint(i, p)

This method extends _Polyline::setPoint() and sets the modified
flag indicating that the Polyline has been altered.
        """
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _pt = self.getPoint(i)
        _Polyline.setPoint(self, i, p)
        self.modified()
        _pt.freeUser(self)
        p.storeUser(self)

    def addPoint(self, i, p):
        """Add a Point to the Polyline.

addPoint(i, p)

This method extends _Polyline::addPoint() and sets the modified
flag indicating the Polyline has been altered.
        """
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _Polyline.addPoint(self, i, p)
        self.modified()
        p.storeUser(self)

    def delPoint(self, i):
        """Remove a Point from the Polyline.

delPoint(i)

This method extends _Polyline::delPoint() and sets the modified
flag indicating the Polyline has been altered.
        """
        _pt = self.getPoint(i)
        _Polyline.delPoint(self, i)
        self.modified()
        _pt.freeUser(self)

    def move(self, dx, dy):
        """Move a Polyline.

move(dx, dy)

This method extends _Polyline::move() and sets the modified
flag indicating that the Polyline has been altered.
        """
        _Polyline.move(self, dx, dy)
        self.modified()

    def clone(self):
        """Create an identical copy of a Polyline

clone()

This method overrides the _Polyline::clone() method.
        """
        _cpts = []
        for _pt in self.getPoints():
            _cpts.append(_pt.clone())
        _st = self.getStyle()
        _lt = self.getLinetype()
        _col = self.getColor()
        _th = self.getThickness()
        return Polyline(_cpts, _st, _lt, _col, _th)

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the Polyline to the point p.

mapPoint(p [, tol])

This function extends _Polyline::mapPoint().
        """
        _pt = _Polyline.mapPoint(self, p, tol)
        if _pt is not None:
            _x, _y = _pt.getCoords()
            _pt = point.Point(_x, _y)
        return _pt

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the Polyline to a coordinate pair.

mapCoords(x, y[, tol])

This function extends _Polyline::mapCoords().
        """
        _pt = _Polyline.mapCoords(self, x, y, tol)
        if _pt is not None:
            _x, _y = _pt.getCoords()
            _pt = point.Point(_x, _y)
        return _pt
