#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# vertical construction lines
#

import conobject
import tolerance
import point
import util
import maptree

class VCLine(conobject.Spcline):
    """A class for Vertical construction lines.

A VCLine object is derived from an Spcline,so it has
all that objects properties.

A VCLine has the following addtional methods:

mapPoint(): Test if a point is within some distance to a VCLine.
mapCoords(): Test if a coordinate pair is within some distance to a VCLine.
inRegion(): Return whether or not a VCLine passes through a bounded region.
clone(): Return an identical copy of a VCLine.
    """

    # class methods
    
    def __init__(self, p):
        """Initialize an VCLine object.

VCLine(p)

The single argument should be a Point object.
        """
        conobject.Spcline.__init__(self, p)
        
    def __cmp__(self, obj):
        """Compare one VCLine object to another or to a Point.

The comparison of two VCLines is simple - the "larger"
of the two is the one with the greater "x" coordinate
value of their Point. If both values are within the
tolerance, the lines are equivalent.

Comparing an VCLine to a Point is equally simple.
If the "x" coordinate of the VCLine is greater than
the Point, return 1. If it is less, return -1. Otherwise,
return 0.
        """
        if not isinstance(obj, (point.Point, VCLine)):
            raise TypeError, "Invalid object for VCLine comparison: " + `obj`
        _xs, _ys = self.getLocation().getCoords()
        if isinstance(obj, VCLine):
            _x, _y = obj.getLocation().getCoords()            
        else:
            _x, _y = obj.getCoords()
        if abs(_xs - _x) < 1e-10:
            _value = 0
        elif _xs > _x:
            _value = 1
        else:
            _value = -1
        return _value

    def __str__(self):
        _x, _y = self.getLocation().getCoords()
        return "Vertical construction line at x = %g" % _x

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the VCLine to the point p.

mapCoords(x, y[, tol])

The function has two required arguments:

x: A Float value giving the "x" coordinate
y: A Float value giving the "y" coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.

This function is used to map a possibly near-by coordinate pair to
an actual Point on the VCLine. If the distance between the actual
Point and the coordinates used as an argument is less than the Tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _vx, _vy = self.getLocation().getCoords()
        _diff = abs(_x - _vx)
        _ap = None
        if _diff < _t or abs(_diff - _t) < 1e-10:
            _ap = point.Point(_vx, _y)
        return _ap

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the VCLine to the point p.

mapPoint(p [, tol])

The function has one required argument:

p: A Point object or a tuple of two floats

There is a single optional argument:

tol: A float value equal or greater than 0.

This function is used to map a possibly near-by Point object to an
actual Point on the VCLine. If the distance between the actual
Point and the Point used as an argument is less than the Tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _p = p
        if not isinstance(_p, point.Point):
            _p = point.Point(p)
        _t = tolerance.toltest(tol)
        _x, _y = _p.getCoords()
        return self.mapCoords(_x, _y, _t)

    def inRegion(self, xmin, ymin, xmax, ymax):
        """Return whether or not an VCLine passes through a region.

inRegion(xmin, ymin, xmax, ymax)

The first four arguments define the boundary. The method
will return True if the VCLine falls between xmin and xmax.
Otherwise the function will return False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _x, _y = self.getLocation().getCoords()
        return _xmin < _x < _xmax

    def clone(self):
        """Return a copy of a VCLine.

clone()
        """
        _loc = self.getLocation().clone()
        return VCLine(_loc)

    def clipToRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _x, _y = self.getLocation().getCoords()
        _coords = None
        if _xmin < _x < _xmax:
            _coords = (_x, _ymin, _x, _ymax)
        return _coords

class VCLineTree(maptree.PointMapTree):
    """A Tree class specific for VCLine objects.

This class is based on the PointMapTree class, and
holds only VCLine objects.
    """
    def __init__(self):
        """Initialize a VCLineTree.

VCLineTree()
        """
        maptree.PointMapTree.__init__(self, VCLine)

    def mapPoint(self, p, tol=tolerance.TOL):
        """See if an object in the VCLineTree exists at a Point.

mapPoint(p [, tol])

This method has one required parameter:

p: Either a Point object or a tuple of two-floats

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if the
     point is close enough to the objects held in the Tree.

The function returns a list of at tuples, each of the form

(obj, pt)

obj: Object that the Point is mapped to
pt: The projected Point on the object

The list will contain at most one tuple.
        """
        _p = p
        if not isinstance(_p, (point.Point, tuple)):
            raise TypeError, "Invalid type for searching in PointMapTree: " + `_p`            
        if isinstance(_p, tuple):
            _x, _y = util.tuple_to_two_floats(_p)
            _p = point.Point(_x, _y)
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "VCLineTree in modified state."
        _px, _py = _p.getCoords()
        _objlist = []
        _scanlist = []
        _lo = 0
        _hi = len(self)
        while _lo < _hi:
            _mid = (_hi+_lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _vcl = self[_mid]
            _vx, _vy = _vcl.getLocation().getCoords()
            if _px > _vx + _t + 1e-10:
                _lo = _mid + 1
            elif _px < _vx - _t - 1e-10:
                _hi = _mid
            else:
                _pt = _vcl.mapPoint(_p, _t)
                if _pt is not None:
                    _objlist.append((_vcl, _pt))
                break
        return _objlist

    def binscan(self, obj, tol=tolerance.TOL):
        """Find a VCLine object held in a VCLineTree via binary search.

binscan(obj [, tol])

This method has one required parameter:

obj: A VCLine object

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     VCLine objects are close enough to be considered equal.

If a VCLine object is found in the VCLineTree within the
tolerance, that object is returned. Otherwise, the function
returns None.
        """
        if not isinstance(obj, VCLine):
            raise TypeError, "Invalid type for searching in VCLineTree: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "VCLineTree in modified state."
        _lo = 0
        _hi = len(self)
        _scanlist = []
        _ro = None
        while _lo < _hi:
            _mid = (_hi+_lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _vcl = self[_mid]
            _res = cmp(obj, _vcl)
            if _res == 0:
                _ro = _vcl
                break
            elif _res == 1:
                _lo = _mid + 1
            else:
                _hi = _mid
        if _ro is None:
            _x, _y = obj.getLocation().getCoords()
            _xmin = _x - _t - 1e-10
            for _i in xrange(_mid, -1, -1):
                _vcl = self[_i]
                _vx, _vy = _vcl.getLocation().getCoords()
                if _vx < _xmin:
                    break
                _xdiff = abs(_vx - _x)
                if _xdiff < _t or abs(_xdiff - _t) < 1e-10:
                    _ro = _vcl
                    break
        return _ro

    def scan(self, obj, tol=tolerance.TOL):
        """Find a VCLine object held in a VCLineTree via linear search.

scan(obj [, tol])

This method has one required parameter:

obj: An VCLine object

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     VCLine objects are close enough to be considered equal.

If a VCLine object is found in the VCLineTree within the
tolerance, that object is returned. Otherwise, the function
returns None.
        """
        if not isinstance(obj, VCLine):
            raise TypeError, "Invalid type for searching in VCLineTree: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "VCLineTree in modified state."
        _ro = None
        _x, _y = obj.getLocation().getCoords()
        _xmax = _x + _t + 1e-10
        for _sobj in self:
            _sx, _sy = _sobj.getLocation().getCoords()
            if _sx > _xmax:
                break
            _xdiff = abs(_sx - _x)
            if _xdiff < _t or abs(_xdiff - _t) < 1e-10:
                _ro = _sobj
                break
        return _ro

    def atCoords(self, x):
        """Find a VCLine in the VCLineTree at a certain location.

atCoords(x)

This method has one required argument:

x: A float value representing the "x" coordinate

If there is a VCLine in the VCLineTree at the designated
coordinates, it is returned. Otherwise this method returns None.

This method is similar to the binscan() method.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        if self.isModified():
            raise ValueError, "VCLineTree in modified state."
        _scanlist = []
        _lo = _mid = 0
        _hi = len(self)
        _ro = None
        while _lo < _hi:
            _mid = (_hi+_lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _vcl = self[_mid]
            _px, _py = _vcl.getLocation().getCoords()
            if abs(_px - _x) < 1e-10:
                _ro = _vcl
                break
            elif _px < _x:
                _lo = _mid + 1
            else:
                _hi = _mid
        return _ro

def intersect_region(vcl, xmin, ymin, xmax, ymax):
    if not isinstance(vcl, VCLine):
        raise TypeError, "Invalid VCLine: " + str(vcl)
    _xmin = xmin
    if not isinstance(_xmin, float):
        _xmin = float(xmin)
    _ymin = ymin
    if not isinstance(_ymin, float):
        _ymin = float(ymin)
    _xmax = xmax
    if not isinstance(_xmax, float):
        _xmax = float(xmax)
    if _xmax < _xmin:
        raise ValueError, "Illegal values: xmax < xmin"
    _ymax = ymax
    if not isinstance(_ymax, float):
        _ymax = float(ymax)
    if _ymax < _ymin:
        raise ValueError, "Illegal values: ymax < ymin"
    _x, _y = vcl.getLocation().getCoords()
    _x1 = _y1 = _x2 = _y2 = None
    if _xmin < _x < _xmax:
        _x1 = _x
        _y1 = _ymin
        _x2 = _x
        _y2 = _ymax
    return _x1, _y1, _x2, _y2
